<?php
/**
 * @file
 * Contains DrunkinsFetcherInterface and DrunkinsFetcher.
 */

/**
 * Interface for fetcher classes that work with Drunkins Job implementations.
 *
 * Using a separate class for fetching items is optional; fetching can just be
 * done inside a DrunkinsJobInterface::start() implementation.
 *
 * Reasons to use a separate fetcher class: code generalization, and preventing
 * issues related to multiple inheritance: if one class extends the processing
 * part with custom code, and another extends the fetching part, your job class
 * cannot use the code from both these classes - unless the fetching part is
 * coded as a DrunkinsFetcherInterface.
 *
 * So job classes which care about their own extensibility, can use (or
 * implement fetching code in) an encapsulated fetcher class, to fetch items.
 * Job classes which don't care, can ignore fetchers for simplicity.
 *
 *
 * About the interface definition:
 *
 * 1.
 * A fetcher uses the same method names as a job, at this moment, which means
 * - a job class could also be used/encapsulated (instead of a fetcher class) by
 *   another job class, for fetching items;
 * - a job class can extend an existing fetcher class (if all it needs to do is
 *   implement the processItem() part).
 * Maybe these things are nice in practice, for quicker code reuse/extension by
 * site-specific job classes; maybe it's unnecessary (because there are obvious
 * other ways of doing these 2 things). At any rate, for now a fetcher does a
 * subset of what a job can do - so until that changes, it uses the same method
 * names too and DrunkinsFetcherInterface is a 'parent'
 * of DrunkinsJobInterface.
 *
 * 2.
 * finish() is defined here as a formalization of the fact that a job class
 * which encapsulates a fetcher, can/should call its fetcher's finish() job
 * (because it may need to do accounting, cleanup, or whatever other action.)
 * However most fetchers won't need finish() and can implement an empty method.
 */
interface DrunkinsFetcherInterface {

  /**
   * NOT part of the formal DrunkinsFetcherInterface specification!
   * This is likely to be moved into some separate UI class but we don't know
   * yet how. Until then you are free to implement settingsForm() in the fetcher
   * or job class, and it will be used. You can expect this to change at some
   * point though.
   *
   * Settings form for 'batch' type processes. (Batch processes can alter
   * runtime settings through a UI, unlike 'queue' type processes.)
   *
   * Fetcher classes that don't have any settings, should make this method
   * simply 'return array();'
   *
   * @return
   *   A FormAPI array.
   */
  //public function settingsForm();

  /**
   * Initializes the environment and context, and fetches the data to process.
   *
   * Any error encountered should be raised by throwing an exception. No custom
   * exception classes have been defined (yet?) for this and code using fetcher/
   * job classes should be able to handle any kind of exception.
   *
   * @param array $context
   *   Context for the process, passed by reference for (further)
   *   initialization. As a rule of thumb, the drunkins module code (to be:
   *   runner class) may use array keys prefixed with "drunkins_", so that
   *   'namespace' should not be used except for purposes defined by the runner.
   *
   * @return array
   *   An array of items to process, or an empty array if there were no items
   *   but there was no error. It should be assumed that array keys are ignored
   *   by the caller, unless otherwise agreed between caller and class (by some
   *   mechanism outside this specification).
   *
   * @throws \RuntimeException
   *   For any error that is supposed to prevent the process from executing.
   *   (Returning an empty array will also not execute further, but will do
   *   improper logging.)
   */
  public function start(array &$context);

  /**
   * Finalizes processing after a process has finished.
   *
   * Any error encountered should be raised by throwing an exception. No custom
   * exception classes have been defined (yet?) for this and code using fetcher/
   * job classes should be able to handle any kind of exception.
   *
   * There is no formal rule regarding whether this method should be called when
   * a process throws an exception and/or returns zero items; it's up to the
   * caller to work this out (and/or agree on this with the class
   * implementation; there is no established way of doing this yet e.g. by
   * 'capabilities' and/or runner settings.)
   * (The current drunkins.module -procedural implementation of 'runner' code-
   * does not call finish() when an exception is thrown, also not when start()
   * returns 0 items, except only for 'batch' type processing where start() sets
   * 'drunkins_process_more' context.)
   *
   * @param array $context
   *   Context for the process.
   *
   * @return string
   *   Result message, which the caller can print / log.
   *
   * @throws \RuntimeException
   *   For any error that prevents the process from finishing cleanly. (It is up
   *   to the caller to handle this as desired.)
   */
  public function finish(array &$context);

}

//@todo use an Exception type with args/tokens/translation context, in various places.

/**
 * Example/base class implementing some helper code useful for most fetchers.
 *
 * It's totally optional to use a separate class for fetching; item-fetching
 * code can be implemented in a job's start() method too.
 *
 * This class uses the following 'context keys', which are meant to be used by
 * child classes:
 * - fetcher_timestamp_from (int): timestamp set by the previous process run, OR
 *   overridden through 'fetcher_timestamp_from_date' setting (often in the UI),
 *   if 'opt_fetcher_timestamp' is on. This could also be set by child classes'
 *   start() in which case this class will leave it alone (regardless the
 *   'opt_fetcher_timestamp' setting, so this is a way for child classes to
 *   'cheat' that setting) - except if items are retrieved from cache.
 * - fetcher_timestamp (int): timestamp of fetching items; set just before
 *   fetching. Not set if items are retrieved from cache or if only a subset of
 *   fetched items will be processed (because the primary use of this value is
 *   storing it for the next run, after our run succeeds.) Should not be set by
 *   child classes.
 */
class DrunkinsFetcher implements DrunkinsFetcherInterface {

  /**
   * The job instance from which this class is instantiated. This property is
   * not set if this class acts as a parent class for some job which extends it;
   * only when used as a separate fetcher class. (And then only when the job
   * class actually thinks of passing itself to the constructor.)
   * It is a way to access properties/methods of the calling class; see log()
   * as an example.
   *
   * This does not introduce cross references because a fetcher instance
   * - is not remembered in the job class (only its name);
   * - usually has a short lifespan, for only one call.
   *
   * @var DrunkinsJobInterface
   */
  protected $jobInstance;

  /**
   * Settings; see DrunkinsJob.
   *
   * Settings used by this class' settingsForm() / start() method:
   * - job_id (string):
   *   Job ID, always present. See DrunkinsJob comments.
   *
   * - pkey (boolean):
   *   Functions as the 'primary key in the source item', for extending fetcher/
   *   job classes that need this. Influences behavior of fetcher_single_item
   *   (and a description on the settingsForm) in this class. See start().
   *
   * - opt_single_item (boolean, interpreted as TRUE by default / if not set):
   *   Enables 'fetcher_single_item' setting in settingsForm.
   * - fetcher_single_item (boolean):
   *   If TRUE/nonempty, limits processing to one specific item.
   *
   * - opt_limit (boolean, interpreted as TRUE by default / if not set):
   *   Enables 'fetcher_limit' setting in settingsForm.
   * - fetcher_limit (int):
   *   The maximum number of items to process. If there are more  items, these
   *   will be discarded.
   *
   * - opt_fetcher_timestamp (boolean|string):
   *   Enables use of a timestamp (by subclasses), which gets updated after each
   *   successful job run. This is useful for implementing jobs that need to get
   *   all items changed since the last run. If this is not a 'true' value, then
   *   all other fetcher_timestamp_* settings are ignored (though child classes
   *   could still manage to get start() to use a timestamp). For UI purposes
   *   this can be a string; the value will be used for the untranslated label
   *   of the date input element.
   * - fetcher_timestamp_ignore (boolean, default FALSE):
   *   If true, ignore the 'current set/stored timestamp' for this job, i.e.
   *   start from the beginning of time. (If no timestamp is set for the current
   *   process or stored from the last process, then providing this option is
   *   probably required to 'kick off' the first process.) Note that 'ignored'
   *   does not mean the _timestamp functionality_ is disabled (because e.g. at
   *   the end of the process, a timestamp will still be stored for the next
   *   process). It only means that the _use of the timestamp_ is disabled. So
   *   in practice, two settings govern whether a timestamp is used:
   *   opt_fetcher_timestamp must be true and fetcher_timestamp_ignore must be
   *   false or not set. To provide more detail, the current usage pattern is:
   *   - fetcher_timestamp_ignore is almost exclusively used as a UI option, for
   *     the user to be able to explicitly say that this functionality should
   *     not be used. It could be set in job configurations, if the use of a
   *     timestamp should be possible through the UI but is disabled by default.
   *   - opt_fetcher_timestamp is meant:
   *     - to enable exposing the _ignore/_from_date options in the UI. (For
   *       this, it can be a string.)
   *     - to enable using / check whether the functionality is used by the
   *       class, e.g. in job configurations. (Regardless of UI; if called
   *       without UI, a timestamp is retrieved from somewhere and
   *       fetcher_timestamp_ignore is supposed to be false. Note that this is
   *       probably the only opt_* setting which influences / is checked in the
   *       actual process instead of just being used to enable something in the
   *       UI.
   *   - checking 'whether the functionality is enabled/used' depends on your
   *     view of what 'enabled' entails, which may not be as straightforward as
   *     it seems. Extensive comments are at helper method getStartTimestamp().
   * - fetcher_timestamp_from_date (string):
   *   Date expression; if set, this will be used as the "fetch items changed
   *   since " date (and be the basis for the 'fetcher_timestamp_from' context
   *   value) instead of whatever 'rolling timestamp' a job stored somewhere.
   * - fetcher_timestamp_overlap (int):
   *   Number of seconds to rewind the stored 'rolling timestamp' (if not
   *   overridden by the fetcher_timestamp_from_date setting) to accommodate for
   *   lag in the remote system which could otherwise cause items to be missed.
   *
   * - opt_cache_items (boolean, interpreted as TRUE by default / if not set):
   *   Enables 'cache_items' setting in settingsForm.
   * - cache_items (int):
   *   The way in which items are cached / currently cached items are used.
   *   Values 0-3, depending on presence of cached items.
   *   This may be useful for a developer testing a custom DrunkinsJob
   *   implementation, if fetching items is slow.
   *   Needs CTools enabled. (We use CTools since it includes the session ID in
   *   the cache key/we don't need to think about clearing stale cache entries.)
   *
   * @var array
   */
  protected $settings;

  /**
   * Constructor function.
   *
   * If used as a fetcher encapsulated by a job class, this constructor is not
   * (or should not be) called on every instantiation of the job class, but only
   * just before calling any of its public functions.
   *
   * @param array $settings
   *   The initial process settings. See property definition, here or in the
   *   job class (or comments at the top of extending classes, hopefully) for
   *   supported keys.
   * @param DrunkinsJobInterface $job
   *   The job instance from which this class is instantiated. (If in fact this
   *   is instantiated from a job; if it is used as a parent class for a job,
   *   this would be NULL.)
   */
  public function __construct(array $settings = array(), $job = NULL) {

    // Remember the calling job class instance so we can call its methods from
    // settingsForm() / start() if needed.
    // @todo change this when we implement a proper logger.
    if (isset($job)) {
      $this->jobInstance = $job;
    }
    $this->settings = $settings;

    // Enable settingsForm options by default.
    if (!isset($this->settings['opt_limit'])) {
      $this->settings['opt_limit'] = TRUE;
    }
    if (!isset($this->settings['opt_single_item'])) {
      $this->settings['opt_single_item'] = TRUE;
    }
    if (!isset($this->settings['opt_cache_items'])) {
      $this->settings['opt_cache_items'] = user_access('access drunkins advanced');
    }

    // In case a subclass sets 'opt_list', we filter away arguments which
    // are useless in the 'display' URL.
    // @todo this does not work for a fetcher. Redo this when we know the future of UI/forms.
    $this->display_arguments_filter = array(
      'opt_limit',
      'opt_single_item',
      'opt_cache_items',
    );
  }

  /**
   * settingsForm(); not part of interface definition; may move to own UI class.
   *
   * It's up to the caller to collect these settings and pass them as the
   * constructor's $settings argument, after which other code can access them.
   * (drunkins module takes care of this.)
   *
   * We provide weights, which are a bit arbitrary, but at least they provide a
   * way for extending classes to place their additional widgets.
   */
  public function settingsForm() {
    $form = array();

    if (!empty($this->settings['opt_fetcher_timestamp'])) {
      if (!empty($this->settings['fetcher_timestamp_from_date'])) {
        $from = $this->settings['fetcher_timestamp_from_date'];
      }
      else {
        $from = variable_get('drunkins_fetcher_ts_' . $this->settings['job_id']);
        if ($from) {
          // Overlap is number of seconds to go back from last recorded timestamp.
          $from -= (isset($this->settings['fetcher_timestamp_overlap']) && is_numeric($this->settings['fetcher_timestamp_overlap'])) ? $this->settings['fetcher_timestamp_overlap'] : 0;
          $from = date('Y-m-d\TH:i:s', $from);
        }
      }

      $form['selection']['fetcher_timestamp_from_date'] = array(
        '#type' => 'textfield',
        '#title' => t(is_string($this->settings['opt_fetcher_timestamp']) ? $this->settings['opt_fetcher_timestamp'] : 'Fetch items changed since'),
        '#description' => 'Format: anything which PHP can parse.',
        '#default_value' => $from,
        '#weight' => 2,
        '#states' => array('enabled' => array(
          ':input[name="fetcher_timestamp_ignore"]' => array('checked' => FALSE),
          ':input[name="cache_items"]' => array('!value' => '2'),
        )),
      );

      $form['selection']['fetcher_timestamp_ignore'] = array(
        '#type' => 'checkbox',
        '#title' => t('Ignore "changed since"'),
        '#default_value' => !empty($this->settings['fetcher_timestamp_ignore']) || empty($from),
        '#weight' => 3,
        '#states' => array('enabled' => array(':input[name="cache_items"]' => array('!value' => '2'))),
      );
    }

    if ($this->settings['opt_limit']) {
      $form['selection']['fetcher_limit'] = array(
        '#type'        => 'textfield',
        '#title'       => t('Limit processing to # items'),
        '#description' => t('Only the first received items will be processed, the rest is ignored.'),
        '#weight'      => 5,
        '#default_value' => !empty($this->settings['fetcher_limit']) ? $this->settings['fetcher_limit'] : NULL,
      );
    }

    if ($this->settings['opt_single_item']) {
      $single_item_title = t('Process one specific item');
      $descn = t('Provide the !id of the item, if you want to process only that one.',
        array('!id' => t(isset($this->settings['pkey']) ? 'ID' : '(1-based) sequence number')));
      $form['selection']['fetcher_single_item'] = array(
        '#type'        => 'textfield',
        '#title'       => $single_item_title,
        '#description' => $descn,
        '#weight'      => 10,
        '#default_value' => !empty($this->settings['fetcher_single_item']) ? $this->settings['fetcher_single_item'] : NULL,
      );
    }

    if ($this->settings['opt_cache_items'] && module_exists('ctools') && isset($this->settings['job_id'])) {
      ctools_include('object-cache');
      $items = ctools_object_cache_get('drunkins_items', $this->settings['job_id']);
      if (is_array($items) && $items) {
        // Don't do this when submitting a form; memory will just be needlessly
        // wasted on caching all items into drupal_static. (The only reason we
        // fetch the first set, is that it changes our options == validation.)
        if (empty($form_state['input'])) {
          $count = count($items);
          // We need to add up items from all cached 'reruns'.
          $rerun = 1;
          while ($items = ctools_object_cache_get("drunkins_items:$rerun", $this->settings['job_id'])) {
            if (is_array($items)) {
              $count += count($items);
            }
            $rerun++;
          }
        }
        else {
          $count = '>= ' . count($items);
        }
        $form['selection']['cache_items'] = array(
          '#type' => 'radios',
          '#title' => t('Cached items'),
          '#description' => format_plural($count, 'Currently 1 fetched item is cached.', 'Currently @count fetched items are cached.') . ' ' . t('(N cached items may not result in N preprocessed items queued by the job.)'),
          '#options' => array(
            // Behavior of options is the same as their equivalents just below.
            // This is important because we don't know if another client
            // (another browser tab?) has populated/cleared the cache already.
            2 => t('Use cached items'),
            1 => t('Fetch & cache new items'),
            0 => t('Fetch new items for this process, but keep the current item(s) cached'),
            3 => t('Fetch new items; clear cache'),
          ),
          '#default_value' => isset($this->settings['cache_items']) ? $this->settings['cache_items'] : 2,
          // We're putting this on top because value 2 influences enabled-ness
          // of so many other things.
          '#weight' => 1,
        );
      }
      else {
        // If there are no cached items yet, it's not a "data selection" thing.
        $form['cache_items'] = array(
          '#type' => 'radios',
          '#title' => t('Cache fetched items locally'),
          '#description' => t('This is useful for testing, if you know your <em>next</em> run of this job will process the same items and caching them locally will speed things up.'),
          '#options' => array(
            0 => t('Do not cache'),
            1 => t('Cache items after fetching them'),
          ),
          '#weight' => 13,
        );
        $form['cache_items']['#default_value'] = isset($this->settings['cache_items']) && isset($form['cache_items']['#options'][$this->settings['cache_items']]) ? $this->settings['cache_items'] : 0;
      }
    }

    // Place fieldset around data selection options, only if there are multiple.
    if (isset($form['selection']) && count($form['selection']) > 1) {
      $form['selection'] += array(
        '#type'   => 'fieldset',
        '#title'  => t('Data selection'),
        '#collapsible' => TRUE,
      );
    }
    // We always need to set weight (to be able to place other form elements),
    // regardless whether 'selection' is an actual element.
    $form['selection']['#weight'] = 10;

    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * This implementation contains code to support all caching/filtering options
   * from the settingsForm.
   */
  public function start(array &$context) {
    // 'fetcher_timestamp' is ours. We allow a job / child class to set it but
    // protest if it is not in the expected format, even if we wouldn't use it
    // (to prevent any misunderstandings in the future).
    if (isset($context['fetcher_timestamp_from'])
        && (!is_int($context['fetcher_timestamp_from']) || $context['fetcher_timestamp_from'] <= 1262304000)) {
      throw new RuntimeException("Illegal fetcher_timestamp_from value set in context:" . drupal_json_encode($context['fetcher_timestamp_from'])  . "'");
    }

    $cached_new_items = FALSE;
    if ($items = $this->checkCache()) {
      // Unset the context value so that other code can depend on the value
      // actually meaning that a start timestamp 'is active / can be used'.
      unset($context['fetcher_timestamp_from']);
      // Note that any context value that a child class might set, only gets
      // overruled by cached items; not by settings (that say e.g. whether the
      // timestamp option is enabled or ignored). A child class setting the
      // context value must take into account that it *may* trump those settings
      // (which is ultimately decided by the implementation of the fetcher).
    }
    else {
      // Set timestamp into context, the first time start() is called / unless
      // a job/child class has set it. We don't force fetchers/jobs to use this
      // context value; 'simple' ones might just call getStartTimestamp() when
      // they need it. However, since its return value depends on an outside
      // environment, we can't guarantee that it doesn't change during job runs.
      // Fetchers/jobs that depend on the value not changing after start() is
      // called for the first time, should use the context value. (Which will
      // also open the possibility for child classes to feed this value in
      // context.)
      if (!isset($context['fetcher_timestamp_from']) && $from_timestamp = $this->getStartTimestamp()) {
        $context['fetcher_timestamp_from'] = $from_timestamp;
      }
      // Just before fetching items, set the timestamp. Our finish() can use it
      // depending on settings, but other classes might have some use for it too
      // so we always set it.
      //
      // If we fetch items in multiple 'batches', then we only set it the first
      // time. This runs the risk that in a future run of this job, items that
      // are already fetched in (2nd-Nth batches of) this job run, will be
      // fetched again in the next run. But updating this timestamp on the
      // 2nd-Nth fetch runs the risk of missing items (that were updated between
      // the 1st and nth fetch in this job run). So... unlike
      // fetcher_timestamp_from, we assume no other code is setting this.
      if (!isset($context['fetcher_timestamp'])) {
        $context['fetcher_timestamp'] = time();
      }

      $items = $this->fetchItems($context);

      $cached_new_items = $this->checkCacheSet($items, $context);
    }

    // Filter items by 'limit' and 'single_item'. We do not filter cached items
    // by 'timestamp from' because it would make processing slower, and it
    // doesn't provide much advantage (given the intended use of a 'rolling
    // timestamp') anyway.

    if (!empty($this->settings['fetcher_single_item'])) {
      // Filter items for single value.
      // For cached items, this should always be done.
      // NOTE: If no items are cached and 'pkey' is set, then hopefully
      // fetchItems() will actually perform a 'filtered query/fetch', so we do
      // not needlessly fetch and then discard items! (In that case, this call
      // should not change anything to $items.)
      $items = $this->filterItemsSingleValue(
        $items,
        empty($this->settings['pkey']) ? '' : $this->settings['pkey'],
        $this->settings['fetcher_single_item']
      );
    }

    if (isset($this->settings['fetcher_limit']) && $this->settings['fetcher_limit'] > 0) {
      /* Note the strangeness here: the 'items previously fetched', which we use
       * to limit the total items fetched, is not a counter _we_ keep, but one
       * the DrunkinsRunner keeps. It's not a disaster because we're only
       * interested in this counter if start() is re-run (which is done by the
       * runner) anyway... but the count can differ from the actual items
       * fetched, if a child class' start() preprocesses items before passing
       * them back to the runner. This could cause confusion.
       */
      $prev_count = !empty($context['drunkins_rerun_start_count']) ? $context['drunkins_rerun_start_count'] : 0;
      // Return only a limited number of items - taking into account the number
      // we already set before.
      if (!is_numeric($prev_count) || $prev_count >= $this->settings['fetcher_limit']) {
        throw new Exception("Internal code error? We have a 'fetcher_limit' setting of " . $this->settings['fetcher_limit'] . " but have already returned more items than the limit ($prev_count)?");
      }
      if ($prev_count + count($items) > $this->settings['fetcher_limit']) {
        $items = array_slice($items, 0, $this->settings['fetcher_limit'] - $prev_count);
        // If a child fetchItems() set 'drunkins_process_more', we unset it;
        // we don't want to fetch more items.
        if (!empty($context['drunkins_process_more'])) {
          if ($cached_new_items) {
            // There is no rule saying the cache must be complete, so we are
            // allowed to do this. However, the cache will now be incomplete and
            // the next time we run the job using these cached items, there will
            // be nothing to signify that fact. We only warn here. (We could
            // make a setting to complete the cache anyway, if ever needed; we'd
            // need to keep fetching and caching items while returning 0, and
            // tweak the above exception logic a bit.)
            $this->log("Because of the 'fetcher_limit' setting, an incomplete data set was cached!", array(), WATCHDOG_WARNING);
          }
          unset($context['drunkins_process_more']);
        }
      }
    }

    return $items;
  }

  /**
   * {@inheritdoc}
   */
  public function finish(array &$context) {
    // Upon successful finish, store the timestamp, only if this functionality
    // is enabled. (Note - also if the timestamp was ignored in this job run.
    // But not if the 'fetcher_timestamp' context value was set by external
    // code without the 'enabled' setting being true.)
    if (!empty($this->settings['opt_fetcher_timestamp']) && $this->hasFullDataSet($context, TRUE)) {
      if (empty($context['fetcher_timestamp'])) {
        // Something emptied the timestamp that was set earlier. This may be
        // legit (for instance, if cached items are used), but let's not finish
        // without logging a notice.
        $this->log('Fetcher timestamp is not set, so next time this job is run, we will not automatically roll through to the next time period.', array());
      }
      else {
        $this->setNextStartTimestamp($context['fetcher_timestamp']);
      }
    }
  }

  /// Extra helper methods, for extending classes to call.
  /// Many methods need to be public because some odd situation requires them to
  /// be called from the encapsulating Job class.

  /**
   * Works out whether we are processing a complete data set in this job.
   *
   * Especially finish() tends to contain functionality which must first check
   * this.
   *
   * 'Completeness' is measured over the data set as a whole, which may be
   * fetched during various calls to start().
   *
   * @param array $context
   *   The process context array.
   * @param bool $ignore_timestamp
   *   (optional) modifier; see return value.
   *
   * @return bool
   *   TRUE if we know we are processing a complete data set - or if
   *   $ignore_timestamp is TRUE: if we would be processing a complete data
   *   set if 'fetcher_timestamp_ignore' was set.
   */
  public function hasFullDataSet(array &$context, $ignore_timestamp = FALSE) {
    $full = empty($this->settings['fetcher_limit'])
            && empty($this->settings['fetcher_single_item'])
            && !$this->usesCachedItems();
    if ($full && !$ignore_timestamp) {
      // IF this job is set up to use 'rolling timestamps' (equivalent to:
      // 'opt_fetcher_timestamp' is nonempty, OR (depending on fetcher) the
      // 'fetcher_timestamp_from' context value is nonempty), THEN we have a
      // full data set if no timestamp is set or we are ignoring it (equivalent
      // to: getStartTimestamp() returns 0 and the context value is empty).
      // IF this job can't use 'rolling timestamps', then we assume we have a
      // full data set. All this can be rolled into a short equation:
      try {
        $full = empty($context['fetcher_timestamp_from']) && $this->getStartTimestamp() == 0;
      }
      catch (RuntimeException $e) {
        // This means no timestamp was set. That still means the data set is not
        // restricted (even though it won't be able to run). So return 'full'.
      }
    }

    return $full;
  }

  /**
   * Fetch items; in separate method so it can be overridden without
   * needing to touch the existing code in start().
   *
   * @param array $context
   *   The process context, passed by reference for (further) initialization.
   *
   * @return array
   *   The items.
   *
   * @throws \RuntimeException
   *   For any error that is supposed to prevent the process from executing.
   */
  protected function fetchItems(array &$context) {
    return array();
  }

  /**
   * Test whether the current job run is using / will use cached items.
   *
   * @return bool
   *   TRUE if cached items will be used. (Also if they are unavailable, in
   *   which case checkCache() will throw an exception.)
   */
  public function usesCachedItems() {
    return isset($this->settings['cache_items']) && $this->settings['cache_items'] == 2;
  }

  /**
   * Gets cached items, IF a setting tells us this must be done.
   *
   * This method assumes that caching is fast, so items which were fetched and
   * cached in multiple runs earlier, are now all returned at once. We'll need
   * to rewrite this to use proper 'drunkins_process_more' if the number of
   * items ever grows so large we run out of memory.
   *
   * @return array|null
   *   The cached items, if applicable. Otherwise NULL.
   *
   * @throws \RuntimeException
   *   If the job is set to use cached items but they can't be retrieved.
   */
  public function checkCache() {
    $items = NULL;
    if ($this->usesCachedItems()) {
      if (isset($this->settings['job_id']) && module_exists('ctools')) {
        ctools_include('object-cache');
        $items = ctools_object_cache_get('drunkins_items', $this->settings['job_id']);
        if (isset($items) && is_array($items)) {
          $rerun = 1;
          while ($more_items = ctools_object_cache_get("drunkins_items:$rerun", $this->settings['job_id'])) {
            if (is_array($more_items)) {
              // Values must be appended, not overwrite existing ones. The
              // caller does not use array_keys anyway. So, just in case our
              // array has non-numeric keys:
              $items = array_merge($items, array_values($more_items));
              $rerun++;
            }
            else {
              throw new RuntimeException("Value cached as drunkins_items:$rerun:{$this->settings['job_id']} is not an array: " . json_encode($more_items));
            }
          }
        }
        elseif (isset($items)) {
          throw new RuntimeException("Value cached as drunkins_items:{$this->settings['job_id']} is not an array: " . json_encode($items));
        }
        else {
          throw new RuntimeException("Job {$this->settings['job_id']} specifies it uses cached items, but no items are present.");
        }
      }
      else {
        throw new RuntimeException("No mechanism for getting cached items for job '{$this->settings['job_id']}''.");
      }
    }
    // If an empty array is cached, it's also returned. It's up to the caller
    // what to do with that (senseless) situation.
    // (We never cache empty arrays, but a child class might....)
    return $items;
  }

  /**
   * Caches items (or clears cache), IF a setting tells us this must be done.
   *
   * @param array|string $items
   *  Items to possibly cache.
   * @param array $context
   *   The process context.
   *
   * @return bool
   *   TRUE if items were cached.
   */
  public function checkCacheSet($items, array $context) {
    $cached = FALSE;

    // Cache is only cleared if items were fetched.
    // We do not cache an array with 0 items.
    if (is_array($items) && $items && isset($this->settings['job_id'])
        && isset($this->settings['cache_items']) && module_exists('ctools')
        && in_array($this->settings['cache_items'], array(1, 3))) { // @todo rework this
      ctools_include('object-cache');

      $rerun = (isset($context['drunkins_rerun_start'])
                && is_numeric($context['drunkins_rerun_start']))
        ? $context['drunkins_rerun_start'] : 0;
      $rerun_id = $rerun ? ":$rerun" : '';

      if ($this->settings['cache_items'] == 1) {
        ctools_object_cache_set("drunkins_items$rerun_id", $this->settings['job_id'], $items);
        $cached = TRUE;
      }
      else {
        ctools_object_cache_clear("drunkins_items$rerun_id", $this->settings['job_id']);
      }

      // We're setting or clearing the cache - either way, the caches set
      // at previous 'rerun's are now invalid. Clear all the ones we find.
      // (This is done on all calls, to prevent the unlikely case of 'gaps' in
      // sequence numbers, which could cause an older set of items to be merged
      // into a new one, on cache get. Usually, the below will only do anything
      // when $rerun is 0.)
      $rerun++;
      while (ctools_object_cache_get("drunkins_items:$rerun", $this->settings['job_id'])) {
        ctools_object_cache_clear("drunkins_items:$rerun", $this->settings['job_id']);
        $rerun++;
      }
    }

    return $cached;
  }

  /**
   * Returns the 'rolling timestamp' start, if the functionality is enabled.
   *
   * This is a separate method mostly so the way of storing such a timestamp
   * over various runs can be abstracted. However if your job needs timestamp
   * functionality, you still may not want to call this directly because this
   * method may not imply what you need it to:
   *
   * For one: if your job / fetcher depends on this value not changing during a
   * run, it may be better off not calling this method but checking the
   * 'fetcher_timestamp_from' context value, because that is not dependent on
   * whatever 'environment' this method may read from. See comments in start().
   *
   * Further: checking if "rolling timestamp functionality is enabled / used"
   * turns out to sometimes be less straightforward than it seems, in part
   * because you need to be very aware of what you mean by that.
   *
   * * Though fetcher code is often better off checking the context value
   *   'fetcher_timestamp_from' than calling this method, the context value does
   *   not necessarily mean that "rolling timestamp functionality is enabled".
   *   It might also be set by a custom job aiming for a fetcher to apply a
   *   one-off date filter, without updating the timestamp afterwards. If a
   *   fetcher wants to explicitly not allow job code to do this, checking just
   *   the context value is not enough.
   *
   * * The most regular definition of 'enabled-ness' is the combination of the
   *   'opt_fetcher_timestamp' & 'fetcher_timestamp_ignore' settings. Which is
   *   almost equivalent to this method returning a non-zero value, except this
   *   method can throw an exception (if 'enabled' but no date can be derived),
   *   which may or may not be what you want.
   *
   * * Checking the above settings / calling this method will yield the same
   *   result, regardless whether the job is using cached items. So take care to
   *   only do this in places where you know the process is not using cached
   *   items OR your code doesn't care about it (meaning it has a slightly
   *   different 'definition of enabled-ness'). Or include the 'cached' check
   *   yourself.
   *
   * * The above definition of 'enabled-ness' is not equal to "a situation when
   *   the timestamp will be set for the next run" (which arguably could also be
   *   seen as 'enabled-ness of rolling timestamp functionality')... because:
   *   - the timestamp is also set when 'fetcher_timestamp_ignore' is true.
   *   - above definition does not take into account completeness of the data
   *     set (e.g. filters being applied) but the timestamp is only set when
   *     (disregarding any timestamp filter) a full data set was processed .
   *
   * There is no helper function to capture all of this. Just be aware what
   * exactly you want to be checking, when writing jobs that use 'rolling
   * timestamp' functionality. Feel free to write a small helper method for your
   * own job, after deciding whether you want to include pre-set context values,
   * whether you want to check whether the functionality 'works' or always
   * return a timestamp (which is mutually exclusive), whether the code in
   * start() has always run already (so context is filled/exceptions are already
   * thrown when appropriate), whether you want to exclude the 'ignored' setting
   * ...and commenting your intentions so you don't get confused when reading
   * back your code later.
   *
   * @return int
   *   The start time to use for fetching items, or 0 if the current process is
   *   not / should not be using a 'rolling timestamp'.
   *
   * @throws \RuntimeException
   *   If 'input' settings are inconsistent. Note this also happens when a job
   *   that needs a rolling timestamp has never successfully run yet and has no
   *   start date (or 'ignore') set explicitly.
   */
  public function getStartTimestamp() {
    $from = 0;
    if (!empty($this->settings['opt_fetcher_timestamp']) && empty($this->settings['fetcher_timestamp_ignore']) && !$this->usesCachedItems()) {
      if (isset($this->settings['fetcher_timestamp_from_date'])) {
        // This setting (which probably came from settingsForm input) overrides
        // the timestamp.
        $from = strtotime($this->settings['fetcher_timestamp_from_date']);
        if (!$from) {
          throw new RuntimeException("Invalid 'fetch items changed since' date specified; this process will not run.");
        }
      }
      elseif ($last_sync = variable_get('drunkins_fetcher_ts_' . $this->settings['job_id'], 0)) {
        // Overlap is number of seconds to go back from last recorded timestamp.
        $overlap = (isset($this->settings['fetcher_timestamp_overlap']) && is_numeric($this->settings['fetcher_timestamp_overlap']))
          ? $this->settings['fetcher_timestamp_overlap'] : 0;
        $from = $last_sync - $overlap;
      }
      else {
        throw new RuntimeException("'fetcher_timestamp_from_date' must be set on this job. (Probably through a user interface.)");
      }
    }

    return $from;
  }

  /**
   * Sets the timestamp to use for the next job run.
   *
   * @param int $timestamp
   *   The timestamp. Must be an int, not a numeric string.
   */
  public function setNextStartTimestamp($timestamp) {
    // Be picky: timestamp must be an int.
    if (!is_int($timestamp) || $timestamp <= 1262304000) {
      throw new RuntimeException("Illegal fetcher timestamp, not stored; the next run may reprocess the same items! -- '" . drupal_json_encode($timestamp)  . "'");
    }
    variable_set('drunkins_fetcher_ts_' . $this->settings['job_id'], $timestamp);
  }

  /**
   * Helper function, filters $items for a certain value
   *
   * @param $items
   *   Items
   * @param string $pkey
   *   Primary key inside item to check for filter value. Must be nonempty if
   *   $filter_value is non-numeric.
   * @param string|int $filter_value
   *   Value to filter primary key on, or (if $pkey is empty) numeric 1-based
   *   index of item to return.
   *
   * @return array
   *   Filtered items.
   *
   * @throws \RuntimeException
   *   If filtering failed because of wrong configuration (rather than 'not
   *   found').
   */
  protected function filterItemsSingleValue($items, $pkey, $filter_value) {

    if ($items) {
      if (!$pkey && !is_numeric($filter_value)) {
        throw new RuntimeException("filterItems() cannot filter 'fetcher_single_item' value \"$filter_value\" because no primary key is set.");
      }
      elseif ($pkey) {

        // Filter items. There could theoretically be multiple items with the
        // same primary key value, so loop through all items.
        $new_items = array();
        foreach ($items as $item) {
          if (!empty($item[$pkey]) && $item[$pkey] == $filter_value) {
            $new_items[] = $item;
          }
        }
        if (!$new_items) {
          $this->log("filterItems() yielded 0 items for 'fetcher_single_item' value %value.", array('%value' => $filter_value));
        }
        $items = $new_items;
      }
      elseif ($filter_value > 0 && $filter_value <= count($items)) {

        // Numeric 'fetcher_single_item': take Nth item. (1-based)
        $items = array_slice($items, $filter_value - 1, 1);
      }
      else {
        // Numeric 'fetcher_single_item' value is out of bounds
        $this->log("filterItems() yielded 0 items for 'fetcher_single_item' value %value.", array('%value' => $filter_value));
      }
    }

    return $items;
  }

  /**
   * Logs or prints a message. Needed for at least filterItemsSingleValue().
   */
  public function log($message, array $variables = array(), $severity = WATCHDOG_NOTICE, $repeat = TRUE) {
    if (isset($this->jobInstance) && is_callable($this->jobInstance, 'log')) {
      return $this->jobInstance->log($message, $variables, $severity, $repeat);
    }

    // Copy of DrunkinsJob::log(). I wish we didn't have to do this but it's
    // safer than dropping messages on the floor.
    // @todo the ideal situation would be for log() not to exist as 'unofficial'
    //   function in here, but inside a DrunkinsRunner class that also
    //   contains most of the .module / .run.inc code.
    if (isset($this->settings['logging']) && $this->settings['logging'] & DrunkinsJob::LOG_SCREEN) {

      $type = 'status';
      // NOTICE maps to 'status'; it would not stand out if we printed a lot of
      // status messages, but this function is expected to be used mostly by
      // warnings/errors - so this way a NOTICE will look different from a
      // WARNING on screen.
      switch ($severity) {
        case WATCHDOG_WARNING:
          $type = 'warning';
          break;
        case WATCHDOG_ERROR:
        case WATCHDOG_CRITICAL:
        case WATCHDOG_ALERT:
        case WATCHDOG_EMERGENCY:
          $type = 'error';
      }
      drupal_set_message(t($message, $variables), $type, $repeat);
    }

    if (!isset($this->settings['logging'])
        || $this->settings['logging'] & DrunkinsJob::LOG_WATCHDOG
        || $this->settings['logging'] & DrunkinsJob::LOG_DRUSH) {

      // Use $_SESSION for $repeat in the same way as drupal_set_message() does.
      $job_id = isset($this->settings['job_id']) ?
        $this->settings['job_id'] : '';

      if ($repeat
          || !isset($_SESSION['messages_drunkins_' . $job_id][$severity])
          || !in_array($message, $_SESSION['messages_drunkins_' . $job_id][$severity])) {

        if (!isset($_SESSION['messages_drunkins_' . $job_id][$severity])) {
          $_SESSION['messages_drunkins_' . $job_id][$severity] = array();
        }
        $_SESSION['messages_drunkins_' . $job_id][$severity][] = $message;

        if (!isset($this->settings['logging']) || $this->settings['logging'] & DrunkinsJob::LOG_WATCHDOG) {
          watchdog('DrunkinsJob', $message . ' (Job ID: @job_id)',
            $variables + array('@job_id' => $job_id), $severity);
        }
        if (isset($this->settings['logging']) && $this->settings['logging'] & DrunkinsJob::LOG_DRUSH) {
          if (DRUSH_MAJOR_VERSION < 7) {
            // _drush_log() ignores 'info' (which is our default log severity) and
            // 'notice' if not in verbose mode - and then they are printed among
            // lots of other Drush logs. Alternatively, we can give severity
            // 'status' and then all messages get a green "[status]" indicator.
            // (Which can again be suppressed by 'quiet'.) Not sure if all logs
            // should be 'green' but we do want the "X items queued" message.
            $type = 'info';
            switch ($severity) {
              case WATCHDOG_INFO:
              case WATCHDOG_NOTICE:
                $type = 'status';
                break;
              case WATCHDOG_WARNING:
                // See just below.
                if ($this->settings['logging'] & DrunkinsJob::LOG_WATCHDOG) {
                  return;
                }
                $type = 'warning';
                break;
              case WATCHDOG_ERROR:
              case WATCHDOG_CRITICAL:
              case WATCHDOG_ALERT:
              case WATCHDOG_EMERGENCY:
                // Watchdog already prints to the screen when Drush executes;
                // don't output duplicate logs.
                // @todo check what exact drush context this depends on.
                // @todo check drush >= 7.
                if ($this->settings['logging'] & DrunkinsJob::LOG_WATCHDOG) {
                  // This only works if DRUSH_LOG is the last if() block...
                  return;
                }
                $type = 'error';
            }
            $severity = $type;
          }
          drush_log(t($message, $variables), $severity);
        }
      }
    }
  }

}
