<?php
/**
 * @file
 * Contains DrunkinsCommerceProductImportJob.
 */

//@todo use an Exception type with args/tokens/translation context, in various places.

/**
 * Class DrunkinsEntityImportJob
 *
 * A class suitable for importing fetched items into Drupal entities.
 *
 * Most code is general. There are some field/entity specific hacks in the code
 * to make import smoother:
 * - field: addressfield in buildEntityValue(), to prevent unneeded saves when
 *   field values have not really changed.
 * - entity: commerce_customer_profile in saveEntity(), to prevent saving a
 *   profile that would be duplicated on save.
 */
class DrunkinsEntityImportJob extends DrunkinsSqlTableImportJob {

  /* Settings in $this->settings, used by this class:
   * - job_id:              defined in parent
   * - pkey:                defined in fetcher
   * - fetcher_limit:       defined in fetcher
   * - fetcher_single_item: defined in fetcher
   *
   * - import_entity_type (string):
   *   Type of Drupal entity to import data into.
   * - import_new_bundle (string):
   *   Bundle used to create new entities during import. Required unless the
   *   entity type does not have a 'bundle' definition in its 'entity keys'.
   * - import_resolve_all_bundles (boolean, default = FALSE):
   *   If TRUE, don't restrict resolveEntityIds() to querying inside
   *   'import_new_bundle' (which is the bundle where supposedly all imported
   *   entities were created inside of).
   * - import_entity_query_filter_field: defined in parent.
   * - import_entity_query_filter_value: defined in parent.
   *   At this moment, the filter_field must be an 'entity property', i.e.
   *   available directly in the entity table.
   *
   * - import_field_mapping (array):     defined in parent.
   * - import_default_values (array):    defined in parent.
   *
   * - import_dst_pkey (string):         defined in parent.
   * - import_src_changed (string):
   * - import_dst_changed (string):
   *   The source item / entity field containing the last-changed date of a
   *   source item. If your fetched data doesn't only contain items which are
   *   changed since the last import process ran, then saving the last-changed
   *   date in the entity and setting either of above properties will prevent
   *   needless saving of entities for unchanged items.
   *   Like the pkey settings, setting only one of 'import_src_changed' and
   *   'import_dst_changed' is allowed, if the other can be derived through the
   *   field mapping.
   *
   * - import_skip_unchanged (integer, default = 1):
   *   If TRUE/1, skip writing items whose 'changed' value is older than the
   *   field value stored in the entity. This only works if 'import_src_changed'
   *   / 'import_dst_changed' is set and the field is present in the entity.
   *   (If 2, only skip writing if the fields are exactly equal. This almost
   *   never makes a difference from 1; see code comments.)
   *
   * - opt_cleanup (string/array): defined in parent, except there is a default
   *   for value TRUE: that gets expanded to array('status' => 0).
   * - import_cleanup (boolean):   defined in parent
   * - import_cleanup_all_bundles (boolean, default = FALSE):
   *   If TRUE, don't restrict resolveEntityIds() to querying inside
   *   'import_new_bundle' (which is the bundle where supposedly all imported
   *   entities were created inside of).
   *
   * - import_trim_values (boolean, TRUE by default): defined in parent.
   * - import_entity_default_owner_id (integer):
   *   User ID that should be used for creating new entities, if user is
   *   'anonymous' -- which is mostly the case on cron runs.
   * - import_entity_always_use_default_owner_id (boolean):
   *   If TRUE, always use import_entity_default_owner_id for new entities, also
   *   for processes that are run by a non-anonymous user.
   *
   * - import_ids_resolve_fast (boolean, TRUE by default):
   *   Makes resolveEntityIds() use faster queries; works only on SQL back-ends.
   * - import_ids_case_sensitive (boolean): defined in parent.
   */

  // Class-wide variables which are filled once per processItem() execution
  // and used by its various called functions to make the code easier.
  //
  // Default field language codes for the current entity (array of fieldname =>
  // langcode pairs). Set at the start of updateEntity() / initEntity()
  protected $field_langcodes;
  // The bundle of the current entity. Set in the same places.
  protected $current_entity_bundle;

  // Class-wide variables used as 'cache' for other purposes.
  //
  // Destination field; often the same as $current_dst, but that may include a
  // language suffix, in which case $current_dst_field has it stripped.
  protected $current_dst_field;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $settings = array()) {

    if (!isset($settings['import_ids_resolve_fast'])) {
      $settings['import_ids_resolve_fast'] = TRUE;
    }
    if (empty($settings['import_ids_resolve_fast'])) {
      // resolveEntityIds() cannot be called with option 'requery_all' so make
      // sure it doesn't: disallow the precaching setting.
      $this->settings['import_precache_entity_ids'] = FALSE;
    }

    if (!isset($settings['import_skip_unchanged'])) {
      $settings['import_skip_unchanged'] = 1;
    }

    parent::__construct($settings);

    // In case a subclass sets 'opt_list', we filter away arguments which
    // are useless in the 'display' URL.
    $this->display_arguments_filter = array_merge($this->display_arguments_filter, array(
      'import_new_bundle',
      'import_resolve_all_bundles',
      'import_cleanup_all_bundles',
      'import_src_changed',
      'import_dst_changed',
      'import_entity_default_owner_id',
      'import_entity_always_use_default_owner_id',
      'import_ids_resolve_fast',
      'import_skip_unchanged',
      'opt_cleanup',
      'import_cleanup',
    ));
  }

  /// Extensions of non-interface methods from parent classes

  /**
   * {@inheritdoc}
   */
  protected function checkSettings(array $context, $quiet_fast = FALSE) {
    $ok = TRUE;
    foreach (array('entity_type', 'field_mapping') as $setting) {
      if (empty($this->settings["import_$setting"])) {
        if ($ok = $quiet_fast) {
          return FALSE;
        }
        $this->log('@setting setting not configured!', array('@setting' => "import_$setting"), WATCHDOG_ERROR);
      }
    }
    if (!empty($this->settings['import_entity_type'])) {
      // Check if bundle is required too.
      $info = entity_get_info($this->settings['import_entity_type']);
      if (!empty($info['entity keys']['bundle']) && empty($this->settings['import_new_bundle'])) {
        if ($ok = $quiet_fast) {
          return FALSE;
        }
        $this->log('@setting setting not configured!', array('@setting' => 'import_new_bundle'), WATCHDOG_ERROR);
      }
      // Bundle settings must be empty (except in above case) or among the
      // allowed bundles. This also goes for bundle-less entities, which always
      // have one default bundle value set.
      if (!empty($this->settings['import_new_bundle'])
          && !isset($info['bundles'][ $this->settings['import_new_bundle'] ])) {
        if ($ok = $quiet_fast) {
          return FALSE;
        }
        $this->log("bundle value @value from '@setting' is invalid for a '@type' entity!", array(
          '@value' => $this->settings['import_new_bundle'],
          '@setting' => 'import_new_bundle',
          '@type' => $this->settings['import_entity_type'],
        ), WATCHDOG_ERROR);
      }
    }

    if (isset($this->settings['import_default_values']) && !is_array($this->settings['import_default_values'])) {
      if ($ok = $quiet_fast) {
        return FALSE;
      }
      $this->log('@setting setting is not an array!', array('@setting' => "import_default_values"), WATCHDOG_ERROR);
    }

    if (!empty($this->settings['import_entity_query_filter_field'])
        xor !empty($this->settings['import_entity_query_filter_value'])) {
      if ($ok = $quiet_fast) {
        return FALSE;
      }
      $this->log("If one of '@setting1' and '@setting2' settings are set, they must both be set.",
        array('@setting1' => 'import_entity_query_filter_field', '@setting2' => 'import_entity_query_filter_value'), WATCHDOG_ERROR);
    }

    // These should all be present after setMappingByDest has run.
    foreach (array('src_pkey', 'dst_pkey') as $setting) {
      if (empty($context["import_$setting"])) {
        if ($ok = $quiet_fast) {
          return FALSE;
        }
        $this->log('@setting context value not found / cannot be derived from other settings!', array('@setting' => "import_$setting"), WATCHDOG_ERROR);
      }
    }

    if (!is_array($this->settings['import_field_mapping'])) {
      if ($ok = $quiet_fast) {
        return FALSE;
      }
      $this->log('@setting setting is not an array!', array('@setting' => 'import_field_mapping'), WATCHDOG_ERROR);
    }
    // We don't do the following block when $quiet_fast - it arguably takes too
    // long to do it on every processItem(), and won't break anything horribly.
    elseif (!$quiet_fast && !empty($this->settings['import_entity_type'])) {
      $instance_info = field_info_instances($this->settings['import_entity_type']);
      foreach (array_keys($this->settings['import_field_mapping']) as $dst_key) {
        if (strpos($dst_key, ':')) {
          $parts = explode(':', $dst_key);
          $lc = array_pop($parts);
          // Field may not have a ':' and we only check FieldAPI fields here:
          if (count($parts) == 1
              && isset($instance_info[$parts[0]])
              && ($field_info = field_info_field($parts[0]))
              && !in_array($lc, field_available_languages($this->settings['import_entity_type'], $field_info), TRUE)
          ) {
            if ($ok = $quiet_fast) {
              return FALSE;
            }
            $this->log('@key is mentioned as a destination in the field mapping, but @lc is not an available language for the @field field!',
                array('@key' => $dst_key, '@lc' => $lc, '@field' => $parts[0]), WATCHDOG_ERROR);
          }
        }
      }
    }

    return $ok;
  }

  /// Interface methods

  /**
   * {@inheritdoc}
   */
  public function settingsForm() {
    $form = parent::settingsForm();

    if (!empty($this->context['import_src_changed'])) {
      // It's unfortunate that we have to put both options 1 and 2 into the form
      // because they almost never differ -- but if we support 2 different
      // values in the code, we can't make one of them totally inaccessible.
      // Extending classes may make this a checkbox (as we used to have here)
      // if they want.
      $form['behavior']['import_skip_unchanged'] = array(
        '#type'          => 'select',
        '#title'         => t('Skip unchanged source items'),
        '#options'       => array(
          0 => t('No (always overwrite)'),
          1 => t('Yes, overwrite if source item is newer'),
          2 => t('Yes, overwrite if local entity changed'),
        ),
        '#description'   => t("The 'Yes' options save time. 'changed' is derived from the 'changed date' of the source item, which is mirrored in Drupal. This the two 'Yes' options behave in the same way, unless your Drupal code sometimes overwrites the same 'changed date' field."),
        '#default_value' => 1,
        '#weight'        => 1,
      );

      if (!empty($form['selection']['fetcher_single_item'])) {
        $form['selection']['fetcher_single_item']['#description'] .=
          ' ' . t('You will probably want to uncheck \'!import_skip_unchanged\' too.',
            array('!import_skip_unchanged' => t('Skip unchanged source items')));
      }
    }

    // Place fieldset around behavior options, only if there are multiple.
    if (isset($form['behavior']) && count($form['behavior']) > 1) {
      $form['behavior'] += array(
        '#type'   => 'fieldset',
        '#title'  => t('Job behavior'),
        '#weight' => 6,
      );
    }

    return $form;
  }

  /// New methods introduced by this/parent class.

  /**
   * {@inheritdoc}
   */
  protected function processEntities(array $actions, array $entity_ids = array()) {

    if ($actions == array('cleanup')) {
      // "cleanup" will be expanded to the 'opt_cleanup' setting value.
      if (empty($this->settings['opt_cleanup'])) {
        // The parent would not be called if !$entity_ids and we don't want to
        // waste time preprocessing, so copy the log message here:
        $this->log("Empty '@option' setting value; ignoring cleanup.", array('@option' => 'opt_cleanup'), WATCHDOG_WARNING, FALSE);
        return;
      }

      // TRUE is not a proper value, but for our class it has a known default:
      // 'unpublish/deactivate'. May do nothing or give error if the entity
      // has no 'status' property.
      if ($this->settings['opt_cleanup'] === TRUE) {
        $this->settings['opt_cleanup'] = array('status' => 0);
      }

      if (!$entity_ids) {
        // Pre-determine the list of entities to clean up.

        $ids = $this->processedEntityIds();
        if (!$ids) {
          $this->log("Cleanup action could not determine processed items; skipping.", array(), WATCHDOG_WARNING);
          return;
        }
        $entity_type = $this->settings['import_entity_type'];

        $query = new EntityFieldQuery();
        $query->entityCondition('entity_type', $entity_type);
        if (!empty($this->settings['import_cleanup_all_bundles'])) {
          // Add a condition on the bundle where supposedly all imported
          // entities were created inside of.
          $query->entityCondition('bundle', $this->settings['import_new_bundle']);
        }
        $result = $query
          ->entityCondition('entity_id', $ids, 'NOT IN')
          ->propertyCondition('status', 1)
          ->execute();

        if (isset($result[$entity_type])) {
          $entity_ids = array_keys($result[$entity_type]);
        }
      }
    }

    if ($entity_ids) {
      // @todo here & in parent:, instead of processing things here, requeue them.
      //    Suggestion: a 'maximum batch size' setting to group '*action' items
      //    because that would probably make processing faster. <-- there is import_cleanup_max_chunk_size now in the parent, we could reuse this.
      //    (possibly different for batch & queue & drush execution.)
      //    (possibly start() could do re-batching of equal *actions without src-id)
      parent::processEntities($actions, $entity_ids);
      // @todo decide on a 'logging strategy'. At the moment, we see no logs
      //       from 'cleanup-on-finish', but 'action-on-processItem' accounts for us which gets logged on finish.
      //       See also entity_import_job processEntities() comment.
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function processEntitiesAction($action, array $entity_ids = array()) {

    // We don't perform actions on an empty list of entities. Default list
    // should be passed by processEntities().
    if ($entity_ids) {

      if (is_array($action)) {
        // Update values.
        //@todo load/un-cache these in batches of a maximum size? In that case, merge this with collect_ids so we don't load twice.
        foreach ($this->loadEntities($entity_ids) as $id => $entity) {
          $changed = FALSE;
          foreach ($action as $dst => $val) {
            if ($this->setEntityValue($val, $entity, $dst)) {
              $changed = TRUE;
            }
          }
          if ($changed) {
            $this->saveEntity($entity);
          }
        }
      }
      else {
        $entity_type = $this->settings['import_entity_type'];
        switch ($action) {
          // The first action in the switch statement is relatively harmless,
          // which is good if $action ever contains TRUE...
          case 'collect_ids':
            // Get the values of the 'entities' that we'll clean up: the
            // original source-item IDs keyed by their Drupal 'entity' ID.

            $dest_field = isset($this->context['import_dst_pkey']) ? $this->context['import_dst_pkey']
              : (isset($this->settings['import_dst_pkey']) ? $this->settings['import_dst_pkey']
                : '');
            if ($dest_field) {
              $info = field_info_field($dest_field);
            }
            if (!isset($this->context['import_processed']) || !is_array($this->context['import_processed'])) {
              $this->context['import_processed'] = array();
            }
            foreach ($this->loadEntities($entity_ids) as $id => $entity) {
              if ($dest_field && $info) {
                // This is the only place in this class where we're using entity
                // wrappers. The value was _set_ using more generic code, and we
                // assume we can _read_ that same value by calling value().
                $wrapper = entity_metadata_wrapper($entity_type, $entity);
                if (isset($wrapper->$dest_field)) {
                  $this->context['import_processed'][$id] = $wrapper->$dest_field->value();
                }
                else {
                  $this->context['import_processed'][$id] = TRUE;
                }
              }
              elseif ($dest_field && isset($entity->$dest_field)) {
                $this->context['import_processed'][$id] = $entity->$dest_field;
              }
              else {
                $this->context['import_processed'][$id] = TRUE;
              }
            }
            break;

          case 'delete':
            // Delete 'entities'.
            if ($entity_type === 'user') {
              // node_user_delete will delete all nodes belonging to the user.
              // We can't let that happen by accident. If you want to delete
              // users, extend this method in your own subclass. Also take care
              // not to delete any important users by accident.
              $this->log("Deleting users is prohibited.", array(), WATCHDOG_ERROR);
            }
            else {
              entity_delete_multiple($entity_type, $entity_ids);
            }
            break;

          default:
            $this->log("Unknown @option value '@value'.",
              array('@option' => 'opt_cleanup', '@value' => $action),
              WATCHDOG_WARNING);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   *
   * An extra option is supported:
   * bundle => (optional) bundle that the resolved entity must have.
   *   If $entity_type is empty, then emptiness of this argument does not mean
   *   there is no 'bundle restriction': the 'import_new_bundle' setting is
   *   checked in that case, unless 'import_resolve_all_bundles' is set to
   *   prevent that.
   *   (So, to search all bundles, the $entity_type argument must be nonempty.)
   *
   * Note on existing option behavior:
   * requery_all only works for one specific 'cache bundle'. If the
   *   $entity_type argument is explicitly specified but the 'bundle' option is
   *   not, all bundles of this type are queried, but they will end up in a
   *   place in the cache that will not be seen by later bundle-specific calls,
   *   so specifying 'requery_all' without 'bundle' is likely useless.
   */
  protected function resolveEntityIds(array $src_ids, $entity_type = '', array $options = array()) {
    // A note: the 'negative caching' could be an option (setting) - we usually
    // don't care because our standard classes populate the cache on
    // saveEntity(), so any effect of negative caching is canceled out in most
    // cases. However we don't need to requery entities usually, so we
    // haven't cared until now.

    // Before determining actions to take, set bundle and cache bin.

    if (empty($options['bundle'])) {
      if ($entity_type || !empty($this->settings['import_resolve_all_bundles'])) {
        // No bundle restriction. Set this explicitly so we don't need to check
        // existence later.
        $options['bundle'] = '';
      }
      else {
        // Implied bundle restriction for standard entity type.
        // If the entity type does not have a bundle, it still has one default
        // defined in 'bundles', and we use that string for the 'cache bundle'.
        // This way, future queries with and without that bundle will match.
        $info = entity_get_info($this->settings['import_entity_type']);
        if (count($info['bundles']) == 1) {
          reset($info['bundles']);
          $options['bundle'] = key($info['bundles']);
        }
        else {
          // Use 'import_new_bundle' by default: all imported entities (which
          // were all 'new' entities once) are created in one bundle, and other
          // bundles are ignored, so it won't matter if entities in another
          // bundle happen to have a matching value.
          $options['bundle'] = $this->settings['import_new_bundle'];
        }
      }
    }

    if (isset($this->entityIds)) {
      if (!$entity_type) {
        $entity_type = $this->settings['import_entity_type'];
      }
      if (empty($options['cache_bin'])) {
        $options['cache_bin'] = "$entity_type:$options[bundle]";
      }
    }

    if ($options['bundle']) {
      // Search only one cache bin: the parent code can handle that (and will
      // use the query in this class if needed; bundle option is set for that.)
      return parent::resolveEntityIds($src_ids, $entity_type, $options);
    }

    // Now that we know we'll execute our own code: first do checks.

    if (empty($src_ids) && empty($options['requery_all'])) {
      $this->log('resolveEntityIds called with empty parameter?', array(),
        WATCHDOG_WARNING, FALSE);
      return array();
    }
    if (isset($options['return_type']) && $options['return_type'] !== 'keyed') {
      throw new \Exception("resolveEntityIds called unknown return_type option value '$options[return_type]'.");
    }

    // Get data from cache.

    $precached_data = array();
    if (isset($this->entityIds)) {

      if (!$this->settings['import_ids_case_sensitive']) {
        // Convert $src_ids to uppercase (because our query system does not
        // distinguish case) - but take care not to change numbers to strings.
        // (It doesn't matter for cache array keys but may matter for e.g.
        // query arguments.)
        foreach ($src_ids as &$src_id) {
          if (is_string($src_id)) {
            $src_id = strtoupper($src_id);
          }
        }
      }

      // Extra code is introduced in various places to hopefully speed things up
      // by not doing array_intersect()'s when we're only looking for one item.
      // That extra code checks $single_id.
      $single_id = count($src_ids) == 1 && empty($options['requery_all'])
        ? ($this->settings['import_ids_case_sensitive']
          ? strtoupper(reset($src_ids)) : reset($src_ids))
        : NULL;

      if (empty($options['requery_all'])) {
        // We need to search all bundle-specific bins for IDs cached there.
        foreach (array_keys($this->entityIds) as $bin) {
          if (strstr($bin, ':', TRUE) === $entity_type) {
            // Optimized code (but same result as falling through to 'else'):
            if (isset($single_id)) {
              if (isset($this->entityIds[$bin][$single_id])) {
                return empty($options['return_type'])
                  ? $this->entityIds[$bin][$single_id]
                  : array($single_id => $this->entityIds[$bin][$single_id]);
              }
            }
            else {
              $precached_data += array_intersect_key($this->entityIds[$bin], array_flip($src_ids));
              if (count($precached_data) === count($src_ids)) {
                break;
              }
            }
          }
        }
      }
      else {
        $this->entityIds[$options['cache_bin']] = array();
      }
    }

    // Get data from database (or: additional data, from database into cache)

    if (!isset($this->entityIds) || count($precached_data) !== count($src_ids) || !empty($options['requery_all'])) {

      $data = $this->executeResolveEntityIdsQuery(
        $precached_data ? array_diff($src_ids, array_keys($precached_data)) : $src_ids,
        $entity_type,
        $options
      );
      if (!isset($this->entityIds)) {
        // If caching is off, the queried data is already in the return format.
        return $data;
      }
    }

    // Status at this point:
    // - Caching is on.
    // - We've returned items for a single cached item already (only for speed).
    // - We have 2 sets of data now:
    //   - $precached_data (the items already cached before this call)
    //   - the cache in $this->entityIds. Also $data *may* hold the full data
    //     for the current bin but that's not better than using the full cache.

    // Compile now-cached data.

    $bin = $this->entityIds[$options['cache_bin']];

    if (isset($single_id)) {
      // Optimized code (would have the same result as the below):
      if (isset($bin[$single_id])) {
        return empty($options['return_type'])
           ? $bin[$single_id] : array($single_id => $bin[$single_id]);
      }
      return array();
    }

    // Add new cached items, turning 'pre-cached data' into 'all needed data'.
    if (empty($options['requery_all'])) {
      // The cached data we need may be divided over different bins, so we can't
      // just get everything from the cache with one assignment. So, add to
      // $precached_data from the bin that got the new items.
      // array_merge() renumbers numeric keys, so use +=.
      $precached_data += array_intersect_key($bin, array_flip($src_ids));
    }
    else {
      $precached_data = $bin;
    }

    if (!empty($options['return_type'])) {
      // Return 'keyed' format, which is as stored in cache bin.
      return $precached_data;
    }
    // Transform $precached_data into a non-keyed array.
    $return = array();
    foreach ($precached_data as $dest_ids) {
      // Probably faster than array_merge() because $dest_ids is short:
      foreach ($dest_ids as $dest_id) {
        $return[] = $dest_id;
      }
    }
    return $return;
  }

  /**
   * {@inheritdoc}
   */
  protected function executeResolveEntityIdsQuery(array $src_ids, $entity_type, array $options) {
    $data = array();

    // By default, the importer stored the 'entity field containig the primary
    // key value' in $this->context, because it's not necessarily filled when
    // using a field mapping.
    // Some classes do not use a field mapping, and only have the field set in
    // $this->settings.
    $dest_field = isset($this->context['import_dst_pkey']) ?  $this->context['import_dst_pkey']
      : (isset($this->settings['import_dst_pkey']) ? $this->settings['import_dst_pkey']
        : '');
    if ($entity_type && $entity_type != $this->settings['import_entity_type']) {
      throw new \Exception("executeResolveEntityIdsQuery() was called for destination entity of type $entity_type, but no functionality has been defined for this.");
    }
    if (!$dest_field) {
      throw new \Exception("executeResolveEntityIdsQuery() was called while 'Source id field in destination entity' was not set.");
    }
    if (!empty($options['requery_all']) && empty($this->settings['import_ids_resolve_fast'])) {
      // Think further if you ever get to this point.
      throw new \Exception("executeResolveEntityIdsQuery() was called with 'requery_all' option while 'import_ids_resolve_fast' was not set. This is way too slow, sorry.");
    }

    $field_info_dest_field = field_info_field($dest_field);

    if (!empty($this->settings['import_ids_resolve_fast'])) {

      // faster way

      if (isset($field_info_dest_field)) {
        // Resolve FieldAPI 'value' in field_data table.

        $db_valuefield = $dest_field . '_value';
        $db_idfield = 'entity_id';

        $query = db_select('field_data_' . $dest_field, 'e');
        $query->condition('entity_type', $this->settings['import_entity_type']);
        if (!empty($options['bundle'])) {
          $query->condition('bundle', $options['bundle']);
        }
      }
      else {
        // Resolve simple property in main entity table.

        $db_valuefield = $dest_field;
        $info = entity_get_info($this->settings['import_entity_type']);
        $db_idfield = $info['entity keys']['id'];

        $query = db_select($info['base table'], 'e');
        if (!empty($options['bundle']) && !empty($info['entity keys']['bundle'])) {
          $query->condition($info['entity keys']['bundle'], $options['bundle']);
        }
      }

      if (empty($options['requery_all'])) {
        if (count($src_ids) == 1) {
          $query->condition($db_valuefield, reset($src_ids));
        }
        else {
          $query->condition($db_valuefield, $src_ids, 'IN');
        }
      }
      if (!empty($this->settings['import_entity_query_filter_field'])
          && !empty($this->settings['import_entity_query_filter_value'])) {

        if (isset($field_info_dest_field)) {
          // We assume this is a field in the entity table so we should join it.
          $info = entity_get_info($this->settings['import_entity_type']);
          $query->innerJoin($info['base table'], 'ee', 'e.entity_id=ee.' . $info['entity keys']['id']);
        }
        $query->condition(
          $this->settings['import_entity_query_filter_field'],
          $this->settings['import_entity_query_filter_value'],
          is_array($this->settings['import_entity_query_filter_value']) ? 'IN' : '='
        );
      }

      // Get results (faster way)
      if (isset($this->entityIds)
          || (isset($options['return_type']) && $options['return_type'] === 'keyed')) {
        // array( Source => array( Eid ) )

        if (isset($this->entityIds)) {
          // Append results directly to cache; construct return value later.
          $data =& $this->entityIds[$options['cache_bin']];
        }

        $query->fields('e', array($db_valuefield, $db_idfield));
        $result = $query->execute();
        foreach ($result as $row) {
          // Uppercase source IDs when necessary. Number to string conversion
          // does not matter here, because they're only used as array key.
          $src_id = $this->settings['import_ids_case_sensitive'] ?
            $row->$db_valuefield : strtoupper($row->$db_valuefield);
          if (!isset($data[$src_id])) {
            $data[$src_id] = array();
          }
          $data[$src_id][] = $row->$db_idfield;
        }
      }
      else {
        // array( Eid )  (0-based numeric keys)
        $query->addField('e', $db_idfield);
        $data = $query->execute()->fetchCol();
      }
    }
    else {

      // portable way

      $keyed = isset($options['return_type']) && $options['return_type'] === 'keyed';
      if (isset($this->entityIds) && !$keyed && count($src_ids) > 1) {
        // We'd have to split up the query into several queries, only to be able
        // to cache it (and do less queries a theoretical next time). Let's not.
        $options['cache_bin'] = '';
      }

      if ($keyed || $options['cache_bin']) {
        // array( Source => array( Eid ) )

        if ($options['cache_bin']) {
          // Append results directly to cache; construct return value later.
          $data =& $this->entityIds[$options['cache_bin']];
        }

        // We need to query each $src_id separately because $src_id cannot be
        // part of the query results.
        foreach ($src_ids as $src_id) {
          // No int-to-string conversion here; it might be bad in the query?
          if (is_string($src_id) && !$this->settings['import_ids_case_sensitive']) {
            $src_id = strtoupper($src_id);
          }

          $query = new EntityFieldQuery();
          $query->entityCondition('entity_type', $this->settings['import_entity_type']);
          if (!empty($options['bundle'])) {
            $query->entityCondition('bundle', $options['bundle']);
          }
          if (empty($options['requery_all'])) {
            if (isset($field_info_dest_field)) {
              $query->fieldCondition($dest_field, 'value', $src_id);
            }
            else {
              $query->propertyCondition($dest_field, $src_id);
            }
          }
          $result = $query->execute();
          $dest_ids = array_keys($result[$this->settings['import_entity_type']]);

          if ($dest_ids) {
            $data[$src_id] = $dest_ids;
          }
        }
      }
      else {
        // array( Eid )  (0-based numeric keys)

        $query = new EntityFieldQuery();
        $query->entityCondition('entity_type', $this->settings['import_entity_type']);
        if (!empty($options['bundle'])) {
          $query->entityCondition('bundle', $options['bundle']);
        }

        if (isset($field_info_dest_field)) {
          // Not keyed: Resolve FieldAPI field (portable)
          if (empty($options['requery_all'])) {
            if (count($src_ids) === 1) {
              $query->fieldCondition($dest_field, 'value', reset($src_ids));
            }
            else {
              $query->fieldCondition($dest_field, 'value', $src_ids, 'IN');
            }
          }
        }
        else {
          // Not keyed: Resolve simple property (portable)
          if (empty($options['requery_all'])) {
            if (count($src_ids) === 1) {
              $query->propertyCondition($dest_field, reset($src_ids));
            }
            else {
              $query->propertyCondition($dest_field, $src_ids, 'IN');
            }
          }
        }
        if (!empty($this->settings['import_entity_query_filter_field'])
            && !empty($this->settings['import_entity_query_filter_value'])) {
          // We assume this is a field in the entity table, like above, i.e. a
          // propertyCondition.
          $query->propertyCondition(
            $this->settings['import_entity_query_filter_field'],
            $this->settings['import_entity_query_filter_value'],
            is_array($this->settings['import_entity_query_filter_value']) ? 'IN' : '='
          );
        }

        $result = $query->execute();
        $data = array_keys($result[$this->settings['import_entity_type']]);
      } // not keyed
    } // portable way

    return $data;
  }

  /**
   * {@inheritdoc}
   */
  protected function getEntityId($entity) {
    if (isset($entity)) {
      return entity_id($this->settings['import_entity_type'], $entity);
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function loadEntities(array $ids) {
    return entity_load($this->settings['import_entity_type'], $ids);
  }

  /**
   * {@inheritdoc}
   */
  protected function updateEntity($entity = FALSE) {

    // Check if we should compare 'changed dates' as per import_skip_unchanged.
    // (Note this isn't necessary if setEntityValue() correctly reports when
    // all values are unchanged. It can still save a little processing though.)
    $update = empty($this->settings['import_skip_unchanged']);
    if (!$update) {
      // import_src/dst_changed must be defined, and the fields must
      // have values, in both the source item and the destination
      // entity, in order to be able to check.
      $update = (empty($this->context['import_src_changed']) ||
                 empty($this->context['import_dst_changed']) ||
                 empty($item[$this->context['import_src_changed']]) ||
                 empty($entity->{$this->context['import_dst_changed']}[LANGUAGE_NONE][0]['value']));
      if (!$update) {
        // Convert to timestamps (if the destination field isn't one
        // already) and compare values.
        $entity_changed_value = $entity->{$this->context['import_dst_changed']}[LANGUAGE_NONE][0]['value'];
        if (!is_numeric($entity_changed_value))  {
          $entity_changed_value = strtotime($entity_changed_value);
        }
        // The 'changed' field in the entity is likely only used by this
        // import, so it doesn't matter if we use > or != for
        // comparison. Only if something else is also setting
        // $entity_changed_value, there will be a difference between:
        // - 1 / >: imported item overwrites local entity if item has
        //   been changed more recently.
        // - 2 / !=: imported item overwrites local entity if the local
        //   entity (date) was changed in any way, so local changes are
        //   more likely to immediately be overwritten again.
        $update = $this->settings['import_skip_unchanged'] == 2
          ? strtotime($this->src_item[$this->context['import_src_changed']]) != $entity_changed_value
          : strtotime($this->src_item[$this->context['import_src_changed']]) > $entity_changed_value;
      }
    }
    if (!$update) {
      return 0;
    }

    if ($entity) {
      // Set bundle & language codes for this entity, for use by other methods.
      // (If the entity is new, we do it in initEntity().)
      $info = entity_get_info($this->settings['import_entity_type']);
      if (!empty($info['entity keys']['bundle'])) {
        $this->current_entity_bundle = $entity->{$info['entity keys']['bundle']};
      }
      else {
        reset($info['bundles']);
        $this->current_entity_bundle = key($info['bundles']);
      }

      if (!empty($info['entity keys']['language'])) {
        $this->field_langcodes = field_language($this->settings['import_entity_type'], $entity, NULL, $entity->{$info['entity keys']['language']});
      }
    }

    return parent::updateEntity($entity);
  }

  /**
   * {@inheritdoc}
   */
  protected function initEntity() {

    // Create new entity; set current_entity_bundle for the remainder of
    // updateEntity().
    $info = entity_get_info($this->settings['import_entity_type']);
    $values = array();
    if (!empty($info['entity keys']['bundle'])) {
      $values[$info['entity keys']['bundle']] = $this->current_entity_bundle =
        $this->settings['import_new_bundle'];
    }
    else {
      reset($info['bundles']);
      $this->current_entity_bundle = key($info['bundles']);
    }
    $entity = entity_create($this->settings['import_entity_type'], $values);

    // Set default owner ID.
    // Hardcoded assumption: (almost) every entity has a 'uid' field.
    if ($this->settings['import_entity_type'] !== 'user') {
      $uid = $GLOBALS['user']->uid;
      if (!empty($this->settings['import_entity_default_owner_id'])
          && is_int($this->settings['import_entity_default_owner_id'])
          && (!$uid || !empty($this->settings['import_entity_always_use_default_owner_id']))) {
        $uid = $this->settings['import_entity_default_owner_id'];
      }
      $entity->uid = $uid;
    }

    if (!empty($info['entity keys']['language'])) {
      // Set default language:
      // - if entity translation is enabled, we follow its language settings for
      //   this entity type;
      // - if not, we set LANGUAGE_NONE. (Not the default site language. There's
      //   no hard reason for one or the other; this just feels a bit better.)
      if (function_exists('entity_translation_enabled')
          && entity_translation_enabled($this->settings['import_entity_type'], $entity)) {
        $handler = entity_translation_get_handler($this->settings['import_entity_type'], $entity);
        $entity->{$info['entity keys']['language']} = $handler->getDefaultLanguage();
      }
      else {
        $entity->{$info['entity keys']['language']} = LANGUAGE_NONE;
      }

      // Set language codes for a new entity, for use by other methods.
      // (For non-new entities we do this in updateEntity().)
      $this->field_langcodes = field_language($this->settings['import_entity_type'], $entity, NULL, $entity->{$info['entity keys']['language']});
    }

    return $entity;
  }

  /**
   * {@inheritdoc}
   * 'parts' are interpreted as language codes that may need special handling.
   */
  protected function saveEntity($entity, array $changed_parts = array('*' => TRUE)) {

    $old_id = $this->getEntityId($entity);
    if (!$changed_parts) {
      return $old_id;
    }

    // Entity specific checks to see if the entity really should be saved:

    if ($this->settings['import_entity_type'] === 'commerce_customer_profile'
        && $old_id
        && function_exists('commerce_order_commerce_customer_profile_can_delete')
        && !commerce_order_commerce_customer_profile_can_delete($entity)) {
      // commerce_order_commerce_customer_profile_presave() will determine that
      // the profile cannot be updated, if it is already attached to a non-cart
      // order, and save a new profile instead.
      // For import jobs like this one, it's not good to let that happen because
      // new duplicate profiles would be created with every job run.
      // Notes:
      // - we take commerce_order_commerce_customer_profile_can_delete(), not
      //   commerce_customer_profile_can_delete(), because that's what the
      //   presave hook does too. (Conceptually I guess it's not better to
      //   use commerce_customer_profile_can_delete() because 'can delete'
      //   could be conceptually different from 'can update'.)
      // - Update is skipped silently. Maybe that's not always good:
      //   - if a user has several customer profiles (which is likely) then
      //     the 'non updatable' ones are skipped, which is good
      //   - but if a user has no updatable profiles, they all will be skipped
      //     and a change of data in the source system will not create a new
      //     profile. This is probably not what you want.
      return $old_id;
    }

    // Handle field translation metadata.

    if (function_exists('entity_translation_enabled')
        && entity_translation_enabled($this->settings['import_entity_type'], $entity)) {

      $uid = $GLOBALS['user']->uid;
      if (!empty($this->settings['import_entity_default_owner_id'])
          && is_int($this->settings['import_entity_default_owner_id'])
          && (!$uid || !empty($this->settings['import_entity_always_use_default_owner_id']))) {
        $uid = $this->settings['import_entity_default_owner_id'];
      }

      // In updateEntity() we have set entity field values as defined in the
      // mapping, irrespective of whether translation is enabled or not.
      // Here, we make sure translation 'metadata' is set, which is used in
      // the content admin UI to signify that a translation exists. (We do not
      // touch any 'actual' entity field values.) This is not done on save
      // automatically: entity_translation_field_attach_insert() only
      // initializes the translation for the entity language (which we've set
      // already) and ..._update() does nothing with translation data. They
      // only save the data that already exists.
      $handler = entity_translation_get_handler($this->settings['import_entity_type'], $entity);
      if ($handler->isNewEntity()) {
        // Before adding translation data, make sure the entity language is
        // initialized.
        $handler->initTranslations();
      }
      // Get existing translations.
      $translations = $handler->getTranslations();
      $source_langcode = $handler->getLanguage();

      // Silently skip $changed_parts that are not a language. (They should
      // not exist, but if they do it's not our responsibility.)
      foreach (array_intersect_key($changed_parts, language_list()) as $langcode => $v) {
        // Unless translate:LC or status:LC is in the field mapping, we set
        // translate to 0 (up-to-date) and status to 1 for new translations
        // and do not change existing translations. We do not try to see if
        // any translations need updating because we don't know if an entity
        // has extra fields that are populated by hand. (If anyone has a
        // bright idea about the value of 'translate', they are welcome.)
        $translation = array(
          'language' => $langcode,
          'status' => 1,
          'translate' => 0,
          'uid' => $uid,
        );
        $force_update = FALSE;
        // If *:LC properties were in the mapping, they were set as entity
        // properties by setProperty(). Check those, and assume no other code
        // has set these unless we should use them here.
        if (isset($entity->{"status:$langcode"})) {
          $translation['status'] = $entity->{"status:$langcode"};
          $force_update = TRUE;
        }
        if (isset($entity->{"translate:$langcode"})) {
          $translation['translate'] = $entity->{"translate:$langcode"};
          $force_update = TRUE;
        }

        if ($force_update || !isset($translations->$langcode)) {
          // Set new translation (for new or updated entity), or update
          // existing translation with values from the mapping.
          // (This means a translation record will be created for the source
          // language too, but _only_ if it's in $changed_parts, i.e. if a
          // language prefix was explicitly mentioned in the mapping. We'll
          // fix this when someone has an opinion about it.)
          if ($langcode != $source_langcode) {
            $translation['source'] = $source_langcode;
          }
          $handler->setTranslation($translation);
        }
      }
      // entity_translation_sync() gets called on ET_field_attach_presave();
      // saveTranslation() gets called on ET_field_attach_insert()/update().
    }

    // Save.

    $return = FALSE;
    try {
      $return = entity_save($this->settings['import_entity_type'], $entity);
    }
    catch (Exception $e) {
      $this->log('Error while saving @entity for item %item: %e',
        array(
          '@entity' => $this->settings['import_entity_type'],
          '%item' => $this->src_item[$this->context['import_src_pkey']],
          '%e' => $e->getMessage(),
        ), WATCHDOG_ERROR);
    }
    // entity_save() can return NULL even though it succeeded, if the save was
    // done through a 'save callback' (which does not specify a return value).
    // Other methods shouldn't return NULL - let's assume it always means "we
    // don't know what happened, let's just return entity_id()".
    // FALSE does mean failure.
    if ($return === FALSE) {
      return $return;
    }

    $id = $this->getEntityId($entity);

    // Check for a change in ID. (This can happen in presave hooks like
    // commerce_order_commerce_customer_profile_presave(); we actually catch
    // that one earlier on, but still it seems worth the check.)
    if ($old_id && $id != $old_id) {
      // We don't have a practical application for this - but for now: log
      // if an ID changes. It may warrant a change in code to e.g. prevent
      // duplicates being saved. (See the check just above for profiles.) Then
      // again, it may be fine, if e.g. the status of the old entity was reset.
      $this->log('After saveEntity() saved an existing @type (source ID @src_id), its entity ID changed from @old_id to @new_id. This may indicate that there is now duplicate data.',
        array('@type' => $this->settings['import_entity_type'], '@old_id' => $old_id, '@new_id' => $id, '@src_id' => $this->src_item[$this->context['import_src_pkey']]), WATCHDOG_WARNING);
    }

    if (isset($this->entityIds) && (!$old_id || $id != $old_id)) {
      // Cache the new id, or the changed existing-id (while discarding the old
      // id).

      // Strings must be uppercased usually.
      $src_id = $this->src_item[$this->context['import_src_pkey']];
      if (is_string($src_id) && !$this->settings['import_ids_case_sensitive']) {
        $src_id = strtoupper($src_id);
      }
      // We don't support creating multiple new entities for one source item.
      if (!$old_id && !empty($this->entityIds[$this->settings['import_entity_type'] . ':' . $this->current_entity_bundle][$src_id])) {
        $this->log('saveEntity() just saved a new entity (@id) for a source item (@src_id) which already has a linked entity-id cached. This probably indicates a code error.',
          array('@id' => $id, '@src_id' => $this->src_item[$this->context['import_src_pkey']]), WATCHDOG_WARNING);
        // We're not recording $id now. Not sure what the best policy is.
      }
      else {
        $this->entityIds[$this->settings['import_entity_type'] . ':' . $this->current_entity_bundle][$src_id] = array($id);
      }
    }
    return $id;
  }

  /**
   * Parses destination string; checks if it's a FieldAPI field and if so, which
   * language is attached.
   *
   * Also sets class-wide 'destination' properties, for use by methods setting
   * the value in this destination. (processedItemValue(), buildEntityValue()
   * and setEntityValue().) The 'type' is not a class-wide property, only a
   * return argument; there is no conscious design decision for this; things
   * just ended up this way.
   *
   * @return array
   *   List of values:
   *   - The real destination field; equal to the 1st parameter except if a
   *     valid language suffix has been found for a FieldAPI field
   *   - 'type' of destination field: 'fieldapi' for a FieldAPI field, or ''.
   *   - the language suffix for a FieldAPI field where one has been found;
   *     the entry from $field_langcodes for FieldAPI field where no suffix
   *     found; any suffix after ':' for non-FieldAPI fields.
   */
  protected function parseDestinationKey($destination) {

    $this->current_dst_field = $this->current_dst = $destination;

    // Split off sub-destination (the suffix after the last colon) if present.
    $sub_dest = '';
    if (strpos($destination, ':')) {
      $parts = explode(':', $destination);
      $sub_dest = array_pop($parts);
      $destination = implode(':', $parts);
      $this->current_dst_field = $destination;
    }

    $dest_type = '';
    $instance_info = field_info_instances($this->settings['import_entity_type'], $this->current_entity_bundle);

    // Check if $destination is an API field with a valid field prefix.
    if (isset($instance_info[$destination])
        && $field_info = field_info_field($destination)) {

      $dest_type = 'fieldapi';

      // Make sure sub_dest is set.
      if (isset($this->field_langcodes[$destination])) {

        if ($sub_dest) {
          if (!in_array($sub_dest,
            field_available_languages($this->settings['import_entity_type'], $field_info),
            TRUE)) {
            // This is not a valid language + field name pair (even though the
            // field exists).
            // Treat it as a non-fieldAPI property because that's 'simple'; this
            // way, an invalid language will probably fall through in a save
            // operation, and not mess up other sub-values of the field.
            // (This should actually never happen because checkSettings()
            // prevents this situation...)
            $dest_type = '';
            $this->current_dst_field = $destination = $this->current_dst;
            $sub_dest = '';
          }
        }
        else {
          // Set default language for this field.
          $sub_dest = $this->field_langcodes[$destination];
        }
      }
      else {
        // Known field without language specification (strange), or entity does
        // not support language: set a default sub-destination.
        $sub_dest = LANGUAGE_NONE;
      }
    }

    return array($this->current_dst_field, $dest_type, $sub_dest);
  }

  /**
   * Sets (or resets) a value inside an entity.
   *
   * @param mixed $value
   *   Value to set.
   * @param $entity
   *   The entity to update.
   * @param string $destination
   *   Destination property name.
   * @param string $dest_type
   *   Type of destination: "fieldapi" = FieldAPI field; '' if we know it's a
   *   simple property. If not provided, this means the type and 'sub'
   *   destination will still be determined, and $value will still be run
   *   through buildEntityValue(). (If provided, it is assumed that the value is
   *   already fully 'built'.)
   * @param string $sub_dest
   *   'sub' destination name. if $dest_type is "fieldapi": the language code.
   *   Otherwise, should be "".
   *
   * @return string
   *   If the value was actually changed: language code. If not: ''.
   */
  protected function setEntityValue($value, $entity, $destination, $dest_type = NULL, $sub_dest = '') {

    $changed = '';

    // Support metatag for 'standard property name', with 'metatag:' as prefix.
    if ($dest_type !== NULL && !$sub_dest && strpos($destination, 'metatag:') === 0) {
      $destination = substr($destination, 8);
      // @todo support explicit language suffix, like we do with FieldAPI fields now.
      $lc = metatag_entity_get_language($this->settings['import_entity_type'], $entity);

      // Set metatag value.
      if (isset($value)) {
        if (!isset($entity->metatags[$lc][$destination]['value'])
            || !$this->isEntityValueEqual($value, $entity->metatags[$lc][$destination]['value'], $dest_type)) {
          $entity->metatags[$lc][$destination]['value'] = $value;
          $changed = '*';
        }
      }
      // Reset metatag value.
      elseif (isset($entity->metatags[$lc][$destination]['value'])) {
        unset($entity->metatags[$lc][$destination]);
        $changed = '*';
      }
    }

    else {
      $changed = parent::setEntityValue($value, $entity, $destination, $dest_type, $sub_dest);
    }

    return $changed;
  }

  /**
   * {@inheritdoc}
   * For FieldAPI values, $dest_type should be set to 'fieldapi'.
   */
  protected function buildEntityValue($value, $destination, $current_built_value = NULL, $dest_type = '', $source_field = '') {

    if ($dest_type === 'fieldapi') {
      // Build FieldAPI value; set initial value first.
      $field_info = field_info_field($this->current_dst_field);
      if (!isset($field_info)) {
        // The caller is responsible for making sure we never end up here!
        $this->log('buildApiFieldValues() was called for destination field %dst, which is not an API-Field for entity type %t. Skipping. (source field: %src)',
          array('%t' => $this->settings['import_entity_type'], '%dst' => $destination, '%src' => $source_field), WATCHDOG_ERROR);
        return array();
      }

      if ($current_built_value === NULL) {
        $current_built_value = array();
      }

      // An array is assumed to be a correct FieldAPI value-structure.
      $values = is_array($value) ? array($value)
        : $this->buildApiFieldValues($value, $current_built_value, $source_field);

      // Post process any FieldAPI value, whether 'built' or passed in directly.
      foreach ($values as &$value) {
        switch ($field_info['type']) {
          case 'addressfield':
            // Unify the addressfield format so that comparisons won't fail - not
            // only the comparison inside this job's setEntityValue(), but also e.g.
            // the one in commerce_order_commerce_customer_profile_presave():
            // Fill first/last name from full, or vice versa.
            // (This is almost the same as in addressfield_field_presave().

            if (isset($value['first_name']) && isset($value['last_name']) && !isset($value['name_line'])) {
              $value['name_line'] = $value['first_name'] . ' ' . $value['last_name'];
            }
            // The following if() is not same as addressfield_field_presave();
            // if the import passes us inconsistent data it should figure this
            // out itself, we won't bluntly overwrite first/last name fields.
            // (Besides, the logic is arguably flawed. Changing it won't matter
            // until addressfield_field_presave() changes too, though.)
            elseif (isset($value['name_line']) && !isset($value['first_name']) && !isset($value['last_name'])) {
              $names = explode(' ', $value['name_line']);
              $value['first_name'] = array_shift($names);
              $value['last_name'] = implode(' ', $names);
            }
            break;
        }
      }

      return array_merge($current_built_value, $values);
    }

    switch ($destination) {

      case 'roles':
        if ($this->settings['import_entity_type'] === 'user') {
          // 'roles' is an array, so it can have multiple values mapped to it.
          if (is_array($value)) {
            // We support setting an array with non-0-based-sequential keys -if
            // it was sequential keys, it wouldn't end up here as an array-
            // which is assumed to be array(RID => name, ...). But in that case
            // we'll assume this is not supposed to be merged into other roles
            // and will throw an error (as the parent does by default) if a
            // value is already set.
            // If you want to be able to merge roles, then pass them as separate
            // values (rids or names) instead of one structure. (First passing
            // a structure and then merging individual values is also OK.)
            //
            // Set 'authenticated' for the isEntityValueEqual() check. (Auth
            // user is never saved but always present; added on user_load().)
            $value[DRUPAL_AUTHENTICATED_RID] = 'authenticated user';
            return parent::buildEntityValue($value, $destination, $current_built_value, $dest_type, $source_field);
          }
          // We support names and IDs as passed values.
          // Note this could turn into generic code sometime later...
          if (!is_scalar($value)) {
            $this->log('%property property cannot have value @value.',
              array('%property' => 'roles', '@value' => print_r($value, TRUE)), WATCHDOG_ERROR);
          }
          else {
            if (!isset($this->context['lookup_cache']['user:roles'])) {
              $this->context['lookup_cache']['user:roles'] =
                db_query('SELECT rid, name FROM {role}')->fetchAllKeyed();
            }
            if (is_numeric($value)) {
              // Eventually user_save() will only check the IDs, but we look up
              // names for the isEntityValueEqual() check & to check existence.
              if (!isset($this->context['lookup_cache']['user:roles'][$value])) {
                $this->log('%property property cannot have value @value.',
                  array('%property' => 'roles', '@value' => $value), WATCHDOG_ERROR);
              }
              else {
                $current_built_value[$value] = $this->context['lookup_cache']['user:roles'][$value];
                $current_built_value[DRUPAL_AUTHENTICATED_RID] = 'authenticated user';
              }
            }
            else {
              // We don't support ID=0 (which is anonymous, and anon user will
              // never be saved by user_save() anyway). So that's fine for the
              // array_search check.
              if (!$id = array_search($value, $this->context['lookup_cache']['user:roles'], TRUE)) {
                $this->log('%property property cannot have value @value.',
                  array('%property' => 'roles', '@value' => $value), WATCHDOG_ERROR);
              }
              else {
                $current_built_value[$id] = $value;
                $current_built_value[DRUPAL_AUTHENTICATED_RID] = 'authenticated user';
              }
            }
          }
        }
        return $current_built_value;

      case 'domain_user':
        // domain_user is an array, so it can have multiple values mapped to it.
        if (is_array($value)) {
          // We support setting an array, see comments at 'roles' above.
          return parent::buildEntityValue($value, $destination, $current_built_value, $dest_type, $source_field);
        }
        // We support machine names and IDs as passed values.
        if (!is_numeric($value)) {
          // Try to look up the ID from the machine name.
          if (!is_string($value)) {
            $this->log('%property property cannot have value @value.',
              array('%property' => 'domain_user', '@value' => print_r($value, TRUE)), WATCHDOG_ERROR);
          }
          else {
            if (!isset($this->context['lookup_cache']['domain_user'])) {
              $this->context['lookup_cache'][$destination] =
                db_query('SELECT machine_name, domain_id FROM {domain_export}')
                ->fetchAllKeyed();
            }
            if (!isset($this->context['lookup_cache']['domain_user'][$value])) {
              $this->log('%property property cannot have value @value.',
                array('%property' => 'domain_user', '@value' => $value), WATCHDOG_ERROR);
            }
            else {
              $id = $this->context['lookup_cache']['domain_user'][$value];
              // For domain_user, values are actually kept in the array _keys_.
              // domain_user_load() duplicates these into the array values, so
              // we'll do that too, for the isEntityValueEqual() check.
              $current_built_value[$id] = $id;
            }
          }
        }
        elseif (!$value) {
          $this->log('%property property cannot have value @value.',
            array('%property' => 'domain_user', '@value' => '0'), WATCHDOG_ERROR);
        }
        else {
          // For domain_user, values are actually kept in the array _keys_.
          $current_built_value[$value] = $value;
        }
        return $current_built_value;

      case 'status':
      case 'promote':
        $value = $this->isItemValueTrue($value) ? 1 : 0;
        break;
    }

    return parent::buildEntityValue($value, $destination, $current_built_value, $dest_type, $source_field);
  }

  /**
   * Transforms a single value so it can be set directly into a specific
   * FieldAPI field.
   *
   * $this->current_dst/_field must be set, for this function to work properly.
   *
   * Not all field values are always added through this method (specifically:
   * buildItemValues() can return FieldAPI structures by itself) so do not let
   * code depend on that.
   *
   * @param $value
   *   Value to transform and add.
   * @param array $current_built_value
   *   (optional) Value(s) already built by a previous call to this method for
   *   the same destination. Empty array means 'no value yet'.
   * @param string $source_field
   *   (optional) Source field from the item where the value came from. This
   *   value should generally not be needed, but it could be useful to reach
   *   properties in the field mapping - or for logging. If not '', mapping must
   *   be present (for $this->current_dst).
   *
   * @return array
   *   Zero, one or more value structures that can be set in the FieldAPI field.
   *   (So, if a single structure, it's enclosed inside an array. Not merged
   *   into $current_built_value; that should be done outside this method.)
   */
  protected function buildApiFieldValues($value, array $current_built_value = array(), $source_field = '') {
    // If we've determined that we can't resolve a value in a previous call,
    // don't bother doing anything.
    if ($this->defer) {
      return array();
    }

    $dst = $this->current_dst;
    $field_info = field_info_field($this->current_dst_field);
    if (!isset($field_info)) {
      // The caller is responsible for making sure we never end up here!
      $this->log('buildApiFieldValues() was called for destination field %dst, which is not an API-Field for entity type %t. Skipping. (source field: %src)',
        array('%t' => $this->settings['import_entity_type'], '%dst' => $dst, '%src' => $source_field), WATCHDOG_ERROR);
      return array();
    }

    // Mapping may not be present for 'unknown source' (e.g. default) values.
    if (isset($this->context['import_field_mapping'][$dst])) {
      $dst_mapping = $this->context['import_field_mapping'][$dst];
    }
    else {
      $dst_mapping = array();
      if ($source_field) {
        // The caller is responsible for making sure we never end up here!
        $this->log('buildApiFieldValues() was called for destination field %dst, which has no field mapping defined. Continuing because the value is already known, but this may be an error in your code. (values passed: source field: %src; value %val)',
          array('%val' => $value, '%dst' => $dst, '%src' => $source_field), WATCHDOG_WARNING);
      }
    }

    if (isset($dst_mapping["process:$source_field"])) {
      $process = $dst_mapping["process:$source_field"];
    }
    else {
      // This will only happen if $source_field = '', for default values.
      $process = array();
      if (!$source_field) {
        // $source_field is -almost- only used for logging strings, so set it
        // to something better.
        $source_field = '<none>';
      }
    }

    switch ($field_info['type']) {
      case 'commerce_product_reference':
      case 'entityreference':
      case 'node_reference':
        // References to other items may not resolve because the destination
        // entity does not exist yet. If this is true for one of the values in a
        // multi-value field, set $this->defer. isProcessingDeferred() will
        // notify the caller later.

        $target_type = strstr($field_info['type'], '_reference', TRUE);
        if (!$target_type) {
//            if ($field_info['type'] == 'entityreference')
          // DB table where field data is stored has 'target_type'.
          // TODO Where does this come from? A required property in the mapping? Or can it be derived from something?
          throw new \Exception("Target type cannot be determined for destination field '$dst'', which has type '$field_info[type]'. This must never happen and is a code error.");
        }
        else {

          // Note: no restriction on bundle. This means that if it's cached
          // already, we can retrieve it - but if it's not, it will get cached
          // in such a way that it's unusable for later bundle-specific queries.
          $dest_ids = $this->resolveEntityIds(array($value), $target_type);
          if ($dest_ids) {
            foreach ($dest_ids as $dest_id) {
              switch ($field_info['type']) {
                case 'commerce_product_reference':
                  $value = array('product_id' => $dest_id);
                  break;

                case 'entityreference':
                  $value = array('target_id' => $dest_id, 'target_type' => '');
                  // TODO needs target_type! A required property in the mapping? Or can it be derived from something?
                  break;

                case 'node_reference':
                  $value = array('nid' => $dest_id);
                  break;
              }
            }
          }
          else {
            // Cut off processing.
            $this->defer = TRUE;
            return array();
          }
        }
        break;

      case 'commerce_price':
        // Set currency from 'currency_src' or 'currency_code';
        // also use 'currency_code' if 'currency_src' is defined but
        // the value in the source item is empty.
        if (!empty($process['currency_src'])
            && $vals = $this->buildItemValues($process['currency_src'])) {
          $cc = $vals[0];
        }
        elseif (!empty($process['currency'])) {
          $cc = $process['currency'];
        }
        else {
          $cc = commerce_default_currency();
        }
        $value = array('amount' => $value * 100, 'currency_code' => $cc);

        // Set included tax from 'tax_src' or 'tax_name'. We assume this holds the
        // system name of the corresponding Drupal commerce_tax_rate! If you have
        // another value in this system, the recommended approach is to define a
        // 'process' function in the mapping for the 'tax_src' field. (This field
        // doesn't even need a 'dest' property.)
        if (!empty($process['tax_src'])
            && $vals = $this->buildItemValues($process['tax_src'])) {
          $value['data'] = array('include_tax' => $vals[0], 'components' => array());
        }
        elseif (!empty($process['tax_name'])) {
          $value['data'] = array('include_tax' => $process['tax_name'], 'components' => array());
        }
        break;

      case 'datestamp':
        $value = array('value' => strtotime($value));
        break;

      case 'list_boolean':
        $value = array('value' => $this->isItemValueTrue($value) ? 1 : 0);
        break;

      case 'list_integer':
      case 'list_text':
        // Default is: you have the value and you need to find the key
        // (needed to save the field) among the allowed values.
        // resolve = FALSE means 'do nothing' (i.e. you have the field already)
        if (!isset($process['resolve']) || $process['resolve']) {
          $key = array_search($value, $field_info['settings']['allowed_values']);
          if ($key === FALSE) {
            $this->log('Value %val is not among the allowed values of the %dst destination field. (source field: %src)',
              array('%val' => $value, '%dst' => $dst, '%src' => $source_field), WATCHDOG_ERROR);
            $value = FALSE;
          }
          else {
            $value = $key;
          }
        }

        if ($value === FALSE) {
          return array();
        }
        $value = array('value' => $value);
        break;

      case 'taxonomy_term_reference':
        // Find vocabulary to save this in (preferably) in field settings. The
        // process key 'vocabulary' is ignored if we can! It will only be used
        // if there are 2 allowed vocabularies (or $field_info is strange).
        $vocab = '';
        if (isset($field_info['settings']['allowed_values'])
            && is_array($field_info['settings']['allowed_values'])
            && count($field_info['settings']['allowed_values']) == 1
            && !empty($field_info['settings']['allowed_values'][0]['vocabulary'])
        ) {
          $vocab = $field_info['settings']['allowed_values'][0]['vocabulary'];
        }
        elseif (!empty($process['vocabulary'])) {
          $vocab = $process['vocabulary'];
        }
        if ($vocab) {
          // Map the value to a taxonomy term in this vocabulary.

          // array(-1) == find terms at any level.
          // array(0) == force terms on first level.

          if (empty($dst_mapping['hierarchy']) &&
              !isset($process['parent_src'])) {
            $pids = array(-1);
          }
          // If this term has a parent field, we need to determine the parent
          // (and its parents etc) again, because we're not sure if these
          // parents themselves are in $current_built_value already. (They aren't, if
          // only leaf terms of a tree are added to the actual field value.)
          //
          // If we have the process value 'hierarchy = TRUE', we don't need to
          // re-determine the parent but can assume that all parents are
          // actually in $current_built_value.
          // This means we can just place our terms on top of:
          // - all values in $current_built_value, IF we are now adding the 2nd level
          //   (source field).
          // - the last single value, IF the number of parent levels is equal
          //   to the number of terms. (There usually is one term per level.)
          //
          // The above 2 documented conditions are likely cases, so it makes
          // sense to check for them, so we don't need to call ensureTerm().
          elseif (!empty($dst_mapping['hierarchy'])
                  && count($dst_mapping['src']) == 2) {
            $pids = $this->fieldValuesToPids($current_built_value);
          }
          elseif (!empty($dst_mapping['hierarchy'])
                  // e.g. if this is 3rd level:
                  // array_search() == 2 == previous levels.
                  && count($current_built_value) ==
                     array_search($source_field, $dst_mapping['src'])
          ) {
            $pids = $this->fieldValuesToPids(array_slice($current_built_value, -1));
          }
          else {
            $pids = array(0);

            // The term goes on the first level of the vocabulary, if:
            // - parent_src was explicitly set to 0 to force this
            // - parent_src value is empty. (If the term should be
            //   discarded instead, process['skip_if_empty_src'] = <parent_src>
            //   should be set so we won't end up here.)
            if (!empty($process['parent_src'])) {
              // Determine the parent. As the result of (a previous) ensureTerm()
              // can be more than one term, the stored parent can also be.

              // Process all ancestors, oldest to newest. (Don't cache tids,
              // just do process & lookup for each ancestor again. There's too
              // much potential-theoretical confusion with ancestor-paths
              // for 2 fields being different; a field along the path not
              // having 'dest' property; wanting raw vs processed values...)

              $ancestors = is_array($process['parent_src']) ?
                $process['parent_src'] : array($process['parent_src']);

              foreach ($ancestors as $ancestor) {
                $tids = array();
                foreach ($pids as $pid) {
                  // If value (buildItemValues) is empty, skip this level.
                  //   (For multi-ancestor terms this is the best equivalent to
                  //   'if empty, put at root'.)
                  // If value is multiple, set all the ancestors.
                  foreach ($this->buildItemValues($ancestor) as $pvalue) {
                    // Also if one term lookup yields multiple existing terms,
                    // set all of them.
                    $tids = array_merge($tids, $this->ensureTerm($pvalue, $vocab, $pid));
                  }
                }
                // If for some reason 'processed values' have all been unset
                // then we have no good way out; disregard this whole 'level'.
                if ($tids) {
                  $pids = $tids;
                }
              }
            }
          }

          // Look up terms for all (most probably one) parents. If there are
          // multiple parents, this will force multiple instances of the term!
          $tids = array();
          foreach ($pids as $pid) {
            // If one lookup yields multiple existing terms, set both.
            $tids = array_merge($tids, $this->ensureTerm($value, $vocab, $pid));
          }
          $values = array();
          foreach ($tids as $tid) {

            // Don't add duplicate term references.
            if (!$current_built_value
                || array_search(array('tid' => $tid), $current_built_value) === FALSE) {
              $values[] = array('tid' => $tid);
            }
          }
        }
        else {
          $this->log("No 'vocabulary' property defined in mapping for source field %source_field.",
            array('%source_field' => $source_field), WATCHDOG_ERROR, FALSE);
          $values = array();
        }
        return $values;

      case 'media':
        // In other cases, $field_info['type'] dictates the format of input data
        // 'media' is so broad, we will probably start needing a hint in
        // $process at some point. For now, we only support youtube URLs,
        // so we'll make that the 'default'. We'll decide on interpreting input
        // value or forcing a mapping property, sometime in the future.
        if (!module_exists('media_youtube')) {
          $this->log('media_youtube module not installed; %source_field cannot be interpreted',
            array('%source_field' => $source_field), WATCHDOG_WARNING, FALSE);
        }
        else {
          $handler = new MediaInternetYouTubeHandler($value);
          // save() will work correctly, OR validId() throws exception.
          try {
            $file = $handler->save();
            $value = array(
              'fid' => $file->fid,
              'title' => !empty($process['title_src'])
                         && isset($this->src_item[$process['title_src']])
                ? $this->src_item[$process['title_src']] : '',
              // field has 'data' key - idk, maybe for when 'fid' not filled?
            );
          }
          catch (Exception $e) {
            $this->log('Saving value %v into field %dst of item %item generated an error: %e.',
              array(
                '%dst' => $dst,
                '%v' => $value,
                '%item' => $this->src_item[$this->context['import_src_pkey']],
                '%e' => $e->getMessage(),
              ), WATCHDOG_ERROR, FALSE);
            return array();
          }
        }

        break;

      // type 'boolean' was not defined? Is always immediately correct?
      // type 'decimal' is not explicitly defined yet, which means there is no
      //   validation. Decimal values will be correctly saved in 'value'.

      default:

        // Transform this single value into a fieldAPI structure.
        // We recognize fields with a 'value' column defined.
        if (isset($field_info['columns']['value'])) {
          $field_value = array('value' => $value);

          // Check for presence of format/summary columns.
          // Only if they are defined for this API field, fill the default values;
          // either from the field mapping, or some general default.
          $format = NULL;
          if (isset($field_info['columns']['format'])) {
            if (isset($process['format'])) {
              $field_value['format'] = $process['format'];
            }
            elseif ($field_info['type'] == 'text_with_summary' || $field_info['type'] == 'text_long') {
              // A 'real filter' should be set. Set a safe default.
              $field_value['format'] = 'filtered_html';
            }
            else {
              $field_value['format'] = NULL;
            }
          }

          // Only call text_summary if the mapping has a 'summary_length'
          // defined; otherwise leave the 'summary' column empty.
          // (Leaving it empty does not give the same result as calling
          // text_summary(,,NULL).)
          if (isset($field_info['columns']['summary'])
              && !empty($process['summary_length'])
              && is_numeric($process['summary_length'])) {
            $field_value['summary'] =
              text_summary($value, $field_value['format'], $process['summary_length']);
          }
          $value = $field_value;
        }
        else {
          $this->log('Destination field %dst (type %type, source field %src) has no \'value\' column defined; unknown processing method.',
            array('%dst' => $dst, '%src' => $source_field, '%type' => $field_info['type']), WATCHDOG_ERROR, FALSE);
          return array();
        }
    }

    return array($value);
  }

  /**
   * Helper method to create array with parent tids from APIField values,
   * which are always assumed to have 'tid's.
   */
  private function fieldValuesToPids(array $array) {
    $return = array();
    foreach ($array as $value) {
      if (is_array($value) && isset($value['tid'])) {
        $return[] = $value['tid'];
      }
    }
    // In our specific case we need array(0) if no parents.
    return $return ? $return : array(0);
  }

  /**
   * Gets term by name from a specific vocabulary with a specific parent.
   * If the vocabulary or term does not exist, it is created.
   * $parent=0 means it's retrieved from the 'first level' and we work
   * with 'flat' vocabularies unless specified otherwise, so then it's no restriction.
   *
   * @param $name string
   *   Term name to get
   * @param $voc_name string
   *   Vocabulary name to get term from
   * @param int $parent
   *   -1 means anywhere in the tree (if the term is newly created it will be a
   *      first-level term); 0 means term must be on first level, otherwise this
   *      is the tid of the parent term.
   * @return array
   *   term id(s)
   */
  protected function ensureTerm($name, $voc_name, $parent = -1) {

    // If $parent = -1, we assume unique term names among the whole vocabulary;
    // otherwise, unique among children of $parent.
    $vid = $this->ensureVocabulary($voc_name);
    $tids = array();
    foreach (taxonomy_get_tree($vid, $parent == -1 ? 0 : $parent, $parent == -1 ? NULL : 1) as $term) {
      if ($term->name == $name) {
        $tids[] = $term->tid;
      }
    }

    // Create term if it does not exist.
    if (!$tids) {
      $term = new stdClass();
      $term->vid = $vid;
      $term->name = $name;
      $term->parent = $parent == -1 ? 0 : $parent;
      taxonomy_term_save($term);
      // tid is now present.
      $tids[] = $term->tid;
    }

    return $tids;
  }

  /**
   * Makes sure vocabulary exists, returns vid.
   */
  protected function ensureVocabulary($machine_name) {
    $vocabulary = taxonomy_vocabulary_machine_name_load($machine_name);
    if (!$vocabulary) {
      // This is only possible if a term field has a nonexistent (deleted?)
      // vocabulary configured, or something strange happened and the code fell
      // back to using a vocabulary name that was defined in the mapping.
      // If for any obscure reason you want to provide your names for newly
      // created vocabularies: override this function.
      $this->log('Missing vocabulary %voc was created during synchronization; no human name was configured!',
        array('%voc' => $machine_name));
      $vocabulary = new stdClass();
      $vocabulary->name = 'NEW VOCAB';
      $vocabulary->machine_name = $machine_name;
      taxonomy_vocabulary_save($vocabulary);
    }
    return $vocabulary->vid;
  }

  /**
   * Update 'deferred' values, in Drupal entities. These should be references
   * to other entities, that could not be determined at first saving the entity.
   * (There is no support for other types of 'deferred' values yet.)
   *
   * @param array $context
   *   The job context, which has arrays mapped by source ID =>:
   *    'import_ids'       : array: Source item ID => array( Drupal entity IDs ) <= only if cache was on
   *    'import_deferred'  : array: fieldname => array( Source values )
   *                                       (source IDs, in case of references)
   *    'import_old_values': array: Entity ID =>
   *                          array: fieldname =>
   *                           array: language => array(Old Drupal field values)
   *                                         (entity IDs, in case of references)
   *       Language is a one-element array, since only one language is filled.
   *       Entity ID can theoretically hold multiple entities; this is unusual.
   *
   * Other input: if $context['defer_update_entities'] is set to an array
   *   no entities are saved, but instead
   *   $context['defer_update_entities'][$this->settings['import_entity_type']] is filled with data
   *   that can be updated later by e.g. calling $this->updateEntityValues().
   *
   *   Dimensions in the array: EntityType -> EntityID -> FieldName -> LangCode.
   *   $context['defer_update_entities'] is unset by default.
   */
  protected function updateDeferred(array &$context) {
    if (!empty($context['import_deferred'])) {

      // Resolve all IDs first, so we only need one query (per entity type).
      $target_types = array();
      $to_resolve = array();
      foreach ($context['import_deferred'] as $item) {
        foreach ($item as $dest_field => $src_ids) {

          if (!isset($target_types[$dest_field])) {

            // Deduce entity type corresponding to this destination field.
            // Half-copy from code in buildApiFieldValues():
            $field_info = field_info_field($dest_field);
            if ($field_info['type'] == 'entityreference') {
              // DB table where field data is stored has 'target_type'.
              // TODO Where does this come from? A required property in the mapping? Or can it be derived from something?
              throw new \Exception("Target type cannot be determined for destination field '$dest_field'', which has type '$field_info[type]'. This must never happen and is a code error.");
            }
            if (empty($field_info['type'])
                || strlen($field_info['type']) <= 10
                || substr($field_info['type'], -10) !== '_reference') {
              throw new \Exception("Unknown action for deferred destination field '$dest_field'', which has type '$field_info[type]'. This must never happen and is a code error.");
            }
            $target_type = substr($field_info['type'], 0, strlen($field_info['type']) - 10);
            $target_types[$dest_field] = $target_type;
          }

          // Collect unknown source IDs.
          // (The 'main source ID' (key in $item) is always resolved already;
          // we only need to worry about the IDs inside the value array.)
          if (!empty($target_types[$dest_field])) {
            $target_type = $target_types[$dest_field];
            if (!isset($to_resolve[$target_type])) {
              $to_resolve[$target_type] = array();
            }

            // Array_merge for small 'src_ids' arrays
            foreach ($src_ids as $src_id) {
              $to_resolve[$target_type][] = $src_id;
            }
          }
        }
      }
      $dest_ids = array();
      if ($to_resolve) {
        // Get destination IDs (cached and uncached) by target type; no bundle.
        foreach ($to_resolve as $target_type => $src_ids) {
          $dest_ids[$target_type] =
            $this->resolveEntityIds(array_unique($src_ids), $target_type, array('return_type' => 'keyed'));
        }
        unset($to_resolve);
      }

      // Process all products, all reference fields per product...

      if ($dest_ids) {
        $update_values = array();
        foreach ($context['import_deferred'] as $item_src_id => $item) {
          if (empty($context['import_old_values'][$item_src_id])) {
            throw new \Exception("No 'import_old_values' recorded for item $item_src_id (which holds deferred values). This must never happen and is a code error.");
          }
          $dest_items = $context['import_old_values'][$item_src_id];

          foreach ($item as $dest_field => $src_values) {
            if (!empty($target_types[$dest_field])) {
              $target_type = $target_types[$dest_field];

              // Construct intended entity field value in double loop:
              // Get all destination IDs for all source field values.
              // If multiple entities hold the same source ID (stored in their
              // 'import_dst_pkey' field), we just put all the import_ids in the
              // list. ('Flattened', not distinguishing the different $src_values.)
              $new_value = array();
              foreach ($src_values as $src_ref_id) {
                if (!$this->settings['import_ids_case_sensitive']) {
                  // 'src_id' keys in $dest_ids are unified-uppercased, so
                  // do the same for all our reference values:
                  $src_ref_id = strtoupper($src_ref_id);
                }

                if (empty($dest_ids[$target_type][$src_ref_id])) {
                  $this->log('Item %referred_id (referenced from item %referrer_id, field %field) does not exist.',
                    array('%referred_id' => $src_ref_id, '%referrer_id' => $item_src_id, '%field' => $dest_field), WATCHDOG_WARNING);
                }
                else {
                  // Again a half-copy from code in buildApiFieldValues().
                  $field_info = field_info_field($dest_field);
                  foreach ($dest_ids[$target_type][$src_ref_id] as $dest_id) {
                    switch ($field_info['type']) {
                      case 'commerce_product_reference':
                        $new_value[] = array('product_id' => $dest_id);
                        break;

                      case 'entityreference':
                        $new_value[] = array('target_id' => $dest_id,
                          'target_type' => '',
                        // TODO needs target_type! A required property in the mapping? Or can it be derived from something?
                        );
                        break;

                      case 'node_reference':
                          $new_value[] = array('nid' => $dest_id);
                          break;

                    }
                  }
                }
              }

              if ($new_value) {
                // Process new field value.
                // Go through all destination entities for this source ID,
                // and check if the value has changed here.
                // (It probably has; if it hadn't, it would probably not have
                // been deferred in the first place. But check anyway.
                // For instance, there are cases where we can't resolve one ID
                // for a multi-value reference field, and then the value stays
                // the same in the end.)

                foreach ($dest_items as $item_dest_id => $dest_item) {
                  // 'old_values' must be a one-value array keyed by language,
                  // whether empty or not. If not, code is broken somewhere.
                  if (!is_array($dest_item[$dest_field])) {
                    $this->log('debug info: source-id @id has the following new-refs: @refs. Full old-dest-refs object: @destrefs',
                      array('@refs' => print_r($item, TRUE), '@destrefs' => print_r($dest_items, TRUE)), WATCHDOG_DEBUG);
                    throw new \Exception("Unknown old-destination-reference for source-id '$item_src_id', destination-field '$dest_field', in updateDeferred()! This must never happen and is a code error. See debug-level log for old-dest-refs dump if you have it.");
                  }
                  $old_value = reset($dest_item[$dest_field]);
                  if ($new_value !== $old_value) {

                    // Record new value for this destination entity, in new
                    // array, keyed by dest_id => dest_field => lang.
                    $langcode = key($dest_item[$dest_field]);
                    $update_values[$item_dest_id][$dest_field][$langcode] = $new_value;
                  }
                } // foreach dest-item (processing new field value)
              }
            }
          } // foreach field in item
        } // foreach item

        // ... and save them, now we have all processed/destination values.

// @todo: defer_update_entities as used in the below line is actually a setting.
// (Though nothing uses it at the moment.)
// Make it a proper setting (maybe) when you redo updateEntityValues() to re-queue the updates.
        if (!isset($context['defer_update_entities']) || !is_array($context['defer_update_entities'])) {
          $this->updateEntityValues($update_values);
        }
        else {
          $context['defer_update_entities'][$this->settings['import_entity_type']] =
            isset($context['defer_update_entities'][$this->settings['import_entity_type']]) ?
              drupal_array_merge_deep($context['defer_update_entities'][$this->settings['import_entity_type']], $update_values) :
              $update_values;
        }
      }
    }
  }

  /**
   * Load all entities mentioned, update specified FieldAPI values and save.
   *
   * Uses: $this->settings['import_entity_type'].
   *
   * @param array $values
   *   Array of destination id => value => langcode => FieldAPI value-structure.
   */
  protected function updateEntityValues(array $values) {
    if ($values) {
      $entities = entity_load($this->settings['import_entity_type'], array_keys($values));
      foreach ($values as $dest_id => $field_values) {
        if (!isset($entities[$dest_id])) {
          $this->log('Entity @id (of type @type) could not be loaded, while trying to update field(s) @fields.',
            array('@id'     => $dest_id,
                  '@type'   => $this->settings['import_entity_type'],
                  '@fields' => implode(', ', array_keys($field_values))),
            WATCHDOG_ERROR);
        }
        else {
          // Update all values for all fields.
          foreach ($field_values as $field_name => $lang_values) {
            // Realistically there is only one language, but why not loop?
            foreach ($lang_values as $langcode => $value) {
              $entities[$dest_id]->{$field_name}[$langcode] = $value;
            }
          }
          entity_save($this->settings['import_entity_type'], $entities[$dest_id]);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function isEntityValueEqual($built_value, $entity_value, $dest_type = '') {
    if ($dest_type !== 'fieldapi' || is_scalar($built_value)) {
      // We want to do the same funny "0 equal to '0' but not to NULL/''" logic
      // as the parent.
      return parent::isEntityValueEqual($built_value, $entity_value, $dest_type);
    }

    // Field API text-type fields have a 'safe_value' sub-value on
    // hook_field_load(), as a form of caching.
    $field_info = field_info_field($this->current_dst_field);
    switch ($field_info['type']) {
      case 'text_with_summary':
        foreach ($entity_value as &$field_value) {
          unset($field_value['safe_summary']);
        }
        // no break!
      case 'text':
      case 'text_long':
        foreach ($entity_value as &$field_value) {
          unset($field_value['safe_value']);
        }
        break;

      case 'addressfield':
        // The field in the entity (at least when loaded from an SQL db) has
        // many NULL sub values. Apparently after saving a profile during
        // checkout, these values are empty strings. A value with these
        // sub values missing should still be considered equal.
        $f = function($v) { return isset($v) && $v !== ''; };
        foreach ($built_value as &$field_value) {
          $field_value = array_filter($field_value, $f);
        }
        foreach ($entity_value as &$field_value) {
          $field_value = array_filter($field_value, $f);
        }
        break;
    }

    return $built_value == $entity_value;
  }
}
