<?php
/**
 * @file
 * Contains DrunkinsLinkedEntityImportJob.
 */

//@todo use an Exception type with args/tokens/translation context, in various places.

/**
 * Class DrunkinsLinkedEntityImportJob
 *
 * Extends import functionality to import data from one source item into a
 * combination of a 'main' entity plus one (or multiple existing) 'linked'
 * entity/ies. A 'main' entity has a property/field containing the source ID; a
 * 'linked' entity has a reference to the main one.
 */
class DrunkinsLinkedEntityImportJob extends DrunkinsEntityImportJob {

  // These should only be touched by switchProfileContext():
  protected $origContextSettings;
  protected $origContextResults;

  /* Settings in $this->settings, used by this class:
   *
   * - all settings mentioned in DrunkinsEntityImportJob
   *
   * The following have N as a 1-based numeric index; there are as many
   * settings as linked entity-type/bundles:
   * - import_default_values:N (array):
   *   Equivalent to the non-'main' setting, for the linked entity if this is
   *   going to be created too. (The 'regular' properties are for the main
   *   entity. Note "import_field_mapping" has no equivalent; the linked fields
   *   are inside the regular import_field_mapping, with the keys prefixed with
   *   N:.)
   * - import_entity_type:N (string):
   *   type of the linked entity.
   * - import_new_bundle:N (string):
   *   Also equivalent to the non-'main' setting, and cannot be overridden by
   *   an equivalent 'import_resolve_all_bundles' / 'import_cleanup_all_bundles'
   *   setting. (Linked entities' creation, query and clean-up is always
   *   confined to one bundle.) This setting is only allowed to be empty if
   *   the corresponding entity type has no bundles.
   *
   * - import_entity_query_filter_field:N:
   * - import_entity_query_filter_value:N: equivalent to the non-'main' setting.
   *
   * - import_main_entity_ref:N (string):
   *   Name of the reference field/property in the 'linked' entity, which holds
   *   the ID of the 'main' entity.
   *
   * - opt_cleanup:N (string/array): as opt_cleanup, for the 'linked' entities.
   *   It's copied to 'opt_cleanup' while switching context (like other
   *   settings), but nothing is being done with it by default: e.g. on finish()
   *   processItem() is only called once, with the 'main' context active. To
   *   clean up items on finish(), a subclass can e.g. implement processItem()
   *   to switch context and do its own thing.
   *   (There is no 'standard' cleanup behavior that is always right, which is
   *   why this is not done by default.)
   */

  /**
   * {@inheritdoc}
   */
  public function __construct(array $settings = array()) {
    parent::__construct($settings);

    // In case a subclass sets 'opt_list', we filter away arguments which
    // are useless in the 'display' URL.
    $i = 0;
    while (!empty($this->settings['import_entity_type:' . ++$i])) {
      $this->display_arguments_filter = array_merge($this->display_arguments_filter, array(
        "import_entity_type:$i",
        "import_new_bundle:$i",
        "import_default_values:$i",
        "import_main_entity_ref:$i",
        "opt_cleanup:$i",
      ));
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function checkSettings(array $context, $quiet_fast = FALSE) {

    $ok = parent::checkSettings($context, $quiet_fast);
    if (!$ok) {
      return FALSE;
    }

    $nr_linked = 1;
    while ($nr_linked) {
      $notfound = $found = array();
      foreach (array('entity_type', 'main_entity_ref') as $setting) {
        $key = "import_$setting:$nr_linked";
        if (empty($this->settings[$key])) {
          $notfound[] = $key;
        }
        else {
          $found[] = $key;
        }
      }
      if ($found) {
        if ($notfound) {
          if ($ok = $quiet_fast) {
            return FALSE;
          }
          $this->log('Discrepancy: @setting setting not configured, while @setting2 is!', array('@setting' => implode(' & ', $notfound), '@setting2' => implode(' & ', $found)), WATCHDOG_ERROR);
          break;
        }
      }
      else {
        break;
      }
      $nr_linked++;
    }

    for ($i = 1; $i <= $nr_linked; $i++) {
      // Linked bundles are required if the entity type has multiple bundles.
      if (!empty($this->settings["import_entity_type:$i"])
          && empty($this->settings["import_new_bundle:$i"])) {
        $info = entity_get_info($this->settings["import_entity_type:$i"]);
        if (count($info['bundles']) > 1) {
          if ($ok = $quiet_fast) {
            return FALSE;
          }
          $this->log("@setting setting is empty, while the configured entity type (@type) has bundles!",
              array('@setting' => "import_new_bundle:$i", '@type' => $this->settings["import_entity_type:$i"]), WATCHDOG_ERROR);
        }
      }

      if (!empty($this->settings["import_entity_query_filter_field:$i"])
          xor !empty($this->settings["import_entity_query_filter_value:$i"])) {
        if ($ok = $quiet_fast) {
          return FALSE;
        }
        $this->log("If one of '@setting1' and '@setting2' settings are set, they must both be set.",
          array('@setting1' => "import_entity_query_filter_field:$i", '@setting2' => "import_entity_query_filter_value:$i"), WATCHDOG_ERROR);
      }
    }

    // We don't do the following block when $quiet_fast - it arguably takes too
    // long to do it on every processItem(), and won't break anything horribly.

    if (!$quiet_fast && is_array($this->settings['import_field_mapping'])) {
      // We checked the mapping for the main entity in the parent method.
      // (There were also 'N:dest' mappings in there but they never generated
      // errors.) Here we generate linked-only mappings, and check those so it
      // does generate errors for nonexistent languages.
      $this->setMappingByDest($context);
      $i = 0;
      while (!empty($this->settings['import_entity_type:' . ++$i])) {
        foreach (array_keys($context["import:$i"]['import_field_mapping']) as $dst_key) {

          // Almost-copy from parent code, except log message has a prefix:
          $instance_info = field_info_instances($this->settings["import_entity_type:$i"]);
          if (strpos($dst_key, ':')) {
            $parts = explode(':', $dst_key);
            $lc = array_pop($parts);
            // Field may not have a ':' and we only check FieldAPI fields here:
            if (count($parts) == 1
                && isset($instance_info[$parts[0]])
                && ($field_info = field_info_field($parts[0]))
                && !in_array($lc, field_available_languages($this->settings["import_entity_type:$i"], $field_info), TRUE)
            ) {
              if ($ok = $quiet_fast) {
                return FALSE;
              }
              $this->log("'@key' is mentioned as a destination in the field mapping for linked entity @nr, but @lc is not an available language for the @field field!",
                array('@key' => ":$dst_key", '@nr' => $i, '@lc' => $lc, '@field' => $parts[0]), WATCHDOG_ERROR);
            }
          }

        }

        // Another check that can only be done after setMappingByDest() because
        // the 'ref_type' needs to be filled: see executeResolveEntityIdsQuery()
        // for details.
        if (!empty($this->settings["import_entity_query_filter_field:$i"])
            && !empty($this->settings["import_entity_query_filter_value:$i"])
            && $context["import:$i"]['import_main_entity_ref_type']) {
          if ($ok = $quiet_fast) {
            return FALSE;
          }
          $this->log("At this moment '@setting1 / @setting2' settings are only supported if @setting3 refers to a field in the linked @type's main table.",
            array('@setting1' => "import_entity_query_filter_field:$i", '@setting2' => "import_entity_query_filter_value:$i", '@setting3' => $context["import:$i"]['import_main_entity_ref_type'], '@type' => $this->settings["import_entity_type:$i"]));
        }
      }
    }


    return $ok;
  }

  /**
   * {@inheritdoc}
   */
  public function start(array &$context) {

    // This needs to be set before parent::start() is called, so
    // setMappingByDest() can use it. But that then caches the mapping - so if
    // start() is called several times to fetch items, do not reset cache.
    $i = 0;
    while (!empty($this->settings['import_entity_type:' . ++$i])) {
      if (empty($context["import:$i"])) {
        $context["import:$i"] = array(
          'import_deferred'   => array(),
          'import_old_values' => array(),
        );
      }
    }

    $items = parent::start($context);

    // Change mapping to stop 'check field mapping' action from complaining.
    // @todo move to ui_actions_class
    if (isset($context['import_op'])
        && $context['import_op'] == t('Check field mapping')) {

      // Merge profile destinations into normal destinations.
      $i = 0;
      while (!empty($this->settings['import_entity_type:' . ++$i])) {
        foreach ($context["import:$i"]['import_field_mapping'] as $dst => $dest_map) {
          $prefix = $this->settings["import_entity_type:$i"]
                    . (empty($this->settings["import_new_bundle:$i"])
              ? '' : ':' . $this->settings["import_new_bundle:$i"]);
          $context['import_field_mapping']["$prefix > $dst"] = $dest_map;
        }
      }
    }

    if (isset($this->origContextSettings)) {
      throw new \Exception("Incomplete context switching occured. This must never happen and can lead to data corruption.");
    }

    return $items;
  }

  /**
   * Largely the same code as DrunkinsEntityImportJob::processItem, with
   * customized lines in between.  (The ones mentioning $entity_id_map.)
   *
   * {@inheritdoc}
   */
  public function processItem($item, array &$context) {

    // TODO check on empty title / possibly other required fields, and warn.
    // The process does not need to be changed to abort, but possibly the user wants to know.

    // TODO or instead, write generic check+warn on required fields.

    if (!$this->checkSettings($context, TRUE)) {
      // We can't do anything. We've warned in start(), so exit silently.
      return;
    }

    // Fill reference to context so called/extending code can access it.
    // (If this method is called more than once on an already-instantiated
    // class, it's a no-op.)
    $this->context =& $context;
    // Turn on resolveEntityIds() caching.
    $this->entityIds =& $context['import_ids'];

    // If a special '(cleanup) action' key is the only thing present besides
    // an ID, we want to perform the mentioned actions instead of the regular
    // import.
    if (isset($item['*action']) && count($item) == 2) {

      // Check what kind of ID we have to operate on.
      if (isset($item['*entity_id'])) {
        // We have no source ID; destination entity already resolved.
        $entity_ids = is_array($item['*entity_id']) ? $item['*entity_id'] : array($item['*entity_id']);
      }
      else {
        // Resolve destination from source. If the second key in $item is not
        // the source ID field, the item will be skipped.
        $src_id = isset($item[$context['import_src_pkey']])
          ? trim($item[$context['import_src_pkey']]) : '';
        $entity_ids = $src_id === '' ? array() : $this->resolveEntityIds(array($src_id));
      }

      // Construct key to account for action in the finish() message. It
      // should be converted from array('action') to 'action' if possible.
      $action_key = $item['*action'];
      if (is_array($item['*action'])) {
        if (count($item['*action']) == 1) {
          $action_key = reset($item['*action']);
          if (!is_numeric(key($item['*action']))) {
            $action_key = $item['*action'];
          }
        }
        if (is_array($action_key)) {
          // We need some string representation.
          $action_key = json_encode($action_key);
        }
      }

      // If no mapped entity exists, skip the action silently. (This is good
      // for 'cleanup'.)
      if ($entity_ids) {

        $this->processEntities(
          is_string($item['*action']) ? array($item['*action']) : $item['*action'],
          $entity_ids
        );

        // Accounting is done in the same way, regardless whether this is an
        // item with a source ID or only an entity ID. In the former case, we
        // add only 1 even if multiple items were cleaned up. There's no real
        // thought behind this yet; it may need to be refined later.
        if (!isset($context['actions'][$action_key])) {
          $context['actions'][$action_key] = 0;
        }
        $context['actions'][$action_key]++;
      }
      else {
        if (!isset($context['actions_skipped'][$action_key])) {
          $context['actions_skipped'][$action_key] = 0;
        }
        $context['actions_skipped'][$action_key]++;
      }
    }

    else {
      // Process 'regular item':

      if (!empty($this->settings['import_trim_values'])) {
        // Trim all values, to prevent hard to trace bugs caused by leading /
        // trailing spaces.
        array_walk_recursive($item, function (&$value) {
          if (is_string($value)) {
            $value = trim($value);
          }
        });
      }

      $this->src_item = $item;
      // The source item's key must have a value since it's the unique ID. Any
      // method called after here can assume it exists.
      if ($this->isItemFieldEmpty($context['import_src_pkey'])) {
        $this->log("Item %nr has no 'Source id' value; will not be processed!", array('%nr' => $context['drunkins_progress']), WATCHDOG_WARNING);
        return;
      }

      $src_id = $item[$context['import_src_pkey']];
      // resolveEntityIds() / its return value has been extended: 'linked'
      // returns an array (main-id => array ( linked-ids )). In this method, the
      // key is important to us, not the value.
      $entity_id_map = $this->resolveEntityIds(array($src_id), '', array('return_type' => 'linked'));

      if (!$entity_id_map) {
        // Create new main + linked entity.

        $updated_dest_id = $this->updateEntity(NULL, array(array()));

        if ($updated_dest_id) {
          $context['new'][$updated_dest_id] = $src_id;
          // Cache new ID. Strings must be uppercased usually.
          if (is_string($src_id) && !$this->settings['import_ids_case_sensitive']) {
            $src_id = strtoupper($src_id);
          }
          $context['import_ids'][$this->settings['import_entity_type'] . ':' . $this->current_entity_bundle][$src_id] = array($updated_dest_id);
        }
        elseif ($updated_dest_id === 0) {
          $context['skipped'][] = $src_id;
        }
        else {
          $context['error'][] = $src_id;
        }
      }
      else {
        // Update all entities which are 'mapped' to this source item.

        $entities = $this->loadEntities(array_keys($entity_id_map));
        if ($entities) {
          $updated_dest_id = 0;
          foreach ($entities as $id => $entity) {
            // Update entity + linked entities; if a certain 'linked-type' has
            // no entities, create a new one of that type.
            $updated_dest_id = $this->updateEntity(
              $entity,
              $entity_id_map[$id] + array(array())
            );

            if ($updated_dest_id) {
              $context['updated'][$updated_dest_id] = $src_id;
            }
            elseif ($updated_dest_id !== 0) {
              $context['error'][] = $src_id;
            }
          }
          // For understandable statistics, report if a source item has more
          // than one entry (otherwise the numbers don't add up, without
          // explanation). We don't report which items had how many entities.
          if (count($entities) > 1) {
            $context['multiple']++;
          }
          elseif ($updated_dest_id === 0) {
            $context['skipped'][] = $src_id;
          }
        }
        else {
          // No loaded entities; should not happen.
          $context['error'][] = $src_id;
        }
      }
    }

    if (isset($this->origContextSettings)) {
      throw new \Exception("Incomplete context switching occured. This must never happen and can lead to data corruption.");
    }

    // It probably makes no difference, but let's not leave references to random
    // large data structures floating around anyway.
    unset($this->context);
    unset($this->context);
  }

  /**
   * {@inheritdoc}
   */
  public function finish(array &$context) {
    // It's a judgment call, what we want to do regarding post processing of
    // linked entities:
    // - call finish() multiple times, after switching context (but then we'd
    //   need to take care to not call fetcher finish() twice, but still to let
    //   fetcher's hasFulleDataSet() return the correct value twice. The commit
    //   history contains working code for this.)
    // - extend all post processing parts individually, and let those all
    //   switch context / call themselves multiple times.
    // The second way means we won't automatically do anything in the linked
    // context, that gets added to the parent finish() later. This could cause
    // as well as prevent bugs...
    // We've decided on the second thing for now, because that (particularly
    // the cleanup part) can prevent a query from being executed twice, if it
    // does the context switching by itself. Plus the things mentioned in the
    // first point are tedious.

    // Until now, updateDeferred() and 'cleanup' are the only things called
    // by the parent finish.
    // UpdateDeferred we'll call here ourselves instead of extending the
    // method, for now. (Note $this->context may or may not be set depending
    // on whether this class gets used, and the parent does not really care)
    $i = 0;
    while (!empty($this->settings['import_entity_type:' . ++$i])) {
      $this->switchLinkedContext($i);
      parent::updateDeferred($context["import:$i"]);
    }
    if ($i > 1) {
      $this->switchLinkedContext(0);
    }

    // The 'cleanup' part is left for processEntities() to extend and call
    // twice.
    $message = parent::finish($context);

    if (isset($this->origContextSettings)) {
      throw new \Exception("Incomplete context switching occured. This must never happen and can lead to data corruption.");
    }

    return $message;
  }

  /**
   * {@inheritdoc}
   *
   * Unlike most other properties (like default values) which are separate, the
   * main 'import_field_mapping' setting contains information for both main
   * and linked entities. Destination fields in the linked entity are prefixed
   * by ':'.
   */
  protected function setMappingByDest(array &$context)  {

    if (!empty($context['import_field_mapping'])) {
      return;
    }

    $nr_linked = 1;
    while (!empty($this->settings['import_entity_type:' . $nr_linked])) {
      $nr_linked++;
    }
    $nr_linked--;

    if ($nr_linked) {
      // First split our mapping in primary & linked entity, by key.
      $mapping = $full_mapping = $this->settings['import_field_mapping'];
      $linked_mappings = array();
      foreach ($full_mapping as $dst => $map) {
        if ($prefix = strstr($dst, ':', TRUE)) {
          if (is_numeric($prefix) && $prefix > 0 && $prefix <= $nr_linked) {
            $linked_mappings[$prefix][substr($dst, strlen($prefix) + 1)] = $map;
            unset($mapping[$dst]);
          }
        }
      }

      // Set normalized main-entity mapping.
      $this->settings['import_field_mapping'] = $mapping;
    }

    parent::setMappingByDest($context);

    if ($nr_linked) {

      $main_context = $context;
      $linked_index = 0;
      while ($linked_index++ < $nr_linked) {

        // Normalize & set linked mappings.
        $this->switchLinkedContext($linked_index);
        // Unset the context stuff that was just filled by our parent methods,
        // which would confuse the next call(s) to it.
        unset($context['import_field_mapping']);
        // We can keep the 'src' equivalents of the below properties; there's no
        // reason that a src-dst pair of context fields (as created by
        // setMappedFieldNames()) would be bad in a linked entity.
        unset($context['import_dst_pkey']);
        unset($context['import_dst_changed']);

        // Set normalized linked-entity mapping.
        $this->settings['import_field_mapping'] = isset($linked_mappings[$linked_index])
          ? $linked_mappings[$linked_index] : array();
        parent::setMappingByDest($context);

        // Restore main-entity context properties.
        $main_context["import:$linked_index"]['import_field_mapping'] = $context['import_field_mapping'];
        if (!empty($context['import_src_pkey'])) {
          $main_context["import:$linked_index"]['import_src_pkey'] = $context['import_src_pkey'];
        }
        if (!empty($context['import_dst_pkey'])) {
          $main_context["import:$linked_index"]['import_dst_pkey'] = $context['import_dst_pkey'];
        }
        // Probably the linked entity has no 'changed' field but who knows.
        if (!empty($context['import_src_changed'])) {
          $main_context["import:$linked_index"]['import_src_changed'] = $context['import_src_changed'];
        }
        if (!empty($context['import_dst_changed'])) {
          $main_context["import:$linked_index"]['import_dst_changed'] = $context['import_dst_changed'];
        }

      }
      $this->switchLinkedContext(0);
      $context = $main_context;

      // Put mapping setting back to 'full'. (We won't expect to use it beyond
      // here, but still, leave settings unchanged to prevent bugs.)
      $this->settings['import_field_mapping'] = $full_mapping;
    }


    $this->setLinkedEntityInfo($context);
  }

  /**
   * Constructs 'cached' (so it does not need to be recalculated every time we
   * need it) info about linked entities:
   * - in the 'main' context, all types/bundles/fieldnames of entity reference
   *   (which isn't actually that hard to derive, but checking 'switched status'
   *   for all properties & 'set-ness' of bundle each time is tiring.)
   * - the types of 'main entity references' in linked entities
   *
   * The reason that this is in a separate method is
   * - caching (of at least the main-reference-types) needs to be done in
   *   setMappingByDest()
   * - but this info is also needed in resolveEntityIds(), and we can't be 100%
   *   sure that all child classes calling this method actually use a mapping.
   *   (It's actually very unlikely! But we do have an example of a
   *   DrunkinsEntityImportJob child that does not use a mapping...)
   *
   * @param array $main_context
   *   Context array to check / set values into. Make sure this is the main one.
   *
   * @throws Exception
   */
  protected function setLinkedEntityInfo(array &$main_context) {

    // At the moment context is never switched, but let's not take the risk:
    $settings = isset($origContextSettings) ? $this->origContextSettings : $this->settings;

    $i = 0;
    $main_context['import']['linked_entities'] = array();
    while (!empty($settings['import_entity_type:' . ++$i])) {

      //@todo this is another reason why there should be a setting to turn caching on/off instead of a dynamic property: we need to trust it to not change so this check never passes by accident.
      if (isset($this->entityIds)
          && $settings["import_entity_type:$i"] === $settings['import_entity_type']) {
        throw new \Exception(t("'linked' entities with the same type as the 'main' entity are not supported with resolveEntityIds() caching on; this could lead to incomplete cached results."));
      }

      // This should always be true but is still checked to guard against PHP
      // notices when we call this from checkSettings():
      if (!empty($settings["import_main_entity_ref:$i"])) {

        // 1) set info in 'main context'
        $main_context['import']['linked_entities'][$i] = array(
          'type' => $settings["import_entity_type:$i"],
          'bundle' => empty($settings["import_new_bundle:$i"]) ? '' : $settings["import_new_bundle:$i"],
          'ref' => $settings["import_main_entity_ref:$i"],
        );
        if (empty($settings["import_new_bundle:$i"])) {
          // We already made this check in checkSettings() but want to be
          // extra careful in this method:
          $info = entity_get_info($settings["import_entity_type:$i"]);
          if (count($info['bundles']) > 1) {
            // We throw error here because we don't trust what the caller will
            // do with the results if we return too few. (e.g. cleanup)
            // @todo decide on translation status for throwing these errors. Want to return untranslated?
            throw new \Exception(t("'@setting' setting is empty, while the configured entity type (@type) has bundles!",
                array('@setting' => "import_new_bundle:$i", '@type' => $settings["import_entity_type:$i"])));
          }
        }

        // 2) set ref-type info in all individual contexts.
        $info = field_info_field($settings["import_main_entity_ref:$i"]);
        if (isset($info)) {
          // We need to know the difference between single and multi value
          // references because the argument to entityAPI set() is different.
          // (At least for commerce_product_reference fields...)
          $main_context["import:$i"]['import_main_entity_ref_type'] = 1;
          //@todo check and set 2 for multi value
          // Also derive fieldname in SQL table.
          // @todo make more complete. Give error for entityref because we can't support it until we add a filter on entity type before using the 'entity_id' field in queries.
          switch ($info['type']) {
            case 'commerce_product_reference':
              $main_context["import:$i"]['import_main_entity_ref_sqlfieldname'] =
                $settings["import_main_entity_ref:$i"] . '_product_id';
              break;
            default:
              $main_context["import:$i"]['import_main_entity_ref_sqlfieldname'] =
                $settings["import_main_entity_ref:$i"] . '_value';
          }
        }
        else {
          $main_context["import:$i"]['import_main_entity_ref_type'] = 0;
        }
      }
      else {
        // won't ever happen but let's be theoretically correct:
        unset($main_context["import:$i"]['import_main_entity_ref_type']);
      }
    }
  }

  /**
   * {@inheritdoc}
   *
   * An extra 'return_type' option value is supported:
   * "linked": return one array (
   *     main-entity-id => array( INDEX => array( linked-entity-ids ), ... )
   *   ) where INDEX is an integer corresponding to the linked entity. (1-based,
   *   but not necessarily consecutive because if no entities exist for this
   *   'linked type', the key does not exist either.)
   *
   * The 'query_type' option is used internally, and always overwritten.
   *
   * $this->context must be set before calling this method!
   *
   * @throws Exception
   */
  protected function resolveEntityIds(array $src_ids, $entity_type = '', array $options = array()) {
    // A note: the 'negative caching' could be an option (setting) - we usually
    // don't care because our standard classes populate the cache on
    // saveEntity(), so any effect of negative caching is canceled out in most
    // cases. However we don't need to requery entities usually, so we
    // haven't cared until now.
    // @todo we DO actually already do a sort of 'negative caching' here, for
    //   main-entities: if the corresponding linked entities are not cached,
    //   we assume they don't exist. Take that assumption out if we make this
    //   'negative caching' configurable.

    // This code assumes that if $options['requery_all'] is set, this does not
    // imply related (either linked or main) caches should be emptied too.
    // Whether that's a good principle... time will tell.

    $orig_return_type = isset($options['return_type']) ? $options['return_type'] : '';

    if ($orig_return_type === 'linked' && !isset($this->entityIds)) {
      // Querying directly is the only way to get correctly formatted data
      // (because we don't have a cache. Which is also why we don't need the
      // parent method to first try the cache.)
      // Overwrite 'query_type' just in case someone set it / to be predictable.
      $options['query_type'] = 'linked';
      return $this->executeResolveEntityIdsQuery($src_ids, $entity_type, $options);
    }

    // We need to know the queried entity type, but also want to pass the
    // empty value to a parent call because passing 'main-entity' explicitly
    // can influence the return value (if 'bundle' option not passed).
    $queried_entity_type = $entity_type ? $entity_type : $this->settings['import_entity_type'];

    $main_entity_type = isset($this->origContextSettings)
      ? $this->origContextSettings['import_entity_type']
      : $this->settings['import_entity_type'];

    if ($orig_return_type === 'linked'
        || (isset($this->entityIds) && $queried_entity_type === $main_entity_type)) {
      // If we're resolving the main entity type and caching is on, always query
      // 'linked', otherwise the cache for linked entities will have incomplete
      // data (which will be returned if source items are requeried later).
      $options['query_type'] = 'linked';
      // The return value from the parent method will still be in the right
      // (normal/keyed) format. Unless it's 'linked':
      if ($orig_return_type === 'linked') {
        // We want to have the parent class check the cache but that can't
        // handle "linked". So have it return 'keyed' type (because we need both
        // source and destination IDs - plus it's fastest)
        $options['return_type'] = 'keyed';

        if ($queried_entity_type !== $main_entity_type) {
          throw new \Exception("'linked' lookups only supported for the 'main' entity type in DrunkinsLinkedEntityImportJob::resolveEntityIds().");
        }
      }
    }
    else {
      $options['query_type'] = '';
    }

    // Get data from cache and/or database.

    $data = parent::resolveEntityIds($src_ids, $entity_type, $options);
    if ($orig_return_type !== 'linked') {
      // The queried data is already in the right format. (If not cached:
      // directly from the query. If cached: it was compiled afterwards.)
      return $data;
    }

    // Status: caching is on. executeResolveEntityIdsQuery() cached everything
    // but we only got the main entities returned.

    // Doublecheck 'cached' info. Note we know we're in the 'main' context.
    if (!isset($this->context['import']['linked_entities'])) {
      $this->setLinkedEntityInfo($this->context);
    }

    // Transform $data into the array we want:
    // array( main-entity-id => array( INDEX => array( linked-entity-ids )).
    // (We lose the connection with the source ID here.)
    $return = array();
    foreach ($data as $src_id => $main_entity_ids) {
      foreach ($main_entity_ids as $main_entity_id) {
        // Main entity should also exist (with empty value) if no linked
        // entities present.
        $return[$main_entity_id] = array();

        foreach ($this->context['import']['linked_entities'] as $i => $linked_info) {
          $linked_bin = "$linked_info[type]:$linked_info[bundle]";
          // We don't do negative caching so if there are no linked entities, the
          // linked-bin cache isn't set.
          if (isset($this->entityIds[$linked_bin][$src_id])) {
            $return[$main_entity_id][$i] = $this->entityIds[$linked_bin][$src_id];
          }
        }
      }
    }

    return $return;
  }

  /**
   * {@inheritdoc}
   *
   * Extra option supported:
   * query_type => the type of query (and/or data to fill in the cache). This is
   *   akin to ResolveEntityIdsQuery()'s 'return_type' but split into a separate
   *   option so we may still use a parent class' method (which has stringent
   *   checks on 'return_type's it does not support).
   *   Only one possible value: 'linked'.
   *
   * $this->context must be set before calling this method for 'linked'!
   */
  protected function executeResolveEntityIdsQuery(array $src_ids, $entity_type, array $options) {
    $data = array();

    // Can't think of a use case for a subclass that doesn't set a mapping,
    // but hey, let's keep it generic: check settings.
    $dest_field = isset($this->context['import_dst_pkey']) ? $this->context['import_dst_pkey']
            : (isset($this->settings['import_dst_pkey']) ? $this->settings['import_dst_pkey'] : '');
    // Generalized checks on query_type even though the caller realistically
    // only uses it to store 'linked'...
    $return_type = isset($options['query_type']) ? $options['query_type']
      : (isset($options['return_type']) ? $options['return_type'] : '');

    if ($return_type !== 'linked' && $dest_field) {
      // The parent can take care of this.
      if (isset($options['query_type'])) {
        $options['return_type'] = $options['query_type'];
      }
      return parent::executeResolveEntityIdsQuery($src_ids, $entity_type, $options);
    }

    if (empty($this->settings['import_ids_resolve_fast'])) {
      throw new \Exception("'Portable' method not supported in DrunkinsLinkedEntityImportJob::executeResolveEntityIdsQuery().");
    }

    // We must never cache only main-entity data, otherwise we risk
    // returning incomplete data a next time we're called with 'linked'.
    if ($return_type === 'linked') {

      if (!$entity_type) {
        $entity_type = $this->settings['import_entity_type'];
      }
      $main_entity_type = isset($this->origContextSettings)
        ? $this->origContextSettings['import_entity_type']
        : $this->settings['import_entity_type'];
      // This was also already checked in resolveEntityIds() (in cache items were
      // cached):
      if ($entity_type !== $main_entity_type) {
        throw new \Exception("'linked' lookups only supported for the 'main' entity type in DrunkinsLinkedEntityImportJob::resolveEntityIds().");
      }

      if (!$dest_field) {
        // @todo throw too. Here and everywhere.
        $this->log("resolveEntityIds was called while 'Source id field in destination entity' was not set.",
          array(), WATCHDOG_ERROR, FALSE);
      }
      else {

        $field_info_dest_field = field_info_field($dest_field);

        // Set up the entity-id plus 'source ID' field, which are both in the
        // main entity, and both in one table.
        if (isset($field_info_dest_field)) {
          // Resolve FieldAPI 'value' in field_data table.

          $dbfield_srcid = $dest_field . '_value';
          $dbfield_eid = 'entity_id';

          $query = db_select('field_data_' . $dest_field, 'm');
          $query->condition('entity_type', $entity_type);
          if (!empty($options['bundle'])) {
            $query->condition('bundle', $options['bundle']);
          }
        }
        else {
          // Resolve simple property in entity table.

          $dbfield_srcid = $dest_field;
          $info = entity_get_info($entity_type);
          $dbfield_eid = $info['entity keys']['id'];

          $query = db_select($info['base table'], 'm');
          if (!empty($options['bundle']) && !empty($info['entity keys']['bundle'])) {
            $query->condition($info['entity keys']['bundle'], $options['bundle']);
          }
        }

        if (empty($options['requery_all'])) {
          if (count($src_ids) == 1) {
            $query->condition($dbfield_srcid, reset($src_ids));
          }
          else {
            $query->condition($dbfield_srcid, $src_ids, 'IN');
          }
        }
        if (!empty($this->settings['import_entity_query_filter_field'])
            && !empty($this->settings['import_entity_query_filter_value'])) {

          if (isset($field_info_dest_field)) {
            // We assume this is a field in the entity table, so we should
            // join it.
            $info = entity_get_info($this->settings['import_entity_type']);
            $query->innerJoin($info['base table'], 'me', 'm.entity_id=me.' . $info['entity keys']['id']);
          }
          $query->condition(
            $this->settings['import_entity_query_filter_field'],
            $this->settings['import_entity_query_filter_value'],
            is_array($this->settings['import_entity_query_filter_value']) ? 'IN' : '='
          );
        }

        $query->fields('m', array($dbfield_srcid, $dbfield_eid));

        // Add linked-entity IDs. This can return multiple rows for the same
        // main-entity, and if we are linking multiple entities it can return
        // potentially really really many rows. We use this query only because
        // we expect there are very few linked-entities of the same type that
        // reference the same main-entity... if this assumption is wrong, the
        // query should be replaced with something else.

        // Doublecheck 'cached' info. Note we know we're in the 'main' context.
        if (!isset($this->context['import']['linked_entities'])) {
          $this->setLinkedEntityInfo($this->context);
        }
        $linked_entities = $this->context['import']['linked_entities'];

        foreach ($linked_entities as $i => $linked_info) {

          if ($this->context["import:$i"]['import_main_entity_ref_type']) {
            // Link is FieldAPI field.
            // ref_sqlfieldname is derived in setLinkedEntityInfo()
            $join_table = "field_data_$linked_info[ref]";
            $join_field = 'entity_id';
            $bundle_field = 'bundle';
            $field = $this->context["import:$i"]['import_main_entity_ref_sqlfieldname'];
            $join_condition = "m.$dbfield_eid = l$i.$field AND l$i.entity_type = :linkedtype$i";
            $join_args = array(":linkedtype$i" => $linked_info['type']);
          }
          else {
            // Link is simple property/db-field in entity table.
            $info = entity_get_info($linked_info['type']);
            $join_table = $info['base table'];
            $join_field = $info['entity keys']['id'];
            $bundle_field = $info['entity keys']['bundle'];
            $join_condition = "m.$dbfield_eid = l$i.$linked_info[ref]";
            $join_args = array();
          }
          if ($linked_info['bundle']) {
            $join_condition .= " AND l$i.$bundle_field = :linkedbundle$i";
            $join_args[":linkedbundle$i"] = $linked_info['bundle'];
          }
          if (!empty($this->settings["import_entity_query_filter_field:$i"])
              && !empty($this->settings["import_entity_query_filter_value:$i"])) {

            // Add another filter on a field in the entity table, restricting
            // returned rows from the left-joined (not from the main) table.
            // This is easy if the entity table is already joined - but if not,
            // we should probably use a subquery instead. Let's put that off
            // until whenever it's necessary.
            if ($this->context["import:$i"]['import_main_entity_ref_type']) {
              throw new \Exception(t("At this moment '@setting1 / @setting2' settings are only supported if @setting3 refers to a field in the linked @type's main table.",
                  array('@setting1' => "import_entity_query_filter_field:$i", '@setting2' => "import_entity_query_filter_value:$i", '@setting3' => $this->context["import:$i"]['import_main_entity_ref_type'], '@type' => $linked_info['type'])));
            }
            $join_condition .= " AND l$i." . $this->settings["import_entity_query_filter_field:$i"]
              . (is_array($this->settings["import_entity_query_filter_value:$i"])
                ? " IN(:linkedfilter$i)" : "=:linkedfilter$i");
            $join_args[":linkedfilter$i"] = $this->settings["import_entity_query_filter_value:$i"];
          }
          $query->leftJoin($join_table, "l$i", $join_condition, $join_args);
          $query->addField("l$i", $join_field, "lid$i");
        }

        // Get context.
        $result = $query->execute();

        if (isset($this->entityIds)) {
          // Append results directly to cache; construct return value later.
          $data =& $this->entityIds[$options['cache_bin']];

          foreach ($result as $row) {
            // Uppercase source IDs when necessary. Number to string conversion
            // does not matter here, because they're only used as array key.
            $src_id = $this->settings['import_ids_case_sensitive'] ?
              $row->$dbfield_srcid : strtoupper($row->$dbfield_srcid);

            // Cache main id and linked ids in separate bins.
            if (!isset($data[$src_id])) {
              $data[$src_id] = array();
            }
            // Dedupe; see comments at building SQL query above.
            if (!in_array($row->$dbfield_eid, $data[$src_id])) {
              $data[$src_id][] = $row->$dbfield_eid;
            }

            foreach ($linked_entities as $i => $linked_info) {
              // lid may not be set because it was not added as a field, or
              // because the contents is null. In both cases we don't cache - as
              // we don't do 'negative caching'.
              if (isset($row->{"lid$i"})) {
                $linked_bin = "$linked_info[type]:$linked_info[bundle]";
                if (!isset($this->entityIds[$linked_bin][$src_id])) {
                  $this->entityIds[$linked_bin][$src_id] = array();
                }
                if (!in_array($row->{"lid$i"}, $this->entityIds[$linked_bin][$src_id])) {
                  $this->entityIds[$linked_bin][$src_id][] = $row->{"lid$i"};
                }
              }
            }
          }
        }
        else {
          // Unlike when caching is on (above), we are compiling the return
          // value directly from the query:
          // array( main-entity-id => array( INDEX => array( linked-entity-ids))
          // several db rows can have the same 'linked' key.
          foreach ($result as $row) {
            // Main entity should also exist (with empty value) if no linked
            // entities present.
            if (!isset($data[$row->$dbfield_eid])) {
              $data[$row->$dbfield_eid] = array();
            }
            foreach ($linked_entities as $i => $linked_info) {

              if (isset($row->{"lid$i"})
                  && (!isset($data[$row->$dbfield_eid])
                      || !in_array($row->{"lid$i"}, $data[$row->$dbfield_eid][$i]))) {

                if (!isset($data[$row->$dbfield_eid])) {
                  $data[$row->$dbfield_eid][$i] = array();
                }
                $data[$row->$dbfield_eid][$i][] = $row->{"lid$i"};
              }
            }
          }
        }
      }
    }
    else {
      // We need to resolve linked-entity IDs without using import_dst_pkey;
      // either because the 'primary-in-current-context' entity does not have
      // src_id values (i.e. the import_dst_pkey field is in another entity
      // type) or because $entity_type != 'primary-in-current-context' type.

      // Add functionality here as time arises. (There is code in
      // DrunkinsCommerceProductImportJob.)
      throw new \Exception("resolveEntityIds was called for destination entity of type $entity_type, but no functionality has been defined for this.");
    }

    return $data;
  }

  /**
   * Update main + linked entities - or create new ones.
   *
   * @param object|null $main_entity
   *   product to update, or null to create new product.
   *   (Note that on return, $entity will be the valid updated entity which
   *   the caller can keep using, ONLY when a non-empty object was passed.)
   * @param array $linked_ids
   *   IDs of 'linked' entities to update, grouped per 'linked-type index'.
   *   An array of arrays, where
   *   - the outer keys must have a corresponding entity type + other necessary
   *     settings defined (must be integers >= 1)
   *   - the inner array is existing linked-entity id(s) to update. (There is no
   *     check if existing entities actually are linked to the main entity.)
   *     Empty array means one new linked entity will be created.
   *   If an outer key of 0 is present (value is ignored), then for any 'linked-
   *   type' which is defined in settings but not present in this array, a new
   *   entity will be created. If the 0 key is not present, then new entities
   *   will only be created for 'linked types' which are present in this array
   *   (with an empty value).
   *   This means: if you want to create new linked entities for every defined
   *   'linked-type', explicitly pass array(array()) as 2nd argument.
   *
   * @return mixed
   *   ID of created/updated entity. 0 means an existing entity was not updated;
   *   FALSE or NULL indicates error (though NULL generally shouldn't happen).
   */
  protected function updateEntity($main_entity = NULL, array $linked_ids = array()) {

    $main_entity_id = $return = parent::updateEntity($main_entity);

    // Handle linked entities, unless there was an error.
    // If an existing entity wasn't updated, still check/update/create linked.
    if ($main_entity_id === 0 && $main_entity) {
      $main_entity_id = $this->getEntityId($main_entity);
    }

    if (isset($linked_ids[0])) {
      // Mark 'need to create new entities' for each 'linked' type that has no
      // existing entity yet.
      $i = 0;
      while (!empty($this->settings['import_entity_type:' . ++$i])) {
        if (!isset($linked_ids[$i])) {
          $linked_ids[$i] = array();
        }
      }
      unset($linked_ids[0]);
    }

    if ($main_entity_id) {
      foreach ($linked_ids as $linked_index => $entity_ids) {

        $this->switchLinkedContext($linked_index);

        // Possible optimization: preloading entities per entity type instead
        // of per 'linked_type', by grouping them and calling loadEntities()
        // beforehand. Meh.
        $linked_entities = $entity_ids ? $this->loadEntities($entity_ids) : array();
        // If entities fail to load, we won't log an error; probably internal
        // code will. If no entities were loaded here, we create a new one, to
        // prevent situations where invalid data in e.g. the 'main-reference
        // field' table is never noticed & fixed.)
        if (!empty($linked_entities)) {

          // Update existing linked-entity/ies.
          if ($this->context['import_field_mapping']) {
            // If field mapping is set for them, then call updateEntity. The
            // mapping is responsible for updating status/title if necessary.
            foreach ($linked_entities as $linked_entity) {
              $updated_dest_id = parent::updateEntity($linked_entity);
              if ($updated_dest_id) {
                // Return a value indicating 'entity was updated' even if the
                // main entity was not (meaning if $return was still 0 here).
                $return = $main_entity_id;
              }
            }
          }
          // If field mapping is not set for linked entities, we assume they
          // hold nothing of interest. A subclass may need to extend this
          // situation to e.g. mirror a name/title or status value.
        }
        // If no existing linked-entities, create a new one.
        elseif (empty($this->settings['import_new_bundle'])
                || empty($this->settings['import_main_entity_ref'])) {
          // This can't happen really - we've already checked properties.
          $this->log('Not creating linked entity for @id, because required settings not present.', array('@id' => $main_entity_id), WATCHDOG_ERROR);
        }
        else {

          // Create entity and set reference link to main entity.
          $linked_entity = $this->initEntity($main_entity_id);

          // If mapping set, call updateEntity on this initialized entity. If
          // not, save it here.
          if (!$this->context['import_field_mapping']
              && !$this->settings['import_default_values']) {

            $this->saveEntity($linked_entity, array(LANGUAGE_NONE => TRUE));
            if ($this->getEntityId($linked_entity)) {
              // Return a value indicating 'entity was updated' even if the main
              // entity was not (meaning if $return was still 0 here).
              $return = $main_entity_id;
            }
          }
          else {
            $updated_dest_id = parent::updateEntity($linked_entity);
            if ($updated_dest_id) {
              $return = $main_entity_id;
            }
          }
        }
      }

      $this->switchLinkedContext(0);
    }

    return $return;
  }

  /**
   * Create new entity; either 'main' or 'linked'. When passing an argument, the
   * context must already be switched.
   *
   * @param int $main_id
   *   When creating a linked entity: the main ID. (This logic assumes that
   *   what we call the 'linked' entity is actually the 'referring' entity; it
   *   holds the reference.) When creating a 'main' entity: do not pass.
   *
   * @return object
   *   The entity, with basic properties initialized.
   *
   * @throws \Exception
   */
  protected function initEntity($main_id = 0) {

    if ($main_id && !isset($this->origContextSettings)) {
      throw new \Exception("initEntity() called with argument \"$main_id\" while context not switched!");
    }

    $entity = parent::initEntity();

    if ($main_id) {

      if (!isset($this->settings['import_main_entity_ref'])
          || !isset($this->context['import_main_entity_ref_type'])) {
        throw new \Exception("initEntity() called while 'import_main_entity_ref(_type)' not set");
      }

      // This logic flow assumes a 'main' entity was created just beforehand.
      // Set reference field/property.
      if ($this->context['import_main_entity_ref_type'] == 2) {
        // Multi-value fieldAPI field. (This hack is necessary for at least
        // commerce_product_reference; don't know about other ones?)
        $main_id = array($main_id);
      }
      // This is the only property that we create an entity wrapper for; the
      // 'simple' properties above are just easier to do without one.
      $wrapper = entity_metadata_wrapper($this->settings['import_entity_type'], $entity);
      $wrapper->{$this->settings['import_main_entity_ref']}->set($main_id);
    }

    return $entity;
  }

  /**
   * Convenience method for switching context/settings of this class while
   * running. Putting all the switching in one method makes sure we don't
   * forget anything / end up with unforeseen inconsistencies.
   *
   * Calling this method is potentially very dangerous; it can mess up the
   * context and with that all the actions done by / output from a job!
   *
   * @param int $linked_index
   *   Index of the linked entity type+bundle / context to switch to. 0 to
   *   switch back to original context.
   *
   * @throws \Exception
   */
  protected function switchLinkedContext($linked_index) {

    if (isset($this->origContextSettings)) {

      $this->settings = $this->origContextSettings;
      unset($this->origContextSettings);

      if (isset($this->context)) {
        $this->context =& $this->origContextResults;
        unset($this->origContextResults);
      }
    }

    // Now switch to linked context if set.
    if ($linked_index) {

      if (!is_int($linked_index) || empty($this->settings["import_entity_type:$linked_index"])) {
        throw new \Exception("unknown argument \"$linked_index\" to switchLinkedContext!");
      }
      if (isset($this->context) && empty($this->context["import:$linked_index"])) {
        $this->log("import:$linked_index part of context not set in switchLinkedContext!");
      }

      $this->origContextSettings = $this->settings;

      // Overwrite some settings and class properties with 'linked equivalents'.
      $this->settings['import_entity_type'] = $this->settings["import_entity_type:$linked_index"];
      $this->settings['import_main_entity_ref'] = $this->settings["import_main_entity_ref:$linked_index"];
      // Unset or set to default, depending on what constructor does:
      $this->settings['import_default_values'] = empty($this->settings["import_default_values:$linked_index"])
        ? array() : $this->settings["import_default_values:$linked_index"];
      if (empty($this->settings["import_new_bundle:$linked_index"])) {
        unset($this->settings['import_new_bundle']);
      }
      else {
        $this->settings['import_new_bundle'] = $this->settings["import_new_bundle:$linked_index"];
      }
      if (empty($this->settings["opt_cleanup:$linked_index"])) {
        unset($this->settings['opt_cleanup']);
      }
      else {
        $this->settings['opt_cleanup'] = $this->settings["opt_cleanup:$linked_index"];
      }
      // pkey / changed fields info is set in context; if we have it and
      // the actual setting is not set, get it. (Because we can't get it from
      // context after switching that.)
      if (empty($this->settings['pkey']) && !empty($this->context['import_src_pkey'])) {
        $this->settings['pkey'] =  $this->context['import_src_pkey'];
      }
      // 'dst' properties by definition refer to the primary entity, so reset.
      // (The mirroring of a key value and/or changed date into linked entities
      // can be made to work, when providing 'import_src_*' settings and a
      // mapping from that field for each entity. This is why we don't reset the
      // import_skip_unchanged setting.)
      unset($this->settings['import_dst_pkey']);
      unset($this->settings['import_dst_changed']);
      // Copy the 'opt_cleanup' setting even though no default functionality
      // uses it (see comments at top of this class)
      if (isset($this->settings["opt_cleanup:$linked_index"])) {
        $this->settings['opt_cleanup'] = $this->settings["opt_cleanup:$linked_index"];
      }
      else {
        unset($this->settings['opt_cleanup']);
      }

      // Point $this->settings to a separate storage, so we can change/populate
      // context for linked. In some cases (setMappingByDest, and possibly on
      // finish?) this won't be set - which will be inconvenient for the caller,
      // but we won't warn.
      if (isset($this->context)) {
        $this->origContextResults =& $this->context;
        $this->context =& $this->origContextResults["import:$linked_index"];
        // Copy anything important we know that is used in standard fetcher, to
        // the 'linked' part of the context. Note that unlike $this->settings,
        // the copied values will linger in $context after we exit this method!
        //
        // If your custom extended/fetcher class uses other context values that
        // influence e.g. the return value of hasFulleDataSet(), they should
        // be copied here too.
        // 'fetcher_timestamp_from' may be needed by hasFulleDataSet().
        if (isset($this->origContextResults['fetcher_timestamp_from'])) {
          $this->context['fetcher_timestamp_from'] = $this->origContextResults['fetcher_timestamp_from'];
        }
      }
    }
  }

}
