<?php
/**
 * @file
 * Contains SqlTableImportJob.
 */

//@todo use an Exception type with args/tokens/translation context, in various places.

/**
 * Class DrunkinsEntityImportJob
 *
 * A class suitable for importing fetched items into SQL database tables.
 *
 * About naming: we've grown accustomed to naming things to be written into
 * Drupal "entities", to distinguish them from the "items" fetched from whatever
 * source system. This means that in this class, we'll often call a plain
 * database record in Drupal an "entity".
 */
class DrunkinsSqlTableImportJob extends DrunkinsJob {

  /* Settings in $this->settings, used by this class:
   * - job_id:              defined in parent
   * - pkey:                defined in fetcher
   * - fetcher_limit:       defined in fetcher
   * - fetcher_single_item: defined in fetcher
   *
   * - import_entity_type (string):
   *   Name of the local table to import data into.
   * - import_field_mapping (array):
   *   Mapping of source field to destination property, keyed by destination.
   *   There are lots of optional properties; see setMappingByDest().
   *   ($context will contain a 'normalized' version.)
   * - import_default_values (array):
   *   default values for destination properties of the entity which are either
   *   not in the mapping, or whose source values are empty.
   *
   * - import_dst_pkey (string):
   *   The entity field containing the unique ID (pkey) value of a source item.
   *   This value must also be saved in the destination entity, in order to do
   *   lookups.
   *   Setting only one of 'import_dst_pkey' and 'pkey' is allowed, if the other
   *   can be derived through the field mapping.
   *
   * - import_entity_query_filter_field (string):
   * - import_entity_query_filter_value:
   *   database fieldname and value forming a filter (WHERE clause) that
   *   restricts the entity (database) query when resolving a source item ID.
   *
   * - opt_cleanup (string/array):
   *   Enables 'import_cleanup' setting in settingsForm, and also describes the
   *   exact cleanup action(s) to take. See processEntities() for possible
   *   values.
   *   If set to a string like "delete", this equals array("delete"). Boolean
   *   TRUE may or may not be translated to a default action, depending on the
   *   implementing class. (This class does not have a default.)
   * - import_cleanup (boolean):
   *   If TRUE/nonempty, clean up all 'entities' in Drupal which are not among
   *   the fetched items upon finish() - but only when we process a complete
   *   data set.
   * - import_cleanup_max_chunk_size (integer, default may vary per class):
   *   The maximum number of IDs to process in one chunk. This internal setting
   *   may or may not be used by various classes to decide whether or not to
   *   optimize processing. (Don't worry about this setting unless you know you
   *   want to.)
   * - import_alt_cleanup (boolean, default FALSE):
   *   Alternative way of counting which items need to be cleaned up. This is
   *   EXPERIMENTAL, untested code which probably does not yield anything; I'm
   *   just too proud to delete it right now. It's not 100% tested (cleanup is
   *   not tested) and may be removed sometime later. Not supported by child
   *   classes.
   * - import_full_data_set_default (boolean):
   *   On finish(), the fetcher's hasFulleDataSet() method is called to check
   *   if items can be cleaned up. The default value (if there is no separate
   *   fetcher class) is FALSE, unless you set this setting to a true value.
   *   (This is because if you are not using a fetcher but extending an import
   *   class with your own start() method, the standard code cannot know whether
   *   the data set is complete so it will err on the side of safety, and not do
   *   cleanup.)
   *
   * - import_precache_entity_ids (boolean, but that may change):
   *   Load a mapping for all existing source->entity IDs into context. This may
   *   create a much larger context blob. At least one large import was slowing
   *   down exponentially toward the end of the process (for reasons unknown),
   *   if this was not set.
   *   The default behavior is to load this mapping if the fetcher's
   *   'hasFullDataSet()' method returns TRUE.
   *   @todo it may also make sense to _not_ cache any entity IDs during the
   *     process, if this is turned off - or, depending on another setting or a
   *     separate value for this setting. (In other words, the behavior
   *     "$this->entityIds =& $context['import_ids'];" should be configurable.)
   *     On the other hand, processEntities(Action)() needs it so it's probably
   *     best to make sure things do get cached if we do cleanup.
   *
   * - import_trim_values (boolean, default TRUE):
   *   If TRUE, all string values from the source data are trim()ed. This is
   *   recommended to prevent obscure bugs.
   *
   * - import_ids_case_sensitive (boolean):
   *   This 'internal setting' must be set to TRUE if Drupal queries its
   *   storage for the source IDs in a case sensitive manner, which is the case
   *   for PostgreSQL on D7.
   *   That means it will matter (and by default it doesn't) if:
   *   - source-ID values stored in Drupal have different case than those in
   *     the source system, or if case gets changed in the source system.(*)
   *   - ID values used in 'reference fields' in the source system have
   *     different case than the original source IDs.
   *   Note this setting is expected to follow your database fields' behavior,
   *   not determine the behavior; there is only one valid value for each
   *   situation. Specifically:
   *   - if you have MySQL (or another case insensitive database) but this
   *     setting is TRUE, resolveEntityIds() will still match IDs with different
   *     case in its database queries. However it may fail to match IDs in the
   *     'entityIds' cache, which may lead to creating duplicate items or fatal
   *     errors (while trying to insert those items with a different-case ID).
   *   - if your database fields are case insensitive, the queries in
   *     resolveEntityIds() will not match a different case even if this setting
   *     is TRUE. This means a new entity will be created. However,
   *     resolveEntityIds() (and other code that behaves according to spec) will
   *     uppercase the 'entityIds' cache, which may lead to strange behavior -
   *     like writing reference fields into, or cleaning up, the wrong entity.
   */

  // Class-wide variables which are filled once per processItem() execution
  // and used by its various called functions to make the code easier.
  //
  // Reference to $context as passed to the public functions. Use wisely in your
  // own code!
  // - if this variable is written to before it is set-by-reference, things will
  //   get lost.
  // - if $this->context['import_ids'] is written to / read, note that the
  //   source IDs used as subkeys of this array must be uppercase (unless
  //   $this->settings['import_ids_case_sensitive'] is TRUE)! See above.
  protected $context;
  // Reference to $context['import_ids']. If set, this enables
  // caching by resolveEntityIds().
  protected $entityIds;
  // The item from the source system (an array of sourcefield => value pairs)
  // which is currently being processed (fields written into an entity).
  protected $src_item;

  // Class-wide variables used as 'cache' for other purposes.
  //
  // Destination (key in mapping) currently being processed. Set repeatedly in
  // updateEntity(); meant to be used by e.g. buildItemValues() to look up
  // 'process' functions from the right destination mapping.
  protected $current_dst;
  //
  // Variable used by isProcessingDeferred() & setApiFieldValue() to keep state:
  protected $defer;

  /**
   * {@inheritdoc}
   * Also see comments at the top of this class for its own settings.
   */
  public function __construct(array $settings = array()) {

    if (!isset($settings['import_trim_values'])) {
      $settings['import_trim_values'] = TRUE;
    }
    if (!isset($settings['import_ids_case_sensitive'])) {
      $settings['import_ids_case_sensitive'] = FALSE;
    }
    // Make sure some variables are arrays so we can use them as such.
    if (!isset($settings['import_default_values'])) {
      $settings['import_default_values'] = array();
    }

    parent::__construct($settings);

    // In case a child class sets 'opt_list', we filter away arguments which
    // are useless in the 'display' URL.
    $this->display_arguments_filter = array_merge($this->display_arguments_filter, array(
      'import_entity_type',
      'import_trim_values',
      'import_dst_pkey',
      'import_field_mapping',
      'import_default_values',
      'import_ids_case_sensitive',
    ));
  }

  /// Extensions of non-interface methods from parent classes

  /**
   * {@inheritdoc}
   */
  public function settingsFormSubmit($op) {
    if ($op == t('Check field mapping')) {

      // Get all fields (to use their field names).

      // Make sure we also get empty fields.
      // @todo this is AFAS specific. In whatever way, move this out.
      $this->settings['afas_include_empty_fields'] = TRUE;

      // Get data. We only need one row but unfortunately we have no generic way
      // to specify that the source system should send one row, so get all data.
      // Provide a hint in 'context' that we are checking the mapping.
      $context = array('import_op' => $op);
      $items = $this->start($context);
      if (is_string($items)) {
        // Return error message literally
        $output = '<h3>' . $items . '</h3>';
      }
      elseif (empty($items)) {
        $output = '<h3>' . t('There are no items to process.') . '</h3>';
      }
      else {
        // Get source fields (names); create simple source -> destination mapping.
        $map_src = array();
        foreach ($context['import_field_mapping'] as $dst => $dest_map) {
          foreach ($dest_map['src'] as $src) {
            if (!isset($map_src[$src])) {
              $map_src[$src] = array();
            }
            $map_src[$src][] = $dst;
          }
        }

        $output = '';

        $missing_afas = array_diff_key($map_src, $items[0]);
        if ($missing_afas) {
          $output .= '<h3>' . t('The following fields have a mapping defined, but are missing in the source data:') . '</h3>';
          $rows = array();
          foreach ($missing_afas as $field => $dest_fields) {
            $rows[] = array($field, implode(', ', $dest_fields));
          }
          $output .= theme('table', array('header' => array('Source field', 'Destination field(s)'), 'rows' => $rows));
        }

        $missing_mapping = array_diff_key($items[0], $map_src);
        if ($missing_mapping) {
          $output .= '<h3>' . t('The following fields in the source data are not defined in a field mapping:') . '</h3>';
          $rows = array();
          foreach (array_keys($missing_mapping) as $field) {
            $rows[] = array($field);
          }
          $output .= theme('table', array('header' => array('Source field'), 'rows' => $rows));
        }

        $both = array_intersect_key($map_src, $items[0]);
        if ($both) {
          if (!$missing_afas && !$missing_mapping) {
            $output .= '<h3>' . t('All source data are correctly defined in a field mapping:') . '</h3>';
          }
          else {
            $output .= '<h3>' . t('Fields correctly defined in the mapping:') . '</h3>';
          }
          $rows = array();
          foreach ($both as $field => $dest_fields) {
            $rows[] = array($field, implode(', ', $dest_fields));
          }
          $output .= theme('table', array('header' => array('Source field', 'Destination field(s)'), 'rows' => $rows));
        }
      }

      return $output;
    }
    else {
      return parent::settingsFormSubmit($op);
    }
  }

  /**
   * Check settings; log anything not OK.
   *
   * @param array $context
   *   Context as passed to start() / processItem().
   * @param bool $quiet_fast
   *   If TRUE, do not log, and exit as soon as something not OK is found.
   *   It can be assumed that whenever this method is called with $quiet_fast
   *   TRUE, it has already been called without this argument for the same
   *   process, and there is no need to do a full accounting anymore.
   *
   * @return bool
   *   TRUE if all OK.
   */
  protected function checkSettings(array $context, $quiet_fast = FALSE) {
    $ok = TRUE;
    foreach (array('entity_type', 'field_mapping') as $setting) {
      if (empty($this->settings["import_$setting"])) {
        // 'shorthand' to set $ok to FALSE when it matters:
        if ($ok = $quiet_fast) {
          return FALSE;
        }
        $this->log('@setting setting not configured!', array('@setting' => "import_$setting"), WATCHDOG_ERROR);
      }
    }
    if (!empty($this->settings['import_entity_type'])) {
      // Until now, we only support tables with one primary key.
      if (!$this->getEntityPrimaryKey()) {
        if ($ok = $quiet_fast) {
          return FALSE;
        }
        $this->log('@table does not have a single primary key defined in the database schema.', array('@table' => $this->settings['import_entity_type']), WATCHDOG_ERROR);
      }
    }

    if (isset($this->settings['import_default_values']) && !is_array($this->settings['import_default_values'])) {
      if ($ok = $quiet_fast) {
        return FALSE;
      }
      $this->log('@setting setting is not an array!', array('@setting' => "import_default_values"), WATCHDOG_ERROR);
    }

    if (!empty($this->settings['import_entity_query_filter_field'])
        xor !empty($this->settings['import_entity_query_filter_value'])) {
      if ($ok = $quiet_fast) {
        return FALSE;
      }
      $this->log("If one of '@setting1' and '@setting2' settings are set, they must both be set.",
        array('@setting1' => 'import_entity_query_filter_field', '@setting2' => 'import_entity_query_filter_value'), WATCHDOG_ERROR);
    }

    // These should all be present after setMappingByDest has run.
    foreach (array('src_pkey', 'dst_pkey') as $setting) {
      if (empty($context["import_$setting"])) {
        if ($ok = $quiet_fast) {
          return FALSE;
        }
        $this->log('@setting context value not found / cannot be derived from other settings!', array('@setting' => "import_$setting"), WATCHDOG_ERROR);
      }
    }

    if (!is_array($this->settings['import_field_mapping'])) {
      if ($ok = $quiet_fast) {
        return FALSE;
      }
      $this->log('@setting setting is not an array!', array('@setting' => 'import_field_mapping'), WATCHDOG_ERROR);
    }

    return $ok;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm() {
    $form = parent::settingsForm();

    $context = array();

    if (isset($this->settings['import_field_mapping'])) {
      $this->setMappingByDest($context);

      if (!empty($this->settings['opt_list']) && user_access('access drunkins advanced')) {
        $form['display']['mapping'] = array(
          '#type'  => 'submit',
          '#value' => t('Check field mapping'),
        );
      }
    }

    if (!empty($this->settings['opt_cleanup'])) {
      // Guess the best description.
      $cleanup_desc = 'Clean up';
      $cleanedup_desc = 'cleaned up';
      if (is_array($this->settings['opt_cleanup']) ? in_array('delete', $this->settings['opt_cleanup'])
        : $this->settings['opt_cleanup'] === 'delete') {
        $cleanup_desc = 'Delete';
        $cleanedup_desc = 'deleted';
      }
      elseif ($this->settings['opt_cleanup'] === array('status' => 0)) {
        $cleanedup_desc = isset($this->settings['import_entity_type'])
                          && $this->settings['import_entity_type'] === 'node' ? 'unpublished' : 'deactivated';
      }
      $form['behavior']['import_cleanup'] = array(
        '#type'          => 'checkbox',
        '#title'         => t('!cleanup items not encountered in data set', array('!cleanup' => t($cleanup_desc))),
        '#description'   => t('If checked, all items present in the website but not encountered in the source data set, will be !cleaned_up - but <em>only</em> if you get a full, un-cached data set, not constrained by any of the below options. If the data set is incomplete/filtered in some other way than the below, you MUST uncheck this.',
          array('!cleaned_up' => t($cleanedup_desc))),
        '#default_value' => 1,
        '#weight'        => 5,
      );
    }

    // Since this form is the one interactive thing in the process, it's also a
    // good place for global checks that should generate user errors.
    $this->checkSettings($context);

    return $form;
  }

  /// Interface methods

  /**
   * {@inheritdoc}
   */
  public function start(array &$context) {

    // If mapping was not set yet in context, normalize and set it.
    // In the rest of the process, we will use the following properties from
    // the context, not look at settings: import_field_mapping,
    // import_src_changed, import_dst_changed. The import_src_pkey and
    // import_dst_pkey properties are a bit trickier: settings _may_ be used by
    // methods that could be called from subclasses that don't use a mapping.
    if (!isset($context['import_field_mapping']) && !empty($this->settings['import_field_mapping'])) {
      $this->setMappingByDest($context);
    }

    // We usually don't change settings, but if 'pkey' could be derived from
    // 'import_dst_pkey', it's good to pass it to the parent/fetcher start().
    if (empty($this->settings['pkey']) && !empty($context['import_src_pkey'])) {
      $this->settings['pkey'] = $context['import_src_pkey'];
    }

    // This is the last place where we can do checks on strangenesses that will
    // make processItem() not do anything.
    // (Better give errors once here than <items> times later.)
    $error = !$this->checkSettings($context);

    // Some actions still need items fetched even when there are errors (because
    // these only serve as warnings).
    // @todo Display and Check Mapping have different ways of specifying their actions - unify?
    if ($error && empty($this->settings['list_format'])
        && !(isset($context['import_op']) && $context['import_op'] == t('Check field mapping'))) {
      throw new \Exception("Improper job configuration prevents it from running. Errors have been logged.");
    }

    $items = parent::start($context);

    // Add the keys we're using to the context array, if we don't want to
    // have to perform isset() checks every time in processItem(). (This is not
    // true for the actions / actions_skipped / import_processed keys.)
    $context = array_merge($context,
      array(
        'new'        => array(),
        'updated'    => array(),
        'skipped'    => array(),
        'error'      => array(),
        'multiple'   => 0,
        'import_ids' => array(),
        'import_deferred'   => array(),
        'import_old_values' => array(),
      )
    );

    $precache = isset($this->settings['import_precache_entity_ids']) ? $this->settings['import_precache_entity_ids']
      : $this->callFetcherMethod('hasFullDataSet', FALSE, $context);
    if ($precache) {
      // resolveEntityIds()' code needs this:
      $this->context =& $context;
      // Prefill cache.
      $this->entityIds =& $context['import_ids'];
      $tmp_cache = $this->resolveEntityIds(array(), $this->settings['import_entity_type'], array('requery_all' => TRUE));
      // It probably makes no difference, but let's not leave references to
      // random large data structures floating around anyway.
      unset($this->context);
      unset($this->entityIds);
    }
    if (!empty($this->settings['import_alt_cleanup'])) {
      $context['skipped'] = 0;
      // It's kinda too bad that we need two nearly identical caches at the
      // same time, but resolveEntityIds one has a different structure and I
      // don't want to rewrite it...
      if (!empty($this->settings['import_precache_entity_ids'])) {
        $context['entity_ids_not_encountered'] = db_select($this->settings['import_entity_type'], 'e')
          ->fields('e', array($this->getEntityPrimaryKey()))
          ->execute()
          ->fetchCol();
      }
      else {
        $context['entity_ids_not_encountered'] = $tmp_cache;
        // So the whole array won't get cloned in memory by the next line:
        unset($tmp_cache);
      }
      $context['entity_ids_not_encountered'] = array_combine($context['entity_ids_not_encountered'], $context['entity_ids_not_encountered']);
    }

    return $items;
  }

  /**
   * {@inheritdoc}
   */
  public function processItem($item, array &$context) {

    // TODO check on empty title / possibly other required fields, and warn.
    // The process does not need to be changed to abort, but possibly the user wants to know.

    // TODO or instead, write generic check+warn on required fields.

    if (!$this->checkSettings($context, TRUE)) {
      // We can't do anything. We've warned in start(), so exit silently.
      return;
    }

    // Fill reference to context so called/extending code can access it.
    // (If this function is called more than once on an already-instantiated
    // class, it's a no-op.)
    $this->context =& $context;
    // Turn on resolveEntityIds() caching.
    // @todo make setting for it; it's not necessary for simple imports.
    $this->entityIds =& $context['import_ids'];

    // If a special '(cleanup) action' key is the only thing present besides
    // an ID, we want to perform the mentioned actions instead of the regular
    // import.
    if (isset($item['*action']) && count($item) == 2) {

      // Check what kind of ID we have to operate on.
      if (isset($item['*entity_id'])) {
        // We have no source ID; destination entity already resolved.
        $dest_ids = is_array($item['*entity_id']) ? $item['*entity_id'] : array($item['*entity_id']);
      }
      else {
        // Resolve destination from source. If the second key in $item is not
        // the source ID field, the item will be skipped.
        $src_id = isset($item[$context['import_src_pkey']])
          ? trim($item[$context['import_src_pkey']]) : '';
        $dest_ids = $src_id === '' ? array() : $this->resolveEntityIds(array($src_id));
      }

      // Construct key to account for action in the finish() message. It
      // should be converted from array('action') to 'action' if possible.
      $action_key = $item['*action'];
      if (is_array($item['*action'])) {
        if (count($item['*action']) == 1) {
          $action_key = reset($item['*action']);
          if (!is_numeric(key($item['*action']))) {
            $action_key = $item['*action'];
          }
        }
        if (is_array($action_key)) {
          // We need some string representation.
          $action_key = json_encode($action_key);
        }
      }

      // If no mapped entity exists, skip the action silently. (This is good
      // for 'cleanup'.)
      if ($dest_ids) {

        $this->processEntities(
          is_string($item['*action']) ? array($item['*action']) : $item['*action'],
          $dest_ids
        );

        // Accounting is done in the same way, regardless whether this is an
        // item with a source ID or only an entity ID. In the former case, we
        // add only 1 even if multiple items were cleaned up. There's no real
        // thought behind this yet; it may need to be refined later.
        if (!isset($context['actions'][$action_key])) {
          $context['actions'][$action_key] = 0;
        }
        $context['actions'][$action_key]++;
      }
      else {
        if (!isset($context['actions_skipped'][$action_key])) {
          $context['actions_skipped'][$action_key] = 0;
        }
        $context['actions_skipped'][$action_key]++;
      }
    }

    else {
      // Process 'regular item':

      if (!empty($this->settings['import_trim_values'])) {
        // Trim all values, to prevent hard to trace bugs caused by leading /
        // trailing spaces.
        array_walk_recursive($item, function (&$value) {
          if (is_string($value)) {
            $value = trim($value);
          }
        });
      }

      $src_id = $item[$context['import_src_pkey']];
      // TODO consider whether it's worth putting the $src_id into
      //   $context['exceptions'] and undo-ing that when filling one of the
      //   other stats below. That way we know which things generated exceptions
      //   and can do proper accounting / can ignore the item in cleanup if we
      //   include 'exceptions' in processedEntityIds(). (This is optional
      //   because the standard behavior of a runner is to never run finish() if
      //   an exception occurred. But it's a nice-to-have when a config setting
      //   is changed that makes the job continue on exceptions?)

      $this->src_item = $item;
      // The source item's key must have a value since it's the unique ID. Any
      // method called after here can assume it exists.
      if ($this->isItemFieldEmpty($context['import_src_pkey'])) {
        $this->log("Item %nr has no 'Source id' value; will not be processed!", array('%nr' => $context['drunkins_progress']), WATCHDOG_WARNING);
        return;
      }

      $dest_ids = $this->resolveEntityIds(array($src_id));
      if (!$dest_ids) {
        // Create new entity.

        $updated_dest_id = $this->updateEntity();

        if ($updated_dest_id) {
          $context['new'][$updated_dest_id] = $src_id;
          // It is highly unlikely that $updated_dest_id would already be
          // present as a destination entity, but still the following is fine:
          unset($context['entity_ids_not_encountered'][$updated_dest_id]);
        }
        elseif ($updated_dest_id === 0) {
          if (!empty($this->settings['import_alt_cleanup'])) {
            $context['skipped']++;
          }
          else {
            $context['skipped'][] = $src_id;
          }
        }
        else {
          $context['error'][] = $src_id;
        }
      }
      else {
        // Update all entities which are 'mapped' to this source item.

        $entities = $this->loadEntities($dest_ids);
        if ($entities) {
          $updated_dest_id = 0;
          foreach ($entities as $entity) {

            if (!empty($this->settings['import_alt_cleanup'])) {
              unset($context['entity_ids_not_encountered'][$this->getEntityId($entity)]);
            }

            $updated_dest_id = $this->updateEntity($entity);

            if ($updated_dest_id) {
              $context['updated'][$updated_dest_id] = $src_id;
              // Again highly unlikely that $updated_dest_id would already be
              // present as a destination entity AND that the entity ID changed
              // (though the latter does happen in exotic cases)...
              unset($context['entity_ids_not_encountered'][$updated_dest_id]);
            }
            elseif ($updated_dest_id !== 0) {
              $context['error'][] = $src_id;
            }
          }
          // For understandable statistics, report if a source item has more
          // than one entry (otherwise the numbers don't add up, without
          // explanation). We don't report which items had how many entities.
          if (count($entities) > 1) {
            $context['multiple']++;
          }
          elseif ($updated_dest_id === 0) {
            if (!empty($this->settings['import_alt_cleanup'])) {
              $context['skipped']++;
            }
            else {
              $context['skipped'][] = $src_id;
            }
          }
        }
        else {
          // No loaded entities; should not happen.
          $context['error'][] = $src_id;
        }
      }
    }

    // It probably makes no difference, but let's not leave references to random
    // large data structures floating around anyway.
    unset($this->context);
    unset($this->entityIds);
  }

  /**
   * Implements DrunkinsJobInterface::finish
   *
   * TODO: instead of doing lots of entity_saves here, queue those actions again!
   */
  public function finish(array &$context) {

    // Enable resolveEntityIds() cache.
    // @todo make setting for it; it's not necessary for simple imports.
    $this->entityIds =& $context['import_ids'];

    $this->updateDeferred($context);

    // Call parent (fetcher) method. (It's a judgment call if it's better doing
    // it before 'cleanup'; this way, the fetcher can do some more processing
    // before cleanup - but we hope it never needs a _cleaned_ data set or
    // statistics on what is/will be cleaned up.)
    $message = parent::finish($context);
    if ($message) {
      // It's very unlikely we'll get a message, but account for it.
      // Just log it separately, until we change our message handling.
      $this->log($message, array(), WATCHDOG_NOTICE);
    }

    // Clean up 'entities' which were not encountered in the data set.
    // (This will not happen by default if no separate fetcher class is used.)
    $default = !empty($this->settings['import_full_data_set_default']);
    $cleanup = !empty($this->settings['import_cleanup'])
               && ($context['updated'] || $context['new'] || $context['skipped'])
               && !$context['error']
               && $this->callFetcherMethod('hasFullDataSet', $default, $context);
    if ($cleanup) {
      // Validate opt_cleanup option.
      if (empty($this->settings['opt_cleanup'])) {
        $this->log("Setting '@setting1' is set while '@setting2' is not! Ignoring.",
          array('@setting1' => 'import_cleanup', '@setting' => 'opt_cleanup'),
          WATCHDOG_WARNING);
      }
      else {
        $this->context =& $context;
        $this->processEntities(array('cleanup'));
        unset($this->context);
      }
    }

    // It probably makes no difference, but let's not leave references to random
    // large data structures floating around anyway.
    unset($this->entityIds);

    // New/updated/skipped should be set by processItem() calls.

    if ($context['new']) {
      $message = format_plural(count($context['new']), '1 new item', '@count new items');
      if ($context['updated']) {
        $message .= ', ' . count($context['updated']) . ' updated';
      }
    }
    else {
      // always 'updated', also if 0
      $message = format_plural(count($context['updated']), '1 item updated', '@count items updated');
    }
    if ($context['skipped']) {
      $message .= ', ' . (empty($this->settings['import_alt_cleanup']) ? count($context['skipped']) : $context['skipped']) . ' skipped';
    }
    if ($context['error']) {
      $message .= ', ' . format_plural(count($context['error']), '1 error', '@count errors');
    }
    if ($context['multiple']) {
      $message .= '; ' . format_plural($context['multiple'], '1 item', '@count items') . ' have multiple corresponding entities (which is why the \'updated\' count is higher)';
    }

    // Add 'actions' as normal messages.
    // Depending on the job, these items may have been in the original data set
    // and/ or passed back from finish() in a 'restarted' job - so, depending on
    // the job they may not always add up. Let's revisit later if needed.
    if (isset($context['actions'])) {
      foreach ($context['actions'] as $action => $amount) {
        $message .= "; '$action' action performed on " . format_plural($amount, '1 item', '@count items');
      }
    }
    if (isset($context['actions_skipped'])) {
      foreach ($context['actions_skipped'] as $action => $amount) {
        $message .= "; '$action' action skipped for " . format_plural($amount, '1 nonexisting entity', '@count nonexisting entities');
      }
    }
    $message .= '.';

    // Add 'cleanup' message - the populating of this needs to be explicitly
    // specified by specifying 'collect_ids' as a cleanup action. It may be
    // duplicate functionality with 'actions' messages now but we'll look at
    // that later.
    if (!empty($context['import_processed'])) {
      // Guess the best description.
      $cleanedup_desc = 'cleaned up';
      if (is_array($this->settings['opt_cleanup']) ? in_array('delete', $this->settings['opt_cleanup'])
        : $this->settings['opt_cleanup'] === 'delete') {
        $cleanedup_desc = 'deleted';
      }
      elseif ($this->settings['opt_cleanup'] === array('status' => 0)) {
        $cleanedup_desc = isset($this->settings['import_entity_type'])
                          && $this->settings['import_entity_type'] === 'node' ? 'unpublished' : 'deactivated';
      }
      $message .= '. ' . format_plural(count($context['import_processed']), '1 existing item !cleaned_up which was not in the import data'
        , '@count existing items !cleaned_up which were not in the import data', array('!cleaned_up' => $cleanedup_desc));
    }

    return $message;
  }

  /// New methods introduced by this class.

  /**
   * Perform a number of actions on an entity.
   *
   * This is not the main processing part of this import class. It doesn't map /
   * process / import values from the source item. This is a 'secondary' line
   * of processing which is primarily used for cleanup of outdated entities that
   * are not present in the import - but it's set up in a more general way so
   * that 'cleanup' can be defined as any action, actions can be added by
   * subclasses. The method could be useful for other things besides 'cleanup'.
   *
   * There is no return value (e.g. for indicating 'success'). (At this moment
   * it's not clear whether that will be needed, or whether an action that needs
   * it should consider 'regular' processItem() flow instead.)
   *
   * @param array $actions
   *   Actions to perform. array('cleanup') is a special value that expands to
   *   whatever actions are configured for cleanup (the 'opt_cleanup' setting
   *   value). Actions are processed sequentially: first the first action on all
   *   entities, then the second, etc. Except actions that are 'set a property's
   *   value': these are grouped together and done at once where possible.
   *   Array key/values can be:
   *   - alphanumeric-key => value: the key is an entity's property, and the
   *     value is the new value to set.
   *   - string values (with standard numeric key): see the definition of
   *     cleanupEntitiesAction(). This class supports "collect_ids" & "delete".
   *     "collect_ids", if specified, should come before any other value or
   *     otherwise statistics may be incomplete.
   * @param array $entity_ids
   *   (optional) IDs of entities to process. This argument should be passed
   *   except for 'after-process cleanup': if $actions is array('cleanup') and
   *   this argument is an empty array, this method will gather a list of
   *   entities 'to be cleaned up after a process has finished'. In other cases,
   *   an empty array will do nothing.
   */
  protected function processEntities(array $actions, array $entity_ids = array()) {
    if (!$actions) {
      $this->log('Empty action passed to processEnities().', array(), WATCHDOG_ERROR);
      return;
    }

    // Expand the 'alias' "cleanup" to the configured setting.
    if ($actions == array('cleanup')) {
      if (empty($this->settings['opt_cleanup'])) {
        $this->log("Empty '@option' setting value; ignoring cleanup.", array('@option' => 'opt_cleanup'), WATCHDOG_WARNING, FALSE);
        return;
      }
      $actions = is_string($this->settings['opt_cleanup']) ?
        array($this->settings['opt_cleanup']) : $this->settings['opt_cleanup'];
      if (!is_array($actions)) {
        // This also makes 'TRUE' an unknown value.
        $this->log("Unknown '@option' setting value '@value'! Ignoring.",
          array('@option' => 'opt_cleanup', '@value' => $this->settings['opt_cleanup']),
          WATCHDOG_WARNING);
        return;
      }
    }
    elseif (!$entity_ids) {
      // The "cleanup" action is performed on a default set of entities. If not
      // "cleanup", we assume the list of entities is empty, and exist silently.
      return;
    }

    // Add a bogus action to $actions as the last item, to make sure that
    // an update action is always executed inside the loop, if necessary.
    $actions[] = NULL;
    $update_fields = array();
    foreach ($actions as $key => $value) {
      if (!is_numeric($key)) {
        // Run update query later.
        $update_fields[$key] = $value;
      }
      else {

        if ($update_fields) {
          // Perform update action with already-collected field pairs, first.
          // (Reason for doing it here: we want to group update field/value
          // pairs into one call, but only if there's not another action in
          // between them. This is a useless feature as long as we only
          // support the "delete" action, but it doesn't hurt having it.)
          $this->processEntitiesAction($update_fields, $entity_ids);
        }
        $update_fields = array();

        if (isset($value)) {
          $this->processEntitiesAction($value, $entity_ids);
        }
      }
    }
  }

  /**
   * Performs a predefined action on specified entities. This method likely
   * only needs to be called by code dealing with 'after-process cleanup'; most
   * code should call processEntities() instead, to prevent messing up w.r.t.
   * the 2nd method argument.
   *
   * @param string|array $action
   *   Action to perform.
   *   - Array: key/value pairs for items to update.
   *   - "collect_ids": gather the values of the 'entities' that will/should be
   *     cleaned up (by another action): the original source-item IDs keyed by
   *     their Drupal 'entity' ID. Put in context['import_processed']. Our class
   *     does not need/use this, but a child class might.
   *   - "delete": delete the 'entities'.
   *   Any other string generates a warning by default, but can be implemented
   *   by child classes.
   * @param array $entity_ids
   *   (optional) IDs of entities to process. Omitting the argument / passing
   *   empty array means this method MAY operate on entities 'to be cleaned up
   *   after a process has finished'. (This can be determined by the specific
   *   method at will.) This is done because some classes do not need to
   *   determine an explicit list of IDs to clean up, rather they take action on
   *   'everything except the items that were processed'. So, take care to not
   *   pass an empty array by accident.
   */
  protected function processEntitiesAction($action, array $entity_ids = array()) {

    $pkey = $this->getEntityPrimaryKey();
    if (!$pkey) {
      $this->log('cleanupEntitiesAction() was called for table @table, which does not have a single-field primary key defined.',
        array('@table' => $this->settings['import_entity_type']), WATCHDOG_ERROR, FALSE);
      return;
    }

    $max_ids = isset($this->settings['import_cleanup_max_chunk_size']) ? $this->settings['import_cleanup_max_chunk_size'] : 1024;

    $in = 'IN';
    if (!$entity_ids) {
      if (!empty($this->settings['import_alt_cleanup'])) {
        // This should have been shrunk.
        $entity_ids = $this->context['entity_ids_not_encountered'];
      }
      else {
        $ids = $this->processedEntityIds();
        if (!$ids) {
          $this->log("Cleanup action could not determine processed items; skipping.", array(), WATCHDOG_WARNING);
          return;
        }
        if (count($ids) > $max_ids) {
          // We won't perform the "NOT IN" on this many IDs because it may break
          // the query. (We do not check the same limit for processedEntityIds()
          // because we expect $ids to be cached already.) So instead, fetch all
          // IDs and 'diff' the just-imported ones.
          $entity_ids = array_diff(
            db_select($this->settings['import_entity_type'], 'e')
              ->fields('e', array($pkey))
              ->execute()
              ->fetchCol(),
            $ids
          );
        }
        else {
          // If $entity_ids, we need to loop through chunks below. If not, we'll
          // use the same loop to facilitate code reuse.
          $entity_ids = $ids;
          $in = 'NOT IN';
        }
      }
    }

    if (is_array($action)) {
      // Update values.
      $updated = 0;
      // Facilitate for list of IDs that is too long by splitting it up.
      foreach (array_chunk($entity_ids, $max_ids) as $ids) {
        $res = db_update($this->settings['import_entity_type'])
          ->fields($action)
          ->condition($pkey, $ids, $in)
          ->execute();
        $updated = (is_numeric($updated) && is_numeric($res)) ? $updated + $res : '?';
      }

      $desc = array();
      foreach ($action as $key => $value) {
        $desc[] = "$key=$value";
      }
      // @todo decide on a 'logging strategy'. At the moment, we have to log here
      //       if we want to see any output from 'cleanup-on-finish', but 'action-on-processItem'
      //       also accounts + logs on finish, which is kind-of duplicate.
      //       BTW at the moment we do it here, not in processEntities(), because this is the only place we know how many
      //       things are updated if !$entity_ids. See also entity_import_job processEntities() comment.
      $this->log("@nr items updated to @action.",
        array('@nr' => $updated, '@action' => implode(', ', $desc)));
    }
    else {
      switch ($action) {
        // The first action in the switch statement is relatively harmless,
        // which is good if $action ever contains TRUE...
        case 'collect_ids':
          if (!isset($this->context['import_processed']) || !is_array($this->context['import_processed'])) {
            $this->context['import_processed'] = array();
          }
          // Get the values of the 'entities' that we'll clean up: the
          // original source-item IDs keyed by their Drupal 'entity' ID.
          $dest_field = isset($this->context['import_dst_pkey']) ?  $this->context['import_dst_pkey']
            : (isset($this->settings['import_dst_pkey']) ? $this->settings['import_dst_pkey']
              : '');
          $query = db_select($this->settings['import_entity_type'], 'f');
          if ($dest_field) {
            $query->fields('f', array($pkey, $dest_field));
          }
          else {
            $query->addField('f', $pkey);
            $query->addExpression('1');
          }
          foreach (array_chunk($entity_ids, $max_ids) as $ids) {
            $query->condition($pkey, $ids, $in);
            $res = $query->execute()->fetchAllKeyed();
            if ($res) {
              // Don't array_merge(); it renumbers numeric keys.
              $this->context['import_processed'] += $res;
            }
          }
          break;

        case 'delete':
          // Delete 'entities'.
          foreach (array_chunk($entity_ids, $max_ids) as $i => $ids) {
            $res = db_delete($this->settings['import_entity_type'])
              ->condition($pkey, $ids, $in)
              ->execute();
            // We can probably interpret $res as the number of rows, but we are
            // not absolutely sure.
            if ($res) {
              // @todo, here and elsewhere in all classes, replace @value by something that properly prints boolean values and arrays, and does not put hard enters around scalars like <pre> does.
              $this->log("Old items deleted. Delete statement@nr for processEntitiesAction 'delete' returned: '@value'.",
                array('@nr' => $i ? (' ' . ($i+1)) : '', '@value' => $res));
            }
            else {
              $this->log("Delete statement@nr for processEntitiesAction 'delete' returned: '@value'.",
                array('@nr' => $i ? (' ' . ($i+1)) : '', '@value' => $res), WATCHDOG_DEBUG);
            }
          }
          break;

        default:
          $this->log("Unknown @option value '@value'.",
            array('@option' => 'opt_cleanup', '@value' => $action),
            WATCHDOG_WARNING);
      }
    }
  }

  /**
   * Returns all IDs of items that were fetched / run through processItem().
   * This method is specifically meant to be called on finish(). Its use is to
   * compare the IDs with any IDs imported by an earlier process; whether it's
   * more generally useful remains to be seen.
   * (This code was split out into a separate function only to ease its
   * customization by child classes.)
   *
   * @return array
   *   Processed IDs. (Note "processed" is supposed to mean "ran through
   *   processItem()", and include any items discarded, not only the ones
   *   'usefully' processed.) Empty array means/can mean that we could not
   *   derive them.
   */
  protected function processedEntityIds() {
    if (!empty($this->settings['import_alt_cleanup'])) {
      return array();
    }
    // Construct an array of all processed items' IDs (and nothing else).
    $src_ids = array();
    foreach (array('new', 'updated', 'skipped', 'error') as $type) {
      $src_ids = array_merge($src_ids, $this->context[$type]);
    }
    return $this->resolveEntityIds($src_ids);
  }

  /**
   * The 'field mapping by destination' is an array, preferably one with at
   * least two levels:
   *
   * * the outer array keys are destination (Drupal) fieldnames, or meta tag
   *   names (metatag module) - the latter should be prefixed with 'metatag:'.
   *
   * * The value may be a single source fieldname or an array of fieldnames
   *   (with sequential 0-based keys), if you don't need other metadata in the
   *   mapping
   *
   * * However, to have a consistent and fully extendable mapping structure,
   *   the value should be an array containing at least the key 'src' OR
   *   'process'. In this case, the keys in the per-destination-field array are
   *   fixed and have the following values:
   *
   * - 'src'    : source fieldname(s). A string value or an array of one or more
   *              strings can be specified. The 'src' key may be left out (or
   *              set to an empty string), only if a 'process' key is present.
   *              In that case the process function is passed NULL as value,
   *              and it's expected to get its needed values from elsewhere.
   *              (See below for info about process functions.)
   *
   * - 'process': a string or array of (key/)values specifying actions to take
   *              on the source value before saving it.
   *              See below for a definition of array key/values (which is by
   *              definition incomplete, because a subclass can implement
   *              actions depending on the keys it chooses). If a string is
   *              specified, this is assumed to be a method name.
   *
   * - 'process:<SRC>': the same, but only applicable for a specific source
   *                    field. <SRC> must be among the values in 'src'.
   *   For field mentioned in 'src', this value is array_merge()'d with
   *   'process', i.e.:
   *   - both 'process' and 'process:<SRC>' actions are applied
   *   - 'process' actions come first (for those where ordering matters)
   *   - except: an alphanumeric array key in 'process:<SRC>' overrides the same
   *     key (action) in 'process'; only the 'process:<SRC>' action is executed.
   *
   * - 'hierarchy': only the value TRUE is supported. (Actions on other values
   *     are unspecified). This can be used with multi-source term reference
   *     fields, as an alternative to specifying 'parent_src' for all fields.
   *     It means each specified source field is the 'parent_src' for the
   *     source field following it.
   *     (So the output will be a multi-value term reference field where the
   *     1st source value is on the 1st level, the 2nd on the 2nd, etc.)
   *     Explicitly specified 'parent_src' actions in 'process' will override
   *     'hierarchy'.
   *
   * The actions inside 'process' and 'process:<SRC>' arrays can mean various
   * things: extra properties of a field, extra fields to process, actual
   * actions to take on the value...
   *
   * For 'actions to take / methods to call' (which generally contain uppercase
   * letters), the order in which they appear in the array is important; the
   * actions are chained together. Also, it is up to the person constructing the
   * mapping, to make sure that a process action can handle all possible input
   * values. Specifically:
   * - it should be able to handle whatever type of field values are passed to
   *   processItem(); this ImportJob does not make assumptions about, or perform
   *   checks on, those field values. However, if an array of values is passed,
   *   the method will be called separately for each individual value.
   * - custom 'process' methods may return (e.g. fieldAPI) structures instead of
   *   scalar values (see below). If these methods are followed by other
   *   actions, those actions should be able to handle the structure as input.
   *
   * The keys are defined as follows:
   *
   * General:
   *
   * - numeric / no key, and the value is not defined below AND contains at
   *   least one uppercase character:
   *
   *   The value is the name of a method, which will be called to convert the
   *   value into another one. (The developer is expected to implement a public
   *   method with this name, in a subclass.)
   *
   *   The method gets passed three arguments:
   *   - the value (which will only be NULL if no 'src' key was in the mapping)
   *   - NULL
   *   - the name of the source field (where the initial value came from).
   *
   *   See below for return values.
   *
   * - anything not mentioned below, which contains at least one uppercase char:
   *
   *   The key is the name of a method, which will be called in the same way
   *   as above, but the 2nd argument passed is the value defined with this key.
   *
   * - anything not mentioned below, which has no uppercase characters:
   *
   *   It's ignored, assuming it is a property used by a child class'
   *   buildEntityValue() implementation. (Indeed, destination-based processing
   *   can also look into the mapping for properties. And it's strange to
   *   - have both actions and properties on the same level
   *   - have destination-based post processing look into the 'process:<SRC>' or
   *     even just 'process' properties...
   *   ...maybe later we'll look if these should be separated.)
   *
   * - 'replaceString':
   *   The value must be an array of two strings. String replacement of the
   *   first into the second will be performed on the source value.
   *
   * - 'booleanSwitch':
   *   The value must be an array of two values. If the source value 'is false',
   *   it is replaced with the first array value, otherwise with the second.
   *
   * - 'skip_if_empty_src':
   *   The current field value will not be used if (any of) the specified source
   *   field(s) has a value.
   *
   * - 'skip_if_nonempty_src':
   *   The current field value will not be used if (any of) the specified source
   *   field(s) has no value.
   *
   *   The value to the 'skip_if_empty_src' / 'skip_if_nonempty_src' keys may be
   *   a string (single source field) or array (multiple source fields).
   *
   * Keys which only have effect on certain types of destination(Drupal) fields:
   *
   * - 'resolve'                - for list_text and list_integer fields
   *   If TRUE or not defined (default), the field value from AFAS is assumed to
   *   be a field display value, so it will be resolved from the field's allowed
   *   values. If the value is not found, this will result in an error being
   *   printed and the field is not updated.
   *   If set to FALSE, the value from AFAS will be stored in the field as-is.
   *   (It's assumed to be a key in the field's allowed values. If not, this
   *   will not yield an error until the field is edited through the UI.)
   *
   * - 'vocabulary'             - for taxonomy_term_reference fields.
   *   Usually NOT necessary to specify, because the linked vocabulary is
   *   retrieved from field settings. It's only for unexpected situations.
   *
   * - 'parent_src'             - for hierarchical terms
   *   A (source) fieldname or an array of fieldnames. The current field value
   *   will be a child of the specified parent field-value (if an array, the
   *   last will be the parent, the next-last the grandparent, etc); if a parent
   *   value does not exist, the term is created on the first level of the tree.
   *   (Unless 'skip_if_empty_src' is set to the parent_field(s) too.)
   *   parent_field=0 forces the term on the first level of a tree.
   *   NOTES:
   *   - This does not make much sense in 'process', only in 'process:<SRC>'.
   *   - For common cases, instead of this property, you can usually just set
   *     'hierarchy' => TRUE in the outer array.
   *   - For 'multi ancestors', the whole path of ancestors must be specified,
   *     i.e. a 3rd level term must have an array with 2 values, even if its
   *     direct parent field also has a 'parent_src' property. (Failure to
   *     specify the grandparent will make all parents appear on the first tree
   *     level too. This cannot be fixed -by interpreting the parent's
   *     'parent_src'- without introducing edge cases where things go wrong,
   *     at least theoretically.)
   *
   * - 'format'                 - for any fields with 'format' FieldAPI property
   *   optional; defaults to filtered_html
   *
   * - 'summary_length'         - for textarea
   *   optional; defaults to text_summary() default,
   *   i.e. 'teaser_length' variable.
   *
   * - 'currency'               - for commerce_price
   *    currency code to attach. (The field value itself holds the price.)
   *
   * - 'currency_src'           - for commerce_price
   *    (single) source field holding currency code to attach.
   *    Overrides 'currency', IF the processed value of this field is non-empty.
   *
   * - 'tax_name'               - for commerce_price
   *    Drupal system name of tax rate object to attach.
   *
   * - 'tax_src'                - for commerce_price:
   *    (single) source  holding tax rate to attach.
   *    Overrides 'tax_name', IF the processed value of this field is non-empty.
   *
   * - 'title_src'              - for media:
   *    Title of video. (The 'main' field holds the URL.)
   *
   * --
   *
   *   Child classes' custom methods in 'process' should log any errors
   *   as needed, using $this->log(). See buildItemValues() for examples.
   *
   *   Return values may be:
   *
   *   - NULL for skipping the value (e.g. on error);
   *
   *   - a single scalar value;
   *
   *   - an array having at least one non-'standard' key (or an empty array);
   *     this will be assumed to be a valid single-value structure.
   *     (e.g. for FieldAPI: an array with keys dependent on the destination
   *     field: 'value', 'tid', etc.)
   *
   *   - an array with 'standard' (sequential 0-based numeric) keys: each array
   *     value must itself be a scalar or a valid single-value structure, and
   *     will be added to the destination entity as an individual value.
   *
   * Each individual value or structure will be passed one by one into
   * buildEntityValue(); whatever value gets built, will be assigned to the
   * entity's property. It's up to each 'process' function to only return
   * structures that are recognized by buildEntityValue() (and by other
   * process functions, if these are chained).
   */

  /**
   * Constructs normalized field mapping, from a non-normalized mapping by
   * destination (Drupal) field, present in the 'import_field_mapping' setting.
   * Sets the normalized mapping in the job context.
   *
   * See above for the mapping structure.
   *
   * @param array $context
   *   Job context variable into which the normalized mapping will be set.
   */
  protected function setMappingByDest(array &$context) {

    // If the field mapping is set already, assume it's complete.
    // (Some child classes may call this method multiple times, so don't do
    // unnecessary processing.)
    if (!empty($context['import_field_mapping'])) {
      return;
    }

    $mapping = array();
    foreach ($this->settings['import_field_mapping'] as $dest_field => $dest_map) {

      // Normalize source field
      if (!is_array($dest_map)) {
        // Only a single source field is present. Normalize.
        $dest_map = array('src' => array($dest_map));
      }
      elseif (isset($dest_map['src'])) {
        // Normalize the 'src' value which is present.
        if (!is_array($dest_map['src'])) {
          $dest_map['src'] = array($dest_map['src']);
        }
      }
      // Value is an array without 'src' set:
      // If 'process' is present, that's fine. Fill 'src' with empty source.
      elseif (isset($dest_map['process'])) {
        $dest_map['src'] = array('');
      }
      // Otherwise, this array value itself is the whole source, only if the
      // array has all sequential 0-based keys. We have a quick way of testing:
      elseif ($dest_map && $dest_map === $this->normalizeValue($dest_map)) {
        // Normalize.
        $dest_map = array('src' => $dest_map);
      }
      else {
        // Array without 'src' or 'process' value: error.
        // (There may be valid custom cases for 'src' being a keyed array, but
        // in that case it should be placed expicitly inside 'src'; it will not
        // automatically be normalized.)
        $this->log("Field mapping for '@dest' misses 'src' or 'process' value; ignoring.", array('@dest' => $dest_field), WATCHDOG_ERROR);
        continue;
      }

      // Test an array value in 'src' (that was either just set, or unchanged):
      // It should be a flat array with simple values. (It's unlikely for the
      // values to be integers, but it's allowed.)
      if (is_array($dest_map['src'])) {
        foreach ($dest_map['src'] as $key => $value) {
          if (!is_string($value) && !is_int($value)) {
            $this->log("Field mapping for '@dest' has invalid value @v; ignoring.", array('@dest' => "$dest_field > $key", '@v' => print_r($value, TRUE)), WATCHDOG_ERROR);
            unset($dest_map['src'][$key]);
          }
        }
      }


      // Normalize 'process:SRC'; make sure it exists (even if it's an empty
      // array) and move all general 'process' functions into it.

      // Get 'process'
      $process_all = array();
      if (isset($dest_map['process'])) {
        if (!is_array($dest_map['process'])) {
          $dest_map['process'] = array($dest_map['process']);
        }
        $process_all = $dest_map['process'];
      }

      // Move 'process' into 'process:<SRC>' for each source field. Put general
      // 'process' last.
      foreach ($dest_map['src'] as $src) {

        if (isset($dest_map["process:$src"])) {
          $dest_map["process:$src"] = array_merge(
            is_array($dest_map["process:$src"]) ?
              $dest_map["process:$src"] : array($dest_map["process:$src"]),
            $process_all
          );
        }
        elseif ($process_all) {
          $dest_map["process:$src"] = $process_all;
        }
      }
      // Empty 'process' which is now unnecessary; only the value of
      // 'process:<SRC>' needs to be checked. (After checking if it's set.)
      if ($process_all) {
        unset($dest_map['process']);
      }

      // Expand 'hierarchy' to 'parent_src'. (Is only necessary in edge cases;
      // see buildApiFieldValue().)
      if (!empty($dest_map['hierarchy'])) {
        $parents = array();
        foreach ($dest_map['src'] as $src) {
          // Do not add to first level, or to anything that has parent_src set.
          if ($parents && !isset($dest_map["process:$src"]['parent_src'])) {
            if (!isset($dest_map['process'])) {
              $dest_map['process'] = array();
            }
            $dest_map["process:$src"]['parent_src'] = $parents;
          }
          $parents[] = $src;
        }
      }

      $mapping[$dest_field] = $dest_map;
    }

    $context['import_field_mapping'] = $mapping;

    // If one of pkey / import_dst_pkey is filled and the other is empty, derive
    // one from the other.
    $this->setMappedFieldNames('pkey', $context);
    // Same with import_src_changed / import_dst_changed.
    $this->setMappedFieldNames('changed', $context);
  }

  /**
   * Helper method that derives source from destination field name, or vice
   * versa, through the field mapping. Both names are stored in context. (Not in
   * settings because
   * - we only need to do this once because context gets remembered;
   * - the general rule is we don't change settings at run time.)
   *
   * @param string $field_setting
   *   Part of the setting name for a field. Either the 'source' or
   *   'destination' field name needs to be already set. (Or both, in which case
   *   the field mapping will not be checked.)
   * @param array $context
   */
  protected function setMappedFieldNames($field_setting, array &$context) {

    // Example: $field_setting is 'changed' => this function will make sure
    // the 'import_src_changed' and 'import_dst_changed' context properties are
    // filled. Both properties contain a field name (in the source item &
    // destination entity, respectively). We take 'src' and 'dst' as an
    // abbreviation for 'source / destination field name'.
    $src_prop = "import_src_$field_setting";
    $dst_prop = "import_dst_$field_setting";
    // Exception: We store 'import_src_pkey' in the context, but the setting
    // name is 'pkey'.
    $src_setting = $field_setting === 'pkey' ? $field_setting : $src_prop;

    if (!empty($this->settings[$src_setting]) && !empty($this->settings[$dst_prop])) {
      // Mapping is not filled automatically. There has to be a mapping between
      // the fields that were independently set, otherwise unexpected things may
      // happen. For instance for the 'pkey' settings (which may be set
      // automatically by a standard job/fetcher), if the mapping is not set,
      // the destination property is likely not filled. resolveEntityIds() will
      // not find matches and the process will keep creating new destination
      // 'entities' every time it is run.
      // Check for mapping.
      if (empty($context['import_field_mapping'][$this->settings[$dst_prop]]['src'])
          || !in_array($this->settings[$src_setting], $context['import_field_mapping'][$this->settings[$dst_prop]]['src'], TRUE)) {
        // It's not an immediate disaster, so we don't have to have a return
        // value here. Just warn.
        $this->log("'@setting' field name properties are set for the source and destination items, but they are not mapped to each other. This can cause strange behavior!",
          array('@setting' => $field_setting), WATCHDOG_WARNING);
      }
    }

    // Derive source property #1: simply from setting
    if (empty($context[$src_prop]) && !empty($this->settings[$src_setting])) {
      $context[$src_prop] = $this->settings[$src_setting];
    }

    // Derive destination property: simply from setting, or from src + mapping
    if (empty($context[$dst_prop])) {

      if (!empty($this->settings[$dst_prop])) {
        $context[$dst_prop] = $this->settings[$dst_prop];
      }
      elseif (!empty($context[$src_prop])) {
        // Destination property is the field (or the first field found) which
        // mentions this source field.

        foreach ($context['import_field_mapping'] as $map_dst => $dest_map) {
          if (array_search($context[$src_prop], $dest_map['src']) !== FALSE) {
            $context[$dst_prop] = $map_dst;
          }
        }
      }
    }

    // Derive source property #2: from destination + mapping.
    if (empty($context[$src_prop])
        && !empty($context[$dst_prop])
        && !empty($context['import_field_mapping'][$context[$dst_prop]]['src'][0])) {

      // Source field is the first field mentioned as source, in the mapping.
      $context[$src_prop] = $context['import_field_mapping'][$context[$dst_prop]]['src'][0];
    }
  }

  /**
   * Translates source IDs (stored in destination 'entity') to destination IDs.
   *
   * Caches results in $this->entityIds if that is non-NULL. This cache is first
   * keyed by "ENTITY_TYPE:BUNDLE" then source ID; the values are arrays of
   * entity IDs.
   * The cache is usually 'enabled' by setting
   *   $this->entityIds =& $context['import_ids']
   *
   * We do not do 'negative caching' here. This is a feature, so an ID
   * that does not exist at first, can be resolved after a new entity is saved.
   *
   * @param array $src_ids
   *   Array of IDs from the source system. (Present in the destination entity's
   *   import_dst_pkey field.)
   * @param string $entity_type
   *   (optional) Type of destination entity. Defaults to 'import_entity_type'
   *   setting.
   * @param array $options
   *   (optional) Query options. Supported keys / values:
   *   - return_type: type/format of the returned array
   *     - empty / not present: regular array( 0 => Eid, ...)
   *     - 'keyed': results keyed by source value, i.e. array( src => array ( 0 => Eid, ...) , ... )
   *         outer array keys (source IDs) will be uppercased if
   *         $this->settings['import_ids_case_sensitive'] == FALSE (which is the
   *         default). The caller is expected to accommodate for this!
   *  - requery_all: if TRUE value, query all items. Overrides $src_id argument.
   *     If cache is enabled, it is cleared and fully repopulated.
   *  - cache_bin: the subgroup in the cache where items are searched / stored
   *     (if caching is on). It should normally not be provided and will be
   *     derived from other settings.
   *
   * @return array
   *   Destination 'entity' IDs in the requested format.
   *
   * @throws \Exception
   */
  protected function resolveEntityIds(array $src_ids, $entity_type = '', array $options = array()) {
    // A note: the 'negative caching' could be an option (setting) - we usually
    // don't care because our standard classes populate the cache on
    // saveEntity(), so any effect of negative caching is canceled out in most
    // cases. However we don't need to requery entities usually, so we
    // haven't cared until now.

    if (empty($src_ids) && empty($options['requery_all'])) {
      $this->log('resolveEntityIds called with empty parameter?', array(),
        WATCHDOG_WARNING, FALSE);
      return array();
    }
    if (isset($options['return_type']) && $options['return_type'] !== 'keyed') {
      throw new \Exception("resolveEntityIds called unknown return_type option value '$options[return_type]'.");
    }

    // Get data from cache.

    if (isset($this->entityIds)) {
      // Make sure cache bin is set.
      if (empty($options['cache_bin'])) {
        if (!$entity_type) {
          $entity_type = $this->settings['import_entity_type'];
        }
        // Even though we don't have a cache subtype (bundle), we'll keep the
        // colon for possible 'compatibility' with extending classes that do.
        $options['cache_bin'] = "$entity_type:";
      }

      if (!isset($this->entityIds[$options['cache_bin']])) {
        $this->entityIds[$options['cache_bin']] = array();
      }
      $bin = $this->entityIds[$options['cache_bin']];

      // Extra code is introduced in various places to hopefully speed things up
      // by not doing array_intersect()'s when we're only looking for one item.
      // That extra code checks $single_id.
      $single_id = count($src_ids) == 1 && empty($options['requery_all'])
        ? ($this->settings['import_ids_case_sensitive']
          ? strtoupper(reset($src_ids)) : reset($src_ids))
        : NULL;

      // Quick-resolve $single_id first. (Would have the same
      // result when falling through to the array_intersect_key().)
      if (isset($single_id) && isset($bin[$single_id])) {
        return empty($options['return_type'])
          ? $bin[$single_id] : array($single_id => $bin[$single_id]);
      }

      if (!$this->settings['import_ids_case_sensitive']) {
        // Convert $src_ids to uppercase (because our query system does not
        // distinguish case) - but take care not to change numbers to strings.
        // (It doesn't matter for cache array keys but may matter for e.g. query
        // arguments.)
        foreach ($src_ids as &$src_id) {
          if (is_string($src_id)) {
            $src_id = strtoupper($src_id);
          }
        }
      }

      if (empty($options['requery_all'])) {
        $precached_data = array_intersect_key($bin, array_flip($src_ids));
      }
      else {
        $this->entityIds[$options['cache_bin']] = $precached_data = array();
      }
    }
    else {
      $precached_data = array();
    }

    // Get data from database (or: additional data, from database into cache)

    if (!isset($this->entityIds) || count($precached_data) !== count($src_ids) || !empty($options['requery_all'])) {

      $data = $this->executeResolveEntityIdsQuery(
        $precached_data ? array_diff($src_ids, array_keys($precached_data)) : $src_ids,
        $entity_type,
        $options
      );
      if (!isset($this->entityIds)) {
        // If caching is off, the queried data is already in the return format.
        return $data;
      }
    }

    // Status at this point:
    // - Caching is on.
    // - We've returned items for a single cached item already (only for speed).
    // - We have 2 sets of data now:
    //   - $precached_data (the items already cached before this call)
    //   - the cache in $this->entityIds. Also $data *may* hold the full data
    //     for the current bin but that's not better than using the full cache.

    // Compile now-cached data.
    // We can ignore $precached_data, because we have everything in one bin.

    $bin = $this->entityIds[$options['cache_bin']];

    if (isset($single_id)) {
      // Optimized code (would have the same result as the below):
      if (isset($bin[$single_id])) {
        return empty($options['return_type'])
          ? $bin[$single_id] : array($single_id => $bin[$single_id]);
      }
      return array();
    }

    // Get all items, including newly cached ones. We can get them all from one
    // cache bin (ignoring $precached_data) because we only have one per entity.
    $data = empty($options['requery_all'])
      ? array_intersect_key($bin, array_flip($src_ids)) : $bin;

    if (!empty($options['return_type'])) {
      // Return 'keyed' format, which is as stored in cache bin.
      return $data;
    }
    // Transform data into a non-keyed array.
    $return = array();
    foreach ($data as $dest_ids) {
      // Probably faster than array_merge() because $dest_ids is short:
      foreach ($dest_ids as $dest_id) {
        $return[] = $dest_id;
      }
    }
    return $return;
  }

  /**
   * Executes the query for resolveEntityIds() and returns data. If caching is
   * on, adds data directly to a cache bin. Split out for easier subclassing.
   *
   * @param array $src_ids
   *   Array of IDs from the source system; must already be uppercased if
   *   caching is on and job settings require it.
   * @param string $entity_type
   *   Type of destination entity. Defaults to 'import_entity_type' setting.
   * @param array $options
   *   Query options. If caching is on, then 'cache_bin' option must already be
   *   set.
   *
   * @return array
   *   Data: the full cache group (not only the data just queried) if caching is
   *   on; otherwise its format depends on the 'return_type' option.
   *
   * @see resolveEntityIds()
   */
  protected function executeResolveEntityIdsQuery(array $src_ids, $entity_type, array $options) {
    $data = array();

    if (!$entity_type) {
      $entity_type = $this->settings['import_entity_type'];
    }

    // By default, the importer stored the 'entity field containig the primary
    // key value' in $this->context, because it's not necessarily filled when
    // using a field mapping.
    // Some classes do not use a field mapping, and only have the field set in
    // $this->settings.
    $dest_field = isset($this->context['import_dst_pkey']) ?  $this->context['import_dst_pkey']
      : (isset($this->settings['import_dst_pkey']) ? $this->settings['import_dst_pkey']
        : '');
    if (!$dest_field) {
      $this->log("executeResolveEntityIdsQuery() was called while 'Source id field in destination entity' was not set.",
        array(), WATCHDOG_ERROR, FALSE);
    }
    elseif (!$db_idfield = $this->getEntityPrimaryKey($entity_type)) {
      // We already checked, but anyway...
      $this->log("executeResolveEntityIdsQuery() was called for table @table, which does not have a single-field primary key defined.",
        array('@table' => $entity_type), WATCHDOG_ERROR, FALSE);
    }
    else {

      $query = db_select($entity_type, 'e');
      if (empty($options['requery_all'])) {
        if (count($src_ids) == 1) {
          $query->condition($dest_field, reset($src_ids));
        }
        else {
          $query->condition($dest_field, $src_ids, 'IN');
        }
      }
      if (!empty($this->settings['import_entity_query_filter_field'])
          && !empty($this->settings['import_entity_query_filter_value'])) {
        $query->condition(
          $this->settings['import_entity_query_filter_field'],
          $this->settings['import_entity_query_filter_value'],
          is_array($this->settings['import_entity_query_filter_value']) ? 'IN' : '='
        );
      }

      // Get results
      if (isset($this->entityIds)
          || (isset($options['return_type']) && $options['return_type'] === 'keyed')) {
        // array( Source => array( Eid ) )

        if (isset($this->entityIds)) {
          // Append results directly to cache; construct return value later.
          $data =& $this->entityIds[$options['cache_bin']];
        }

        $query->fields('e', array($dest_field, $db_idfield));
        $result = $query->execute();
        foreach ($result as $row) {
          // Uppercase source IDs when necessary. Number to string conversion
          // does not matter here, because they're only used as array key.
          $src_id = $this->settings['import_ids_case_sensitive'] ?
            $row->$dest_field : strtoupper($row->$dest_field);
          if (!isset($data[$src_id])) {
            $data[$src_id] = array();
          }
          $data[$src_id][] = $row->$db_idfield;
        }
      }
      else {
        // array( Eid )  (0-based numeric keys)
        $query->addField('e', $db_idfield);
        $data = $query->execute()->fetchCol();
      }
    }

    return $data;
  }

  /**
   * Gets the primary key field name for an 'entity' (SQL table).
   *
   * @param string $entity_type
   *   Type identifier. Empty to take the configured 'import_entity_type'
   *   setting for this job.
   *
   * @return string
   *   Primary key field name; '' if not found.
   */
  private function getEntityPrimaryKey($entity_type = '') {
    if (!$entity_type) {
      $entity_type = $this->settings['import_entity_type'];
    }
    $schema = drupal_get_schema($entity_type);
    if (empty($schema['primary key']) || !is_array($schema['primary key']) || count($schema['primary key']) != 1) {
      // We don't support this (and have checked in checkSettings()).
      return '';
    }
    return reset($schema['primary key']);
  }

  /**
   * Gets the ID of a (destination) entity.
   *
   * @param object $entity
   *   The entity.
   *
   * @return mixed
   *   The ID value or NULL if the entity has none (if it's new).
   *
   * @throws \Exception
   */
  protected function getEntityId($entity) {
    if (!$pkey = $this->getEntityPrimaryKey()) {
      throw new \Exception('Entity type "' . $this->settings['import_entity_type'] . "' has no primary key.");
    }
    if (isset($entity->$pkey)) {
      return $entity->$pkey;
    }
  }

  /**
   * Loads destination entities.
   *
   * @param array $ids
   *   IDs of entities to load.
   *
   * @return array
   *   Entities (objects), keyed by their IDs.
   *
   * @throws \Exception
   */
  protected function loadEntities(array $ids) {
    // @todo often, this call is preceded by an equivalent call to resolveEntityIds(). Can we dedupe somehow?

    // Get default values for the entity from the schema. This is different from
    // the 'import_default_values' setting, because those must/will be
    // explicitly set in the destination entity, and can/will be preprocessed.
    // The values here don't need to be set; they are supposed to be
    // automatically set by the database backend if the field is empty. The
    // purpose of getting them is to prevent setEntityValue() from explicitly
    // setting a 'changed' flag when the destination field is NULL (which will
    // be turned into the default value on save). We do it in loadEntities() to
    // keep start() / processItem() free of SQL backend specific commands.
    if (!isset($this->context['import_default_entity_values'])) {
      $this->context['import_default_entity_values'] = array();
      $schema = drupal_get_schema($this->settings['import_entity_type']);
      foreach ($schema['fields'] as $destination => $info) {
        if (isset($info['default'])) {
          $this->context['import_default_entity_values'][$destination] = $info['default'];
        }
      }
    }

    if (!$pkey = $this->getEntityPrimaryKey()) {
      throw new \Exception('Entity type "' . $this->settings['import_entity_type'] . "' has no primary key.");
    }

    $query = db_select($this->settings['import_entity_type'], 'f');
    $query->condition($pkey, $ids);
    $query->fields('f');
    return $query->execute()->fetchAllAssoc($pkey);
  }

  /**
   * Updates or creates a (destination) entity.
   *
   * @param object|null $entity
   *   The entity to update, or null to create new entity.
   *   (Note that on return, $entity will be the valid updated entity which
   *   the caller can keep using, ONLY when a non-empty object was passed.)
   *
   * @return mixed
   *   ID of created/updated entity. 0 means an existing entity was not updated;
   *   FALSE or NULL indicates error (though NULL generally shouldn't happen).
   */
  protected function updateEntity($entity = NULL) {

    $new_entity = !(bool)$entity;
    if ($new_entity) {
      $entity = $this->initEntity();
    }

    // Process & set the item values.
    $src_id = $this->src_item[$this->context['import_src_pkey']];
    $dsts = $changed_parts = $deferred_values = $remember_old_values = array();

    // Allow items to defer until others are processed, and check for deadlocks,
    // with an outer loop.
    $needs_processing = array_keys($this->context['import_field_mapping']
                                   + $this->settings['import_default_values']);
    while (count($needs_processing) != count($dsts)) {
      $dsts = $needs_processing;
      $needs_processing = array();

      // Process all 'remaining' destination properties.
      foreach ($dsts as $dst) {

        list($dst_prop, $dst_type, $sub_dst) = $this->parseDestinationKey($dst);

        if ($this->isProcessingDeferred()) {
          $needs_processing[$dst] = TRUE;
        }
        else {
          // Actually process one destination.

          // Set initial value to add source values into.
          $dst_value = NULL;
          // Set initial 'deferred' property value. This is always an array, to
          // make sure that no values get lost. If code 'defers' simple property
          // values (which this class never does), it should often convert
          // this one-element array back to a single value, afterwards.
          $unbuilt_values = array();

          // Gather values from various sources:
          // 1) value(s) from source field(s) in the mapping
          if (isset($this->context['import_field_mapping'][$dst])) {
            $dst_mapping = $this->context['import_field_mapping'][$dst];

            // If 'src' has 0-based sequential keys, the processed source values
            // will be added (i.e. run through buildEntityValue() as individual
            // values.
            // If it has other keys, this is considered to be an 'atomic' value-
            // structure consisting of several sub-values, which are all derived
            // individually but then run through buildEntityValue() as a whole.
            // (The primary difference between this and individual mappings for
            // 'dest:sub-dest', is that setting the value as a whole could
            // influence sub-destinations that are present in the entity but not
            // in the mapping. e.g. in an addressfield, if the address is set as
            // a whole, a little-used subfield like 'sub_premise' will be reset
            // and not 'pollute' the new address by accident.)
            // We have a helper function for checking array keys:
            $sequential_keys =
              $dst_mapping['src'] === $this->normalizeValue($dst_mapping['src']);

            foreach ($dst_mapping['src'] as $key => $src) {

              // Get (possibly processed) item value(s) from source field.
              // (It's usually one value but can theoretically be multiple,
              // from the source being an array, or from processing.)
              $src_values = $this->buildItemValues($src);

              // $src_values must be a 0-based numerically keyed array. We could
              // run things through normalizeValue() yet again, but instead
              // we'll check and fail. buildItemValues() should really not
              // return scalars or single structures.
              if (!is_array($src_values)) {
                $this->log('Expected an array from buildItemValues()! Process functions for dst/src %d/%s: %m; resulting non-array value: %v',
                  array('%d' => $dst, '%s' => $src, '%v' => print_r($src_values, TRUE), '%m' => (isset($dst_mapping["process:$src"]) ? print_r($dst_mapping["process:$src"], TRUE) : '-')),
                  WATCHDOG_ERROR, FALSE);
              }
              else {

                foreach ($src_values as $k => $val) {
                  if (!is_numeric($k)) {
                    // buildItemValues() did something strange. We refuse to
                    // process values with non-numeric keys. (This may be a
                    // single value-structure that was returned by accident, and
                    // we want to catch this rather than set strange values.)
                    $this->log('Expected an array with numeric keys from buildItemValues()! The following value will not be processed: %v (key: %k). Process functions for dst/src %d/%s: %m; resulting array value: %a.',
                      array('%d' => $dst, '%s' => $src, '%a' => print_r($src_values, TRUE), '%k' => $k, '%v' => print_r($val, TRUE), '%m' => (isset($dst_mapping["process:$src"]) ? print_r($dst_mapping["process:$src"], TRUE) : '-')),
                      WATCHDOG_ERROR, FALSE);
                  }
                  else {

                    $unbuilt_values[$key] = $val;
                    if ($sequential_keys) {
                      $dst_value = $this->buildEntityValue($val, $dst_prop, $dst_value, $dst_type, $src);
                    }
                  }
                }
              }
            }
            if (!$sequential_keys && isset($unbuilt_values)) {
              // Build one value. (We have no single source field to pass.)
              $dst_value = $this->buildEntityValue($unbuilt_values, $dst_prop, $dst_value, $dst_type);
            }
          }
          // 2) default value(s), only if mapping was not present or the
          //    value was empty. (Emptiness is best checked by $unbuilt_values.)
          if (!$unbuilt_values && isset($this->settings['import_default_values'][$dst])) {
            foreach ($this->normalizeValue($this->settings['import_default_values'][$dst])
              as $val) {
              $dst_value = $this->buildEntityValue($val, $dst_prop, $dst_value, $dst_type);
              $unbuilt_values[] = $val;
            }
          }

          // Check deferring for the second time, after constructing value.
          if ($this->isProcessingDeferred($dst_value)) {

            // Save the value in 'context', to reprocess on finish().
            $deferred_values[$dst] = $unbuilt_values;

            // Also remember the existing Drupal entity value; otherwise we'd
            // need to load the entity again when comparing later.
            // @todo make setting for this, so we have an option not to overload $context?
            if ($sub_dst) {
              // Remember the 'destination + sub' separately from other subs.
              $remember_old_values[$dst_prop][$sub_dst] =
                !empty($entity->{$dst_prop}[$sub_dst]) ? $entity->{$dst_prop}[$sub_dst] : array();
            }
            else {
              $remember_old_values[$dst] = !empty($entity->{$dst}) ? $entity->{$dst} : array();
            }

            $needs_processing[$dst] = FALSE;
          }
          // If not deferred: set our value, or clear existing value.
          elseif ($changed_part = $this->setEntityValue($dst_value, $entity, $dst_prop, $dst_type, $sub_dst)) {
            $changed_parts[$changed_part] = TRUE;
          }
        }

      }
    }
    if (array_filter($needs_processing)) {
      $this->log("Item %id has fields which could not be processed: %fields",
        array('%id' => $this->src_item[$this->context['import_src_pkey']],
              '%fields' => implode(', ', $needs_processing)),
        WATCHDOG_WARNING);
    }

    if ($changed_parts || ($new_entity && $deferred_values)) {
      $return = $id = $this->saveEntity($entity, $changed_parts);
    }
    else {
      $return = 0;
      $id = $this->getEntityId($entity);
    }

    // Remember deferred values, unless saving failed.
    if ($id && $deferred_values) {
      $this->context['import_deferred'][$src_id] = $deferred_values;
      // We want to remember old values by their source AND destination ID:
      $this->context['import_old_values'][$src_id][$id] = $remember_old_values;
    }
    return $return;
  }

  /**
   * Creates a new entity with initial properties.
   *
   * @return object
   *   The entity.
   */
  protected function initEntity() {
    return new stdClass();
  }

  /**
   * Saves a (destination) entity.
   *
   * @param object $entity
   *   The entity to save.
   * @param array $changed_parts
   *   Indicates if the entity has changed, and if yes, which parts (in the
   *   array keys). If empty, no save is done by the default implementation.
   *
   * @return mixed
   *   ID of the (un)saved entity. FALSE or NULL indicates error (though NULL
   *   generally shouldn't happen).
   */
  protected function saveEntity($entity, array $changed_parts = array('*' => TRUE)) {

    $id = $this->getEntityId($entity);
    if ($id === FALSE) {
      return FALSE;
    }

    if ($changed_parts) {
      if ($id) {
        if (!$pkey = $this->getEntityPrimaryKey()) {
          return FALSE;
        }
        $pkeys_arg = array($pkey);
      }
      else {
        $pkeys_arg = array();
      }
      if (!drupal_write_record($this->settings['import_entity_type'], $entity, $pkeys_arg)) {
        return FALSE;
      }

      if (!$id && isset($this->entityIds)) {
        $id = $this->getEntityId($entity);

        // Cache new ID. Strings must be uppercased usually.
        $src_id = $this->src_item[$this->context['import_src_pkey']];
        if (is_string($src_id) && !$this->settings['import_ids_case_sensitive']) {
          $src_id = strtoupper($src_id);
        }
        if (isset($context['import_ids'][$this->settings['import_entity_type'] . ':'][$src_id])) {
          $this->log('saveEntity() just saved a new entity (@id) for a source item (@src_id) which already has a linked entity-id cached. This probably indicates a code error.',
            array('@id' => $id, '@src_id' => $src_id), WATCHDOG_WARNING);
          // We're not recording $id now. Not sure what the best policy is.
        }
        else {
          $context['import_ids'][$this->settings['import_entity_type'] . ':'][$src_id] = array($id);
        }
      }

    }

    return $id;
  }

  /**
   * Update 'deferred' values, in destination 'entities'.
   */
  protected function updateDeferred(array &$context) {
    // Do nothing, for now. Just an empty stub method called from finish().
  }

  /**
   * Checks whether processing should be deferred until later, on a 'destination
   * field' level.
   *
   * This method is called twice; once at the start of processing a field and
   * once when the buildItemValues() + add*Value() call(s) have been done.
   * In the first case, $value is not passed.
   * Various information can/should be determined from class-wide properties,
   * e.g. $this->current_dst.
   *
   * @param $value
   *   Value to be set in destination property.
   *
   * @return boolean
   *   FALSE if the item can be processed now.
   */
  protected function isProcessingDeferred($value = NULL) {
    // This default method implementation:
    // - never wants to defer on the first call;
    // - relies on outside code setting $this->defer (during
    //   buildItemValues() / buildEntityValue() calls), to indicate when
    //   processing for a destination should be deferred.

    // We distinguish between the first call and $value being NULL on the
    // second call (though that last thing realistically doesn't happen):
    if (func_num_args() == 0) {
      // Initialize our state variable.
      $this->defer = FALSE;
    }
    return $this->defer;
  }

  /**
   * Parses destination string; deduces type, and sub-destination string.
   * (In e.g. the default Drupal-entity importer, the sub-destination will be a
   * language code for the 'fieldapi' type.)
   *
   * Also sets class-wide property for 'destination', for use by methods setting
   * the value in this destination. (processedItemValue(), buildEntityValue()
   * and setEntityValue().)
   *
   * This implementation returns empty type & destination, but still returns
   * a 3-element array because calling methods are designed to deal with
   * child classes that return a sub-destination.
   *
   * @return array
   *   List of values:
   *   - The real destination property
   *   - 'type' of destination property. Always '' in our case because we don't
   *     have any different types.
   *   - a named sub-part of the destination property / entity. Always '' in our
   *     case because we don't support any 'part' destinations yet.
   */
  protected function parseDestinationKey($destination) {

    $this->current_dst = $destination;

    return array($destination, '', '');
  }

  /**
   * Sets (or resets) a value inside an entity.
   *
   * @param mixed $value
   *   Value to set.
   * @param $entity
   *   The entity to update.
   * @param string $destination
   *   Destination name, as defined in mapping.
   * @param string $dest_type
   *   'type' of destination property. (This class has no support for them;
   *   subclasses can support different logic for setting different
   *   property types.) If not provided, this means the type and 'sub'
   *   destination will still be determined, and $value must be a single value,
   *   which will still be run through buildEntityValue(). (If provided, it is
   *   assumed that the value is already fully 'built'.)
   * @param string $sub_dest
   *   'sub' destination name, if only a part of the destination should be set /
   *   overwritten.
   *
   * @return string
   *   If the value was actually changed: a nonempty string, that might
   *   represent a certain 'sub' part code. If not: ''.
   */
  protected function setEntityValue($value, $entity, $destination, $dest_type = NULL, $sub_dest = '') {
    $changed = '';

    if ($dest_type === NULL) {
      list($destination, $dest_type, $sub_dest) = $this->parseDestinationKey($destination);

      // Run the value through buildEntityValue(), for destination(type)
      // based transformation.
      $this->buildEntityValue($value, $destination, NULL, $dest_type);
    }

    if ($sub_dest) {
      // Set value into 'destination + sub' separately from other subs.
      // (In a FieldAPI case: per language code.)
      if ($value) {
        if (!isset($entity->{$destination}[$sub_dest])
            || !$this->isEntityValueEqual($value, $entity->{$destination}[$sub_dest], $dest_type)) {
          $entity->{$destination}[$sub_dest] = $value;
          $changed = $sub_dest;
        }
      }
      elseif (!empty($entity->{$destination}[$sub_dest])) {
        // Is this actually generic? Or should we unset here (and let
        // DrunkinsEntityImportJob extend things where necessary)?
        $entity->{$destination}[$sub_dest] = array();
        $changed = $sub_dest;
      }
    }

    // Set or clear property.
    // Since there are no 'sub parts' for the entity, we just need to return
    // an arbitrary non-empty string when a property is changed.
    elseif (isset($value)) {
      if (!isset($entity->$destination)
          || !$this->isEntityValueEqual($value, $entity->$destination, $dest_type)) {
        $entity->$destination = $value;
        $changed = '*';
      }
    }
    // Clear property, except if the value is equal to the db schema default.
    elseif (isset($entity->$destination)
            && !(isset($this->context['import_default_entity_values'][$destination])
                 && $this->isEntityValueEqual($this->context['import_default_entity_values'][$destination], $entity->$destination))) {
      unset($entity->$destination);
      $changed = '*';
    }

    return $changed;
  }

  /**
   * Gets a value from the source item and transforms it depending on the
   * mapping configuration for this field. The value will be interpreted as
   * multiple separate values if it's an array with 0-based sequential keys.
   *
   * See documentation on the 'process' key(s) in the field mapping.
   *
   * @param string $source_field
   *   Field name in the source item (to get the value from).
   * @param array $process
   *   'process' actions for the item currently being processed. Only needs to
   *   be specified if the default value in the mapping should be overridden.
   *
   * @return array
   *   An array of 'processed' values, all with numeric keys. Usually if a
   *   source field contains one value, this will be an array with one element.
   *
   * Note: extending this method with a custom 'process key' is not easy /
   * involves copying code. (Arguably this method should call a
   * processedItemValue() method.) Will be implemented as need arises.
   * Compatibility may be broken in the future!
   */
  protected function buildItemValues($source_field, array $process = NULL) {

    if ($source_field === '') {
      // With our code / normalized mapping, this happens only if a 'process'
      // function is defined which can further construct a value.
      $values = array(NULL);
    }
    else {
      if ($this->isItemFieldEmpty($source_field)) {
        return array();
      }
      $values = $this->normalizeValue($this->src_item[$source_field]);
    }

    // Get default 'process' function for this dst/src, if necessary.
    if (!isset($process)) {
      $process = isset($this->context['import_field_mapping'][$this->current_dst]["process:$source_field"])
        ? $this->context['import_field_mapping'][$this->current_dst]["process:$source_field"] : array();
    }

    // Execute all actions in order
    foreach ($process as $action => $parameters) {

      // Check for actions 'hardcoded' in this function.
      // (This makes for easier logging / somewhat faster processing than also
      // having those actions all be separate method. It's not very extensible
      // by child classes, though.)
      // If not a known action name, call a method.

      // Prevent 0 from matching the first 'case':
      switch ($action === 0 ? 1 : $action) {

        // The following actions do not process the value, but are checks:

        case 'skip_if_nonempty_src':
          // Skip field if _any_ of the fields defined here is nonempty.

          $fields_to_check = is_array($parameters) ?
            $parameters : array($parameters);

          foreach ($fields_to_check as $field) {
            if (!$this->isItemFieldEmpty($field)) {
              $values = array();
            }
          }
          break;

        case 'skip_if_empty_src':
          // Skip field if _any_ of the fields defined here is empty.

          $fields_to_check = is_array($parameters) ?
            $parameters : array($parameters);

          foreach ($fields_to_check as $field) {
            if ($this->isItemFieldEmpty($field)) {
              $values = array();
            }
          }
          break;

        // The following actions do more processing on the value(s):

        case 'replaceString':
          // Replace substring with another.

          if (!is_array($parameters) || count($parameters) != 2) {
            $this->log('Expected a two-element array as parameter to \'replace\'. Mapping: %d/%s; parameter: %v',
              array('%d' => $this->current_dst, '%s' => $source_field, '%v' => print_r($parameters, TRUE)),
              WATCHDOG_ERROR, FALSE);
          }
          else {
            foreach ($values as &$val) {
              // Take care not to fail on arrays with non-numeric keys.
              $val = str_replace(reset($parameters), end($parameters), $val);
            }
          }
          break;

        case 'booleanSwitch':
          // Replace with 'true or false value'.
          // (Or only replace 'true' value, if first array element is NULL.)

          if (!is_array($parameters) || count($parameters) != 2) {
            $this->log('Expected a two-element array as parameter to \'booleanSwitch\'. Mapping: %d/%s; parameter: %v',
              array('%d' => $this->current_dst, '%s' => $source_field, '%v' => print_r($parameters, TRUE)),
              WATCHDOG_ERROR, FALSE);
          }
          else {
            $vals = array();
            foreach ($values as $val) {
              // Take care not to fail on arrays with non-numeric keys.
              // Go around the possibility that this array has keys:
              $val = $this->isItemValueTrue($val) ?
                end($parameters) : reset($parameters);
              if (isset($val)) {
                $vals[] = $val;
              }
            }
            $values = $vals;
          }
          break;

        default:

          // (maybe) Call a custom method which is defined as
          // method_name => argument, or [numeric] => method_name.
          //
          // (Custom methods do not need to be able to handle NULL as input.)

          $arg = NULL;
          if (is_numeric($action)) {
            if (!is_string($parameters) || !$parameters) {
              $this->log('Expected a string value to call as method. Mapping: %d/%s; non-string value: %v',
                array('%d' => $this->current_dst, '%s' => $source_field, '%v' => print_r($parameters, TRUE)),
                WATCHDOG_ERROR, FALSE);
            }
            else {
              $action = $parameters;
              $parameters = NULL;
            }
          }

          // If it has no uppercase letters; skip it. We assume it's a property
          // used by a child class.
          if (strtolower($action) != $action) {

            // Call method on each individual value we have.
            $vals = array();
            foreach ($values as $val) {

              // It's PHP; we don't have to think about whether or not the method
              // actually implements the parameters.
              try {
                $vals = array_merge(
                  $vals,
                  $this->normalizeValue($this->$action($val, $parameters, $source_field))
                );
              }
              catch (Exception $e) {
                $this->log('Error occured while calling process action %a; value %v is left unchanged: @e',
                  array('%a' => $action, '%v' => print_r($val, TRUE), '@e' => $e->getMessage()), WATCHDOG_ERROR);
              }
            }

            $values = $vals;
          }
      }

      // If any action yields empty array (i.e. all values have been reset),
      // further processing is no use.
      if (!$values) {
        break;
      }
    }

    return $values;
  }

  /**
   * Transforms a single value so it can be set directly into a specific
   * destination/entity property.
   *
   * @param mixed $value
   *   Value to transform. This is one 'atomic' value; for a multi-value
   *   property, this method must be called multiple times to transform each
   *   value separately (passing the result value as $dst_value on
   *   subsequent calls). So if an array is passed in this argument, this means
   *   the array represents one value - and the caller should ensure that its
   *   structure is recognized for this suitable property.
   * @param string $destination
   *   Destination property name.
   * @param mixed $current_built_value
   *   (optional) Value(s) already built by a previous call to this method for
   *   the same destination.
   * @param string $dst_type
   *   (optional) 'type' of destination property. (This class has no support for
   *   them; subclasses can support different logic for transforming different
   *   property types.)
   * @param string $source_field
   *   (optional) Source field from the item where the value came from. This
   *   value should generally not be needed, but it could be useful to reach
   *   properties in the field mapping - or for logging.
   *
   * @return mixed
   *   Property value, transformed if necessary; usually merged into (or
   *   otherwise influenced by) $current_built_value if that was provided.
   */
  protected function buildEntityValue($value, $destination, $current_built_value = NULL, $dst_type = '', $source_field = '') {

    switch ($destination) {

      case 'created':
      case 'changed':
        $value = strtotime($value);
        break;
    }

    // We don't know anything about the property. This standard method
    // implementation assumes it needs to be a scalar value, so we _set_ (not
    // _add to_) $current_built_value. If it has already been set previously, we
    // issue a warning.
    // This method can be extended for each destination (type / name) which
    // knows what it wants to do with multiple values.
    if ($current_built_value && $current_built_value != $value) {
      $this->log('While processing item %item, multiple values (%v and %vv) are set into property %dst? We ignore the second value.',
        array(
          '%v' => print_r($current_built_value, TRUE),
          '%vv' => print_r($value, TRUE),
          '%item' => $this->src_item[$this->context['import_src_pkey']],
        ), WATCHDOG_ERROR, FALSE);
    }
    else {
      $current_built_value = $value;
    }

    return $current_built_value;
  }

  /**
   * Expands a value with unknown structure to an array with zero, one or
   * multiple atomic values.
   *
   * The input value can be various things:
   * - empty, i.e. will not add anything to the output array.
   * - an array of individual 'atoms' with sequential 0-based numeric keys.
   * - a single atomic value.
   *
   * An 'atom' can be
   * - a single scalar value,
   * - a custom value-structure (e.g. a fieldAPI value). Any array with at least
   *   one alphanumeric or non-sequential key is considered an individual
   *   structure.
   *
   * This is used to normalize e.g. input from an item's field, the return value
   * from a process function, or a 'default_values' setting - which can be
   * different things like:
   * - NULL or array():                                  nothing; skip.
   * - 'value1':                                         one value
   * - array('value1', 'value2'):                        two values
   * - array('value' => 'value1', 'format' => 'value2'): one value (struct)
   * - array('value1', 3 => 'value2'):                   also seen as one value.
   *
   * @param mixed $value
   *   Input value which is checked for needed expansion.
   * @return array
   *   A normalized array of atomic values (which are each scalars or structs).
   */
  protected function normalizeValue($value) {

    if (isset($value) && $value !== array()) {
      $i = 0;
      if (is_array($value)) {
        // Check for sequential 0-based numeric keys
        foreach ($value as $key => $v) {
          if ($key !== $i++) {
            // Not 0-based: this is a single value-structure.
            return array($value);
          }
        }
      }
      return $i ? $value : array($value);
    }

    return array();
  }

  /**
   * Checks if a field 'has an empty value', in the sense that it should not be
   * processed/saved.
   */
  protected function isItemFieldEmpty($source_field) {

    if (!isset($this->src_item[$source_field])) {
      return TRUE;
    }

    // By default, we're regarding 'empty string' as 'empty', but not 0 or '0'.
    $value = $this->src_item[$source_field];
    return (!isset($value) || $value === '' || $value === array());
  }

  /**
   * Checks if a value is 'true'.
   * This is for source systems that can encode values, e.g. (string) "false"
   * being not a true value (which it is by default).
   */
  protected function isItemValueTrue($value) {
    return (bool)$value;
  }

  /**
   * Checks if values are equal. (For some structures, that's not so
   * straightforward because the current destination entity's value structure
   * may have cached/meta info which should be disregarded for value
   * comparison.)
   *
   * Info about the destination can be deduced from $this->current_dst.
   * (That's a class property while $dst_type is an argument; it's slightly
   * inconsistent but that's how the code grew - so far.)
   */
  protected function isEntityValueEqual($built_value, $entity_value, $dst_type = '') {
    // This represents what is most likely for mysql storage vs. an unknown
    // fetcher: string values are equal to their numeric equivalent. (The
    // database backend represents every value as a string.) NULL is also equal
    // to ''... because. However 0/"0" is not equal to ''/NULL.
    return (empty($built_value) && empty($entity_value))
      ? ($built_value === $entity_value
         || is_numeric($built_value) && is_numeric($entity_value)
         || $built_value === NULL && $entity_value === ''
         || $built_value === '' && $entity_value === NULL)
      : $built_value == $entity_value;
  }

}
